/*
 *   Copyright © 2021 NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 *     * Neither the name of Intel Corporation nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include <stdint.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <errno.h>
#include "nvme_emu_log.h"
#include "effdma_address_space.h"

typedef struct {
    uint64_t siova;
    uint64_t tiova;
    size_t length;
} address_range_t;

struct address_space {
    size_t len;
    size_t capacity;
    address_range_t *map;
};

static int compare(const void *arg1, const void *arg2)
{
    const address_range_t *a1 = arg1;
    const address_range_t *a2 = arg2;

    if (a1->siova < a2->siova) return -1;
    if (a1->siova > a2->siova) return 1;
    return 0;
}

address_space_t *address_space_init()
{
    int status;
    address_space_t *address_space;

    status = posix_memalign((void **)&address_space, 256, sizeof(address_space_t));

    if (status != 0) {
        nvmx_error("Failed allocating address space, err: %d", status);
        return NULL;
    }

    memset(address_space, 0, sizeof(address_space_t));
    return address_space;
}

void address_space_destroy(address_space_t *address_space)
{
    free(address_space->map);
    free(address_space);
}

address_space_t *address_space_clone(const address_space_t *address_space)
{
    address_space_t *new_address_space;

    new_address_space = address_space_init();
    if (!new_address_space)
        return NULL;

    new_address_space->len = address_space->len;
    new_address_space->capacity = new_address_space->len;
    new_address_space->map = malloc(new_address_space->capacity * sizeof(address_range_t));

    if (!new_address_space->map) {
        address_space_destroy(new_address_space);
        return NULL;
    }

    memcpy(new_address_space->map, address_space->map, new_address_space->capacity * sizeof(address_range_t));

    return new_address_space;
}

int address_space_map(address_space_t *address_space, uint64_t siova, uint64_t tiova, size_t length)
{
    size_t i;
    const uint64_t startA = siova;
    const uint64_t endA = startA + length - 1;

    if (!length) {
        nvmx_error("Zero regions are not allowed");
        return -EINVAL;
    }

    // Check for overlapping. Overlap is not allowed
    for (i = 0; i < address_space->len; i++) {
        const uint64_t startB = address_space->map[i].siova;
        const uint64_t endB = startB + address_space->map[i].length - 1;

        if (startA <= endB && endA >= startB) {
            nvmx_error("Overlapping regions are not allowed");
            return -EINVAL;
        }
    }

    if (address_space->len == address_space->capacity) {
        address_range_t *new_map;
        size_t new_capacity = address_space->capacity + 1;

        new_map = realloc(address_space->map, new_capacity * sizeof(address_range_t));
        if (!new_map)
            return -ENOMEM;

        address_space->map = new_map;
        address_space->capacity = new_capacity;
    }

    address_space->map[address_space->len].siova = siova;
    address_space->map[address_space->len].tiova = tiova;
    address_space->map[address_space->len].length = length;

    address_space->len++;

    qsort(address_space->map, address_space->len, sizeof(address_range_t), compare);

    return 0;
}

int address_space_unmap(address_space_t *address_space, uint64_t siova, size_t length)
{
    size_t index;
    const address_range_t key = {
            .siova = siova
    };
    address_range_t *address_range;

    address_range = bsearch(&key, address_space->map, address_space->len, sizeof(address_range_t), compare);

    if (!address_range)
        return -ENOENT;

    if (length != address_range->length) {
        nvmx_error("Partial unmap not allowed");
        return -EINVAL;
    }

    index = address_range - address_space->map;
    memmove(address_range, address_range + 1, (address_space->len - index - 1) * sizeof(address_range_t));

    address_space->len--;

    return 0;
}

int address_space_unmap_all(address_space_t *address_space)
{
    free(address_space->map);
    memset(address_space, 0, sizeof(address_space_t));
    return 0;
}

int address_space_translate(const address_space_t *address_space, uint64_t siova, uint64_t *tiova)
{
    size_t begin = 0;
    size_t end = address_space->len;

    // Advanced binary search which looks for closest element
    while (begin < end) {
        const size_t pos = (begin + end) / 2;
        const address_range_t *current_range = &address_space->map[pos];
        const address_range_t *next_range = &address_space->map[pos + 1];

        if (siova >= current_range->siova && (pos >= address_space->len - 1 || siova < next_range->siova)) {
            const size_t offset = siova - current_range->siova;
            if (offset >= current_range->length)
                return -ENOENT;
            *tiova = current_range->tiova + offset;
            return 0;
        }

        if (current_range->siova > siova)
            end = pos;
        else
            begin = pos + 1;
    }

    return -ENOENT;
}
