#include "config.h"
#include "json.h"
#include "utf.h"
#include "compiler.h"

#define NVME_JSON_MAX_NESTING_DEPTH  64

enum nvme_json_parse_type {
    NVME_JSON_PARSE_ARRAY,
    NVME_JSON_PARSE_OBJECT,
};

struct nvme_json_obj {
    char name[NVME_JSON_MAX_OBJ_NAME];
    struct nvme_json_obj *next;
    struct nvme_json_obj *prev;
    enum nvme_json_parse_type type;
    unsigned index;
};

static int
hex_value(uint8_t c)
{
#define V(x, y) [x] = y + 1
	static const int8_t val[256] = {
		V('0', 0), V('1', 1), V('2', 2), V('3', 3), V('4', 4),
		V('5', 5), V('6', 6), V('7', 7), V('8', 8), V('9', 9),
		V('A', 0xA), V('B', 0xB), V('C', 0xC), V('D', 0xD), V('E', 0xE), V('F', 0xF),
		V('a', 0xA), V('b', 0xB), V('c', 0xC), V('d', 0xD), V('e', 0xE), V('f', 0xF),
	};
#undef V

	return val[c] - 1;
}

static int
json_decode_string_escape_unicode(uint8_t **strp, uint8_t *buf_end, uint8_t *out)
{
	uint8_t *str = *strp;
	int v0, v1, v2, v3;
	uint32_t val;
	uint32_t surrogate_high = 0;
	int rc;
decode:
	/* \uXXXX */
	assert(buf_end > str);

	if (*str++ != '\\') { return NVME_JSON_PARSE_INVALID; }
	if (buf_end == str) { return NVME_JSON_PARSE_INCOMPLETE; }

	if (*str++ != 'u') { return NVME_JSON_PARSE_INVALID; }
	if (buf_end == str) { return NVME_JSON_PARSE_INCOMPLETE; }

	if ((v3 = hex_value(*str++)) < 0) { return NVME_JSON_PARSE_INVALID; }
	if (buf_end == str) { return NVME_JSON_PARSE_INCOMPLETE; }

	if ((v2 = hex_value(*str++)) < 0) { return NVME_JSON_PARSE_INVALID; }
	if (buf_end == str) { return NVME_JSON_PARSE_INCOMPLETE; }

	if ((v1 = hex_value(*str++)) < 0) { return NVME_JSON_PARSE_INVALID; }
	if (buf_end == str) { return NVME_JSON_PARSE_INCOMPLETE; }

	if ((v0 = hex_value(*str++)) < 0) { return NVME_JSON_PARSE_INVALID; }
	if (buf_end == str) { return NVME_JSON_PARSE_INCOMPLETE; }

	val = v0 | (v1 << 4) | (v2 << 8) | (v3 << 12);

	if (surrogate_high) {
		/* We already parsed the high surrogate, so this should be the low part. */
		if (!utf16_valid_surrogate_low(val)) {
			return NVME_JSON_PARSE_INVALID;
		}

		/* Convert UTF-16 surrogate pair into codepoint and fall through to utf8_encode. */
		val = utf16_decode_surrogate_pair(surrogate_high, val);
	} else if (utf16_valid_surrogate_high(val)) {
		surrogate_high = val;

		/*
		 * We parsed a \uXXXX sequence that decoded to the first half of a
		 *  UTF-16 surrogate pair, so it must be immediately followed by another
		 *  \uXXXX escape.
		 *
		 * Loop around to get the low half of the surrogate pair.
		 */
		if (buf_end == str) { return NVME_JSON_PARSE_INCOMPLETE; }
		goto decode;
	} else if (utf16_valid_surrogate_low(val)) {
		/*
		 * We found the second half of surrogate pair without the first half;
		 *  this is an invalid encoding.
		 */
		return NVME_JSON_PARSE_INVALID;
	}

	/*
	 * Convert Unicode escape (or surrogate pair) to UTF-8 in place.
	 *
	 * This is safe (will not write beyond the buffer) because the \uXXXX sequence is 6 bytes
	 *  (or 12 bytes for surrogate pairs), and the longest possible UTF-8 encoding of a
	 *  single codepoint is 4 bytes.
	 */
	if (out) {
		rc = utf8_encode_unsafe(out, val);
	} else {
		rc = utf8_codepoint_len(val);
	}
	if (rc < 0) {
		return NVME_JSON_PARSE_INVALID;
	}

	*strp = str; /* update input pointer */
	return rc; /* return number of bytes decoded */
}

static int
json_decode_string_escape_twochar(uint8_t **strp, uint8_t *buf_end, uint8_t *out)
{
	static const uint8_t escapes[256] = {
		['b'] = '\b',
		['f'] = '\f',
		['n'] = '\n',
		['r'] = '\r',
		['t'] = '\t',
		['/'] = '/',
		['"'] = '"',
		['\\'] = '\\',
	};
	uint8_t *str = *strp;
	uint8_t c;

	assert(buf_end > str);
	if (buf_end - str < 2) {
		return NVME_JSON_PARSE_INCOMPLETE;
	}

	assert(str[0] == '\\');

	c = escapes[str[1]];
	if (c) {
		if (out) {
			*out = c;
		}
		*strp += 2; /* consumed two bytes */
		return 1; /* produced one byte */
	}

	return NVME_JSON_PARSE_INVALID;
}

/*
 * Decode JSON string backslash escape.
 * \param strp pointer to pointer to first character of escape (the backslash).
 *  *strp is also advanced to indicate how much input was consumed.
 *
 * \return Number of bytes appended to out
 */
static int
json_decode_string_escape(uint8_t **strp, uint8_t *buf_end, uint8_t *out)
{
	int rc;

	rc = json_decode_string_escape_twochar(strp, buf_end, out);
	if (rc > 0) {
		return rc;
	}

	return json_decode_string_escape_unicode(strp, buf_end, out);
}

/*
 * Decode JSON string in place.
 *
 * \param str_start Pointer to the beginning of the string (the opening " character).
 *
 * \return Number of bytes in decoded string (beginning from start).
 */
static int
json_decode_string(uint8_t *str_start, uint8_t *buf_end, uint8_t **str_end, uint32_t flags)
{
	uint8_t *str = str_start;
	uint8_t *out = str_start + 1; /* Decode string in place (skip the initial quote) */
	int rc;

	if (buf_end - str_start < 2) {
		/*
		 * Shortest valid string (the empty string) is two bytes (""),
		 *  so this can't possibly be valid
		 */
		*str_end = str;
		return NVME_JSON_PARSE_INCOMPLETE;
	}

	if (*str++ != '"') {
		*str_end = str;
		return NVME_JSON_PARSE_INVALID;
	}

	while (str < buf_end) {
		if (str[0] == '"') {
			/*
			 * End of string.
			 * Update str_end to point at next input byte and return output length.
			 */
			*str_end = str + 1;
			return out - str_start - 1;
		} else if (str[0] == '\\') {
			rc = json_decode_string_escape(&str, buf_end,
						       flags & NVME_JSON_PARSE_FLAG_DECODE_IN_PLACE ? out : NULL);
			assert(rc != 0);
			if (rc < 0) {
				*str_end = str;
				return rc;
			}
			out += rc;
		} else if (str[0] <= 0x1f) {
			/* control characters must be escaped */
			*str_end = str;
			return NVME_JSON_PARSE_INVALID;
		} else {
			rc = utf8_valid(str, buf_end);
			if (rc == 0) {
				*str_end = str;
				return NVME_JSON_PARSE_INCOMPLETE;
			} else if (rc < 0) {
				*str_end = str;
				return NVME_JSON_PARSE_INVALID;
			}

			if (out && out != str && (flags & NVME_JSON_PARSE_FLAG_DECODE_IN_PLACE)) {
				memmove(out, str, rc);
			}
			out += rc;
			str += rc;
		}
	}

	/* If execution gets here, we ran out of buffer. */
	*str_end = str;
	return NVME_JSON_PARSE_INCOMPLETE;
}

static int
json_valid_number(uint8_t *start, uint8_t *buf_end)
{
    uint8_t *p = start;
    uint8_t c;

    if (p >= buf_end) { return -1; }

    c = *p++;
    if (c >= '1' && c <= '9') { goto num_int_digits; }

    if (c == '0') {
        c = *p;

        if (c == 'x' || c == 'X') {
            ++p;
            goto num_hex_first_digit;
        }

        goto num_frac_or_exp;
    }

    if (c == '-') { goto num_int_first_digit; }
    p--;
    goto done_invalid;

num_int_first_digit:
    if (likely(p != buf_end)) {
        c = *p++;
        if (c == '0') { goto num_frac_or_exp; }
        if (c >= '1' && c <= '9') { goto num_int_digits; }
        p--;
    }
    goto done_invalid;

num_int_digits:
	if (likely(p != buf_end)) {
		c = *p++;
		if (c >= '0' && c <= '9') { goto num_int_digits; }
		if (c == '.') { goto num_frac_first_digit; }
		if (c == 'e' || c == 'E') { goto num_exp_sign; }
		p--;
	}
	goto done_valid;

num_hex_first_digit:
    if (hex_value(*p) < 0) {
        return NVME_JSON_PARSE_INVALID;
    }
    goto num_hex_digits;

num_hex_digits:
    if (likely(p != buf_end)) {
        c = *p++;

        if (hex_value(c) >= 0) {
            goto num_hex_digits;
        }

        p--;
    }
    goto done_valid;

num_frac_or_exp:
	if (likely(p != buf_end)) {
		c = *p++;
		if (c == '.') { goto num_frac_first_digit; }
		if (c == 'e' || c == 'E') { goto num_exp_sign; }
		p--;
	}
	goto done_valid;

num_frac_first_digit:
	if (likely(p != buf_end)) {
		c = *p++;
		if (c >= '0' && c <= '9') { goto num_frac_digits; }
		p--;
	}
	goto done_invalid;

num_frac_digits:
	if (likely(p != buf_end)) {
		c = *p++;
		if (c >= '0' && c <= '9') { goto num_frac_digits; }
		if (c == 'e' || c == 'E') { goto num_exp_sign; }
		p--;
	}
	goto done_valid;

num_exp_sign:
	if (likely(p != buf_end)) {
		c = *p++;
		if (c >= '0' && c <= '9') { goto num_exp_digits; }
		if (c == '-' || c == '+') { goto num_exp_first_digit; }
		p--;
	}
	goto done_invalid;

num_exp_first_digit:
	if (likely(p != buf_end)) {
		c = *p++;
		if (c >= '0' && c <= '9') { goto num_exp_digits; }
		p--;
	}
	goto done_invalid;

num_exp_digits:
	if (likely(p != buf_end)) {
		c = *p++;
		if (c >= '0' && c <= '9') { goto num_exp_digits; }
		p--;
	}
	goto done_valid;

done_valid:
	/* Valid end state */
	return p - start;

done_invalid:
	/* Invalid end state */
	if (p == buf_end) {
		/* Hit the end of the buffer - the stream is incomplete. */
		return NVME_JSON_PARSE_INCOMPLETE;
	}

	/* Found an invalid character in an invalid end state */
	return NVME_JSON_PARSE_INVALID;
}

static int
json_valid_comment(const uint8_t *start, const uint8_t *buf_end)
{
	const uint8_t *p = start;
	bool multiline;

	assert(buf_end > p);
	if (buf_end - p < 2) {
		return NVME_JSON_PARSE_INCOMPLETE;
	}

	if (p[0] != '/') {
		return NVME_JSON_PARSE_INVALID;
	}
	if (p[1] == '*') {
		multiline = true;
	} else if (p[1] == '/') {
		multiline = false;
	} else {
		return NVME_JSON_PARSE_INVALID;
	}
	p += 2;

	if (multiline) {
		while (p != buf_end - 1) {
			if (p[0] == '*' && p[1] == '/') {
				/* Include the terminating star and slash in the comment */
				return p - start + 2;
			}
			p++;
		}
	} else {
		while (p != buf_end) {
			if (*p == '\r' || *p == '\n') {
				/* Do not include the line terminator in the comment */
				return p - start;
			}
			p++;
		}
	}

	return NVME_JSON_PARSE_INCOMPLETE;
}

struct json_literal {
	enum nvme_json_val_type type;
	uint32_t len;
	uint8_t str[8];
};

/*
 * JSON only defines 3 possible literals; they can be uniquely identified by bits
 *  3 and 4 of the first character:
 *   'f' = 0b11[00]110
 *   'n' = 0b11[01]110
 *   't' = 0b11[10]100
 * These two bits can be used as an index into the g_json_literals array.
 */
static const struct json_literal g_json_literals[] = {
	{NVME_JSON_VAL_FALSE, 5, "false"},
	{NVME_JSON_VAL_NULL,  4, "null"},
	{NVME_JSON_VAL_TRUE,  4, "true"},
	{}
};

static int
match_literal(const uint8_t *start, const uint8_t *end, const uint8_t *literal, size_t len)
{
	assert(end >= start);
	if ((size_t)(end - start) < len) {
		return NVME_JSON_PARSE_INCOMPLETE;
	}

	if (memcmp(start, literal, len) != 0) {
		return NVME_JSON_PARSE_INVALID;
	}

	return len;
}

ssize_t
nvme_json_parse(void *json, size_t size, struct nvme_json_val *values, size_t num_values,
		void **end, uint32_t flags)
{
	uint8_t *json_end = json + size;
	enum nvme_json_val_type containers[NVME_JSON_MAX_NESTING_DEPTH];
	size_t con_value[NVME_JSON_MAX_NESTING_DEPTH];
	enum nvme_json_val_type con_type = NVME_JSON_VAL_INVALID;
	bool trailing_comma = false;
	size_t depth = 0; /* index into containers */
	size_t cur_value = 0; /* index into values */
	size_t con_start_value;
	uint8_t *data = json;
	uint8_t *new_data;
	int rc = 0;
	const struct json_literal *lit;
	enum {
		STATE_VALUE, /* initial state */
		STATE_VALUE_SEPARATOR, /* value separator (comma) */
		STATE_NAME, /* "name": value */
		STATE_NAME_SEPARATOR, /* colon */
		STATE_END, /* parsed the complete value, so only whitespace is valid */
	} state = STATE_VALUE;

#define ADD_VALUE(t, val_start_ptr, val_end_ptr) \
	if (values && cur_value < num_values) { \
		values[cur_value].type = t; \
		values[cur_value].start = val_start_ptr; \
		values[cur_value].len = val_end_ptr - val_start_ptr; \
	} \
	cur_value++

	while (data < json_end) {
		uint8_t c = *data;

		switch (c) {
		case ' ':
		case '\t':
		case '\r':
		case '\n':
			/* Whitespace is allowed between any tokens. */
			data++;
			break;

		case 't':
		case 'f':
		case 'n':
			/* true, false, or null */
			if (state != STATE_VALUE) { goto done_invalid; }
			lit = &g_json_literals[(c >> 3) & 3]; /* See comment above g_json_literals[] */
			assert(lit->str[0] == c);
			rc = match_literal(data, json_end, lit->str, lit->len);
			if (rc < 0) { goto done_rc; }
			ADD_VALUE(lit->type, data, data + rc);
			data += rc;
			state = depth ? STATE_VALUE_SEPARATOR : STATE_END;
			trailing_comma = false;
			break;

		case '"':
			if (state != STATE_VALUE && state != STATE_NAME) { goto done_invalid; }
			rc = json_decode_string(data, json_end, &new_data, flags);
			if (rc < 0) {
				data = new_data;
				goto done_rc;
			}
			/*
			 * Start is data + 1 to skip initial quote.
			 * Length is data + rc - 1 to skip both quotes.
			 */
			ADD_VALUE(state == STATE_VALUE ? NVME_JSON_VAL_STRING : NVME_JSON_VAL_NAME,
				  data + 1, data + rc - 1);
			data = new_data;
			if (state == STATE_NAME) {
				state = STATE_NAME_SEPARATOR;
			} else {
				state = depth ? STATE_VALUE_SEPARATOR : STATE_END;
			}
			trailing_comma = false;
			break;

		case '-':
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
			if (state != STATE_VALUE) { goto done_invalid; }
			rc = json_valid_number(data, json_end);
			if (rc < 0) { goto done_rc; }
			ADD_VALUE(NVME_JSON_VAL_NUMBER, data, data + rc);
			data += rc;
			state = depth ? STATE_VALUE_SEPARATOR : STATE_END;
			trailing_comma = false;
			break;

		case '{':
		case '[':
			if (state != STATE_VALUE) { goto done_invalid; }
			if (depth == NVME_JSON_MAX_NESTING_DEPTH) {
				rc = NVME_JSON_PARSE_MAX_DEPTH_EXCEEDED;
				goto done_rc;
			}
			if (c == '{') {
				con_type = NVME_JSON_VAL_OBJECT_BEGIN;
				state = STATE_NAME;
			} else {
				con_type = NVME_JSON_VAL_ARRAY_BEGIN;
				state = STATE_VALUE;
			}
			con_value[depth] = cur_value;
			containers[depth++] = con_type;
			ADD_VALUE(con_type, data, data + 1);
			data++;
			trailing_comma = false;
			break;

		case '}':
		case ']':
			if (trailing_comma) { goto done_invalid; }
			if (depth == 0) { goto done_invalid; }
			con_type = containers[--depth];
			con_start_value = con_value[depth];
			if (values && con_start_value < num_values) {
				values[con_start_value].len = cur_value - con_start_value - 1;
			}
			if (c == '}') {
				if (state != STATE_NAME && state != STATE_VALUE_SEPARATOR) {
					goto done_invalid;
				}
				if (con_type != NVME_JSON_VAL_OBJECT_BEGIN) {
					goto done_invalid;
				}
				ADD_VALUE(NVME_JSON_VAL_OBJECT_END, data, data + 1);
			} else {
				if (state != STATE_VALUE && state != STATE_VALUE_SEPARATOR) {
					goto done_invalid;
				}
				if (con_type != NVME_JSON_VAL_ARRAY_BEGIN) {
					goto done_invalid;
				}
				ADD_VALUE(NVME_JSON_VAL_ARRAY_END, data, data + 1);
			}
			con_type = depth == 0 ? NVME_JSON_VAL_INVALID : containers[depth - 1];
			data++;
			state = depth ? STATE_VALUE_SEPARATOR : STATE_END;
			trailing_comma = false;
			break;

		case ',':
			if (state != STATE_VALUE_SEPARATOR) { goto done_invalid; }
			data++;
			assert(con_type == NVME_JSON_VAL_ARRAY_BEGIN ||
			       con_type == NVME_JSON_VAL_OBJECT_BEGIN);
			state = con_type == NVME_JSON_VAL_ARRAY_BEGIN ? STATE_VALUE : STATE_NAME;
			trailing_comma = true;
			break;

		case ':':
			if (state != STATE_NAME_SEPARATOR) { goto done_invalid; }
			data++;
			state = STATE_VALUE;
			break;

		case '/':
			if (!(flags & NVME_JSON_PARSE_FLAG_ALLOW_COMMENTS)) {
				goto done_invalid;
			}
			rc = json_valid_comment(data, json_end);
			if (rc < 0) { goto done_rc; }
			/* Skip over comment */
			data += rc;
			break;

		default:
			goto done_invalid;
		}

		if (state == STATE_END) {
			break;
		}
	}

	if (state == STATE_END) {
		/* Skip trailing whitespace */
		while (data < json_end) {
			uint8_t c = *data;

			if (c == ' ' || c == '\t' || c == '\r' || c == '\n') {
				data++;
			} else {
				break;
			}
		}

		/*
		 * These asserts are just for sanity checking - they are guaranteed by the allowed
		 *  state transitions.
		 */
		assert(depth == 0);
		assert(trailing_comma == false);
		assert(data <= json_end);
		if (end) {
			*end = data;
		}
		return cur_value;
	}

	/* Invalid end state - ran out of data */
	rc = NVME_JSON_PARSE_INCOMPLETE;

done_rc:
	assert(rc < 0);
	if (end) {
		*end = data;
	}
	return rc;

done_invalid:
	rc = NVME_JSON_PARSE_INVALID;
	goto done_rc;
}

nvme_json_t *nvme_str_to_json(char *str)
{
    size_t str_len, num_values;
    nvme_json_t *json;

    if (!str)
        return NULL;

    json = calloc(1, sizeof(nvme_json_t));
    if (!json)
        return NULL;

    str_len = strlen(str);
    num_values = nvme_json_parse(str, str_len, NULL, 0, NULL,
                    NVME_JSON_PARSE_FLAG_ALLOW_COMMENTS);
    if (num_values <= 0)
        return NULL;

    json->values = calloc(num_values, sizeof(struct nvme_json_val));
    if (!json->values)
        goto free_json;

    num_values = nvme_json_parse(str, str_len, json->values, num_values, NULL,
        NVME_JSON_PARSE_FLAG_ALLOW_COMMENTS | NVME_JSON_PARSE_FLAG_DECODE_IN_PLACE);
    if (num_values <= 0)
        goto free_values;

    json->buf = strdup(str);
    if (!json->buf)
        goto free_values;

    json->num_values = num_values;

    return json;

free_values:
    free(json->values);

free_json:
    free(json);

    return NULL;
}

static void *
read_file(FILE *f, size_t *psize)
{
	void *buf, *newbuf;
	size_t cur_size, buf_size, rc;

	buf = NULL;
	cur_size = 0;
	buf_size = 128 * 1024;

	while (buf_size <= 1024 * 1024 * 1024) {
		newbuf = realloc(buf, buf_size);
		if (newbuf == NULL) {
			free(buf);
			return NULL;
		}
		buf = newbuf;

		rc = fread(buf + cur_size, 1, buf_size - cur_size, f);
		cur_size += rc;

		if (feof(f)) {
			*psize = cur_size;
			return buf;
		}

		if (ferror(f)) {
			free(buf);
			return NULL;
		}

		buf_size *= 2;
	}

	free(buf);
	return NULL;
}

void nvme_json_config_close(nvme_config_t *config)
{
    free(config->values);
    free(config->buf);
}

void nvme_json_free(nvme_config_t *json)
{
    nvme_json_config_close(json);
    free(json);
}

int
nvme_json_parse_config(const char *filename,
                       nvme_config_t *config)
{
	size_t size;
	void *buf, *end;
	ssize_t rc;
	struct nvme_json_val *values;
	size_t num_values;

    FILE *f = fopen(filename, "r");
	if (f == NULL) {
		fprintf(stderr, "%s: file open error\n", filename);
		return -1;
	}

	buf = read_file(f, &size);
	if (buf == NULL) {
		fprintf(stderr, "%s: file read error\n", filename);
		goto close_file;
	}

	rc = nvme_json_parse(buf, size, NULL, 0, NULL, NVME_JSON_PARSE_FLAG_ALLOW_COMMENTS);
	if (rc <= 0) {
		print_json_error(stderr, rc, filename);
        goto free_buf;
	}

	num_values = (size_t)rc;
	values = calloc(num_values, sizeof(*values));
	if (values == NULL) {
		perror("values calloc");
        goto free_buf;
	}

	rc = nvme_json_parse(buf, size, values, num_values, &end,
			     NVME_JSON_PARSE_FLAG_ALLOW_COMMENTS |
                 NVME_JSON_PARSE_FLAG_DECODE_IN_PLACE);
	if (rc <= 0) {
		print_json_error(stderr, rc, filename);
        goto free_values;
	}

    fclose(f);

    config->buf = buf;
    config->values = values;
    config->num_values = num_values;

    return 0;

free_values:
    free(values);
free_buf:
    free(buf);
close_file:
    fclose(f);

    return -1;
}

static int __get_number(struct nvme_json_val *value,
                        void *out, size_t size)
{
    long long int v;

    switch (size) {
        case 8:
            return nvme_json_decode_int64(value, out);
        case 4:
            return nvme_json_decode_int32(value, out);
        case 2:
            return nvme_json_decode_int16(value, out);
        case 1:
            v = strtoull((char *) value->start, NULL, 0);
            if (v > 255 || -v > 127)
                return -1;

            *(int8_t *) out = (int8_t) v;
            break;
        default:
            fprintf(stderr, "Wrong size %ld for integer value\n", size);
            return -1;
    }

    return 0;
}

static int __get_string(struct nvme_json_val *value,
                        void *out, size_t size)
{
    size_t len = (size < value->len + 1 ? size : value->len + 1);

    if (value && out)
        snprintf((char *) out, len, "%s", (char *) value->start);
    return 0;
}

static int __json_covert_value(struct nvme_json_val *value,
                               void *out, size_t size)
{
    switch (value->type) {
        case NVME_JSON_VAL_TRUE:
        case NVME_JSON_VAL_FALSE:
            return nvme_json_decode_bool(value, out);
        case NVME_JSON_VAL_NUMBER:
            return __get_number(value, out, size);
        case NVME_JSON_VAL_STRING:
            return __get_string(value, out, size);
        case NVME_JSON_VAL_NULL:
        case NVME_JSON_VAL_INVALID:
        default:
            break;
    }

    return -1;
}

static void nvme_json_parse_free(struct nvme_json_obj *objs)
{
    while (NULL != objs) {
        struct nvme_json_obj *curr = objs;

        if (NULL != curr->next) {
            curr->next->prev = curr->prev;
            objs = objs->next;
        } else if (NULL != curr->prev) {
            curr->prev->next = curr->next;
            objs = objs->prev;
        } else {
            objs = NULL;
        }

        free(curr);
    }
}

struct nvme_json_obj *nvme_json_parse_path(const char *path)
{
    int i;

    char *token, *tmp;
    char buff[1024], name[NVME_JSON_MAX_OBJ_NAME];

    struct nvme_json_obj *objs = NULL;
    struct nvme_json_obj *curr = NULL;

    struct nvme_json_obj *prev = NULL;
    struct nvme_json_obj **next = &objs;

    sprintf(buff, "%s", path);
    for (i = 0; '\0' != buff[i]; ++i) {
        if (' ' == buff[i] || '\t' == buff[i]) {
            return NULL;
        }

        if ('[' == buff[i]) {
            buff[i] = ' ';
            continue;
        }

        if (']' == buff[i]) {
            buff[i] = '\t';
        }
    }

    for (token = strtok_r(buff, ".", &tmp); NULL != token; token = strtok_r(NULL, ".", &tmp)) {
        /*
         * Each next/prev pointer inside the struct
         * will be set to NULL by calloc
         */
        *next = calloc(1, sizeof *objs);
        if (NULL == *next) {
            goto error_path;
        }

        curr = *next;
        next = &curr->next;

        curr->prev = prev;
        prev = curr;

        i = -1;

        sscanf(token, "%s %d\t", name, &i);
        sprintf(curr->name, "%s", name);

        curr->index = i;
        curr->type = (i < 0) ? NVME_JSON_PARSE_OBJECT :
                               NVME_JSON_PARSE_ARRAY;
    }

    return objs;

error_path:
    nvme_json_parse_free(objs);

    return NULL;
}

static int __retrieve_value(struct nvme_json_val *values,
                            size_t num_values,
                            const char *param,
                            void *out,
                            size_t size)
{
    int i, in_obj = 0;

    /* Retrieve the value form the given object */
    for (i = 0; i < num_values; ++i) {
        if (NVME_JSON_VAL_NAME == values[i].type &&
                values[i].len == strlen(param) &&
                    !strncmp(values[i].start, param, values[i].len) &&
                    in_obj == 1) {
            return __json_covert_value(&values[i+1], out, size);
        }

        if (NVME_JSON_VAL_OBJECT_BEGIN == values[i].type) {
            ++in_obj;
            continue;
        }

        if (NVME_JSON_VAL_OBJECT_END == values[i].type && --in_obj == 0) {
            break;
        }
    }

    return -1;
}

static int __obj_index(struct nvme_json_val *values,
                       size_t num_values,
                       const char *obj_name)
{
    int i;

    for (i = 0; i < num_values; ++i) {
        if (NVME_JSON_VAL_NAME == values[i].type &&
               !strncmp(values[i].start, obj_name, values[i].len) &&
                  values[i].len == strlen(obj_name) &&
                  NVME_JSON_VAL_NULL != values[i+1].type) {
            break;
        }
    }

    return i;
}

static int __arr_elem_index(struct nvme_json_val *values,
                            size_t num_values,
                            unsigned index)
{
    int i, k = 0, in_obj = 0, in_arr = 0;

    for (i = 0; i < num_values && k < index; ++i) {
        if (NVME_JSON_VAL_ARRAY_BEGIN == values[i].type) {
            ++in_arr;
            continue;
        }

        if (NVME_JSON_VAL_ARRAY_END == values[i].type && --in_arr == 0) {
            return -1;
        }

        if (NVME_JSON_VAL_OBJECT_BEGIN == values[i].type) {
            ++in_obj;
            continue;
        }

        if (NVME_JSON_VAL_OBJECT_END == values[i].type && --in_obj == 0) {
            ++k;
        }
    }

    return i;
}

static int __enter_object(struct nvme_json_val *values,
                          size_t num_values,
                          struct nvme_json_obj *obj)
{
    int i, j;

    if (0 == num_values) {
        return -1;
    }

    i = __obj_index(values, num_values, obj->name);

    if (NVME_JSON_PARSE_ARRAY == obj->type) {
        j = __arr_elem_index(&values[i], num_values - i, obj->index);

        if (j < 0) {
            return -1;
        }

        i += j;
    }

    return i;
}

static int __pass_objs(struct nvme_json_val *values,
                       size_t num_values,
                       struct nvme_json_obj *objs)
{
    int i, id = 0;

    for (i = 0; NULL != objs; objs = objs->next, id += i) {
        i = __enter_object(&values[id], num_values - id, objs);
        if (i < 0) {
            return NVME_JSON_OBJ_INVALID_ID;
        }
    }

    return id;
}

static int __calc_elems(struct nvme_json_val *values,
                        size_t num_values)
{
    int i, k = 0, in_obj = 0, in_arr = 0;

    for (i = 0; i < num_values; ++i) {
        if (NVME_JSON_VAL_ARRAY_BEGIN == values[i].type) {
            ++in_arr;
            continue;
        }

        if (NVME_JSON_VAL_ARRAY_END == values[i].type && --in_arr == 0) {
            return k;
        }

        if (NVME_JSON_VAL_OBJECT_BEGIN == values[i].type) {
            ++in_obj;
            continue;
        }

        if (NVME_JSON_VAL_OBJECT_END == values[i].type && --in_obj == 0) {
            ++k;
        }
    }

    return -1;
}

static int __calc_array_elems(struct nvme_json_val *values,
                              size_t num_values,
                              struct nvme_json_obj *arr)
{
    int i;

    i = __enter_object(values, num_values, arr);
    if (i < 0) {
        return i;
    }

    return __calc_elems(&values[i], num_values - i);
}

struct nvme_json_obj*
__get_tail(struct nvme_json_obj *objs, int remove)
{
    while (NULL != objs->next) {
        objs = objs->next;
    }

    if (remove && NULL != objs->prev) {
        objs->prev->next = NULL;
        objs->prev = NULL;
    }

    return objs;
}

int nvme_json_get_obj_value(nvme_config_t *config,
                            nvme_json_obj_id_t id,
                            const char *param,
                            void *out,
                            size_t size)
{
    if (id < 0) {
        return -1;
    }

    return __retrieve_value(&config->values[id],
                            config->num_values - id,
                            param, out, size);
}

nvme_json_obj_id_t nvme_json_get_obj(nvme_config_t *config,
                                     const char *path)
{
    nvme_json_obj_id_t id;
    struct nvme_json_obj *objs = nvme_json_parse_path(path);

    if (NULL == objs) {
        return NVME_JSON_OBJ_INVALID_ID;
    }

    id = __pass_objs(config->values, config->num_values, objs);
    nvme_json_parse_free(objs);

    return id;
}

static int __get_array_size(struct nvme_json_val *values,
                            size_t num_values,
                            struct nvme_json_obj *objs)
{
    nvme_json_obj_id_t id;
    struct nvme_json_obj *param;

    if (NULL == objs || 0 == num_values) {
        return -1;
    }

    id = __pass_objs(values, num_values, objs);
    if (NVME_JSON_OBJ_INVALID_ID == id) {
        return -1;
    }

    param = __get_tail(objs, 0);

    return __calc_array_elems(&values[id], num_values - id, param);
}

int nvme_json_get_array_size(nvme_config_t *config,
                             const char *path)
{
    int size;
    struct nvme_json_obj *objs = nvme_json_parse_path(path);

    if (NULL == objs) {
        return -1;
    }

    size = __get_array_size(config->values, config->num_values, objs);
    nvme_json_parse_free(objs);

    return size;
}

static int __get_value(struct nvme_json_val *values,
                       size_t num_values,
                       struct nvme_json_obj *objs,
                       void *out,
                       size_t size)
{
    int rc;

    nvme_json_obj_id_t id;
    struct nvme_json_obj *param;

    if (NULL == objs || 0 == num_values) {
        return -1;
    }

    param = __get_tail(objs, 1);

    if (param == objs)
        return __retrieve_value(values, num_values,
                               param->name, out, size);

    id = __pass_objs(values, num_values, objs);
    if (NVME_JSON_OBJ_INVALID_ID == id) {
        rc = -1;
        goto exit;
    }

    rc = __retrieve_value(&values[id], num_values - id,
                          param->name, out, size);

exit:
     free(param);
     return rc;
}

int nvme_json_get_value(nvme_config_t *config,
                        const char *path,
                        void *out,
                        size_t size)
{
    int rc;
    struct nvme_json_obj *objs = nvme_json_parse_path(path);

    if (NULL == objs) {
        return -1;
    }

    rc = __get_value(config->values, config->num_values, objs, out, size);
    nvme_json_parse_free(objs);

    return rc;
}

int nvme_json_verify_value(nvme_config_t *config, const char* path, int value)
{
    int rc, conf_value;

    rc = nvme_json_get_value(config, path, &conf_value, sizeof(conf_value));
    if (rc < 0) {
        return rc;
    } else if (conf_value != value) {
        return -EINVAL;
    }

    return 0;
}

int nvme_json_verify_lte_value(nvme_config_t *config, const char* path,
                               int exp_value)
{
    int rc, conf_value;

    rc = nvme_json_get_value(config, path, &conf_value, sizeof(conf_value));
    if (rc < 0) {
        return rc;
    } else if (conf_value > exp_value) {
        return -EINVAL;
    }

    return 0;
}

int nvme_read_config_file(nvme_config_t *config,
                          const char *env,
                          const char *def)
{
    const char *user_config = getenv(env);

    return nvme_json_parse_config(NULL == user_config ?
                                  def : user_config,
                                  config);
}
