#include "config.h"
#include "nvme_json.h"
#include "nvme_emu_log.h"
#include "utf.h"


void nvme_json_get_quirks(nvme_config_t *config, enum nvme_quirks *quirks)
{
    int rc;
    enum nvme_quirks json_quirks;

    rc = nvme_json_get_value(config, "ctrl.quirks",
                             &json_quirks, sizeof(json_quirks));
    if (rc < 0)
        json_quirks = 0;

    if (json_quirks)
        nvmx_info("host NVME driver quirks: 0x%x", json_quirks);

    *quirks = json_quirks;
}

static int nvme_parse_hex_be(uint8_t *raw, char *value, int num_bytes)
{
    int i, b;
    uint8_t byte;
    size_t len = strlen(value);

    memset(raw, 0, num_bytes);

    /*
     * Go through the string that represents num_bytes bit
     * hex integer in BE and convert each two adjacent hex digits to the byte.
     * We are going from the end in order to keep BE order of the final result.
     */
    b = num_bytes - 1;
    for (i = len - 1; i >= 0 && b >= 0; --i) {
        char buf[3] = {0};

        if (i - 1 >= 0)
            buf[1] = value[i--];

        buf[0] = value[i];

        byte = (uint8_t) strtoul(buf, NULL, 16);
        memcpy(&raw[b--], &byte, sizeof(byte));
    }

    return 0;
}

static int nvme_eui64_parse(nvme_eui64_t *id, char *value)
{
    int rc;

    rc = nvme_parse_hex_be(id->raw, value, 8);
    if (rc) {
        nvmx_info("failed to parse eui64");
        return rc;
    }

    nvmx_debug("eui64: %x%x%x%x%x%x%x%x", id->raw[0], id->raw[1], id->raw[2],
               id->raw[3], id->raw[4], id->raw[5], id->raw[6], id->raw[7]);

    return 0;
}

static int nvme_nguid_parse(nvme_ngiud_t *id, char *value)
{
    int rc;

    rc = nvme_parse_hex_be(id->raw, value, 16);
    if (rc) {
        nvmx_info("failed to parse nguid");
        return rc;
    }

    nvmx_debug("nguid: %x%x%x%x%x%x%x%x%x%x%x%x%x%x%x%x",
               id->raw[0], id->raw[1], id->raw[2], id->raw[3], id->raw[4],
               id->raw[5], id->raw[6], id->raw[7], id->raw[8], id->raw[9],
               id->raw[10], id->raw[11], id->raw[12], id->raw[13], id->raw[14],
               id->raw[15]);

    return 0;
}

static int nvme_uuid_parse(nvme_uuid_t *id, char *value)
{
    int i, rc;
    long long node;
    int time_low;
    int time_mid;
    int time_hi_and_version;
    int clk_seq;
    char *tmp;
    size_t len = strlen(value);

    tmp = (char *) malloc(len + 1);
    if (!tmp)
        return -ENOMEM;

    memcpy(tmp, value, len + 1);

    for (i = 0; '\0' != tmp[i]; ++i) {
        if ('-' == tmp[i]) {
            tmp[i] = ' ';
            continue;
        }
    }

    rc = sscanf(tmp, "%x %x %x %x %llx", &time_low, &time_mid,
                &time_hi_and_version, &clk_seq, &node);
    if (rc == 1 && time_low == 0) {
        /* uuid set to zero, i.e. not defined */
        rc = 0;
        goto exit;
    } else if (rc != 5) {
        nvmx_error("Wrong uuid format");
        rc = -EINVAL;
        goto exit;
    } else {
        rc = 0;
    }

    memcpy(&id->bits.node.byte1, &node, sizeof(id->bits.node.byte1));
    memcpy(&id->bits.node.byte5, (char *) &node + sizeof(id->bits.node.byte1),
           sizeof(id->bits.node.byte5));

    nvmx_debug("uuid: %x-%x-%x-%x-%x%x",
               time_low, time_mid, time_hi_and_version, clk_seq,
               id->bits.node.byte5, id->bits.node.byte1);

    id->bits.time_low = htonl(time_low);
    id->bits.time_mid = htons(time_mid);
    id->bits.time_hi_and_version = htons(time_hi_and_version);
    id->bits.clk_seq = htons(clk_seq);
    id->bits.node.byte1 = htons(id->bits.node.byte1);
    id->bits.node.byte5 = htonl(id->bits.node.byte5);

exit:
    free(tmp);
    return rc;
}

static int nvme_json_get_global_id(nvme_config_t *config, const char *ns_path,
                                   const char *type, nvme_id_ns_global_t *nid)
{
    int rc;
    char buf[256];
    char param[64];

    snprintf(param, sizeof(param), "%s.global.%s", ns_path, type);
    rc = nvme_json_get_value(config, param, buf, sizeof(buf));
    if (rc) {
        nvmx_warn("%s: failed to retrieve %s value from the config file",
                  ns_path, type);
        return -EINVAL;
    }

    memset(nid, 0, sizeof(*nid));

    if (strncmp(type, "eui64", 5) == 0) {
        rc = nvme_eui64_parse(&nid->eui64, buf);
    } else if (strncmp(type, "nguid", 5) == 0) {
        rc = nvme_nguid_parse(&nid->nguid, buf);
    } else if (strncmp(type, "uuid", 4) == 0) {
        rc = nvme_uuid_parse(&nid->uuid, buf);
    } else {
        nvmx_warn("%s: wrong global ID type %s", ns_path, type);
        rc = -EINVAL;
    }

    return rc;
}

int nvme_json_get_global_id_list(nvme_config_t *config, const char *ns_path,
                                 nvme_id_ns_descriptor_t id_descs[])
{
    int rc, count = 0, i = 0;
    nvme_id_ns_descriptor_t *id_desc;

    id_desc = &id_descs[0];

    rc = nvme_json_get_global_id(config, ns_path, "eui64", &id_desc->nid);
    if (!rc) {
        ++count;
        id_desc->nidl = 8;
        id_desc->nidt = NVME_NS_IDENTIFIER_EUI64;
        id_desc = &id_descs[++i];
    }

    rc = nvme_json_get_global_id(config, ns_path, "nguid", &id_desc->nid);
    if (!rc) {
        ++count;
        id_desc->nidl = 16;
        id_desc->nidt = NVME_NS_IDENTIFIER_NGUID;
        id_desc = &id_descs[++i];
    }

    rc = nvme_json_get_global_id(config, ns_path, "uuid", &id_desc->nid);
    if (!rc) {
        ++count;
        id_desc->nidl = 16;
        id_desc->nidt = NVME_NS_IDENTIFIER_UUID;
    }

    return count;
}
