/*
 *   Copyright © 2021 NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 *     * Neither the name of Intel Corporation nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "mlnx_snap_pci_manager.h"
#include "queue.h"
#include "nvme_emu_log.h"


struct nsid_list_entry {
    TAILQ_ENTRY(nsid_list_entry) entry;
    uint32_t nsid;
};

#if HAVE_SNAP
struct mlnx_snap_emu_manager {
    struct snap_context *sctx;
    struct ibv_context *ibctx;
    struct ibv_device *ibdev;
    LIST_ENTRY(mlnx_snap_emu_manager) entry;
    TAILQ_HEAD(, nsid_list_entry) nsid_list;
};

LIST_HEAD(, mlnx_snap_emu_manager) mlnx_snap_emu_manager_list =
                              LIST_HEAD_INITIALIZER(mlnx_snap_emu_manager_list);


struct ibv_context *mlnx_snap_get_ibv_context(const char *rdma_dev)
{
    struct mlnx_snap_emu_manager *emu_manager;
    struct ibv_context *ibctx = NULL;

    LIST_FOREACH(emu_manager, &mlnx_snap_emu_manager_list, entry) {
        if (!strncmp(rdma_dev, ibv_get_device_name(emu_manager->ibdev), 16)) {
            ibctx = emu_manager->ibctx;
            break;
        }
    }

    return ibctx;
}

struct snap_context *mlnx_snap_get_snap_context(const char *rdma_dev)
{
    struct mlnx_snap_emu_manager *emu_manager;
    struct snap_context *sctx = NULL;

    LIST_FOREACH(emu_manager, &mlnx_snap_emu_manager_list, entry) {
        if (!strncmp(rdma_dev, ibv_get_device_name(emu_manager->ibdev), 16)) {
            sctx = emu_manager->sctx;
            break;
        }
    }

    return sctx;
}

struct snap_pci **mlnx_snap_get_snap_pci_list(enum snap_emulation_type type,
                                              const char *rdma_dev,
                                              int *spci_list_sz)
{
    struct snap_context *sctx;
    struct snap_pci **pf_list;
    const struct snap_pfs_ctx *pfs;

    sctx = mlnx_snap_get_snap_context(rdma_dev);
    if (!sctx) {
        nvmx_debug("Cannot find snap context on %s", rdma_dev);
        goto err;
    }

    switch (type) {
    case SNAP_NVME:
        pfs = &sctx->nvme_pfs;
        break;
    case SNAP_VIRTIO_BLK:
        pfs = &sctx->virtio_blk_pfs;
        break;
    case SNAP_VIRTIO_FS:
        pfs = &sctx->virtio_fs_pfs;
        break;
    default:
        pfs = NULL;
    }
    if (!pfs || !pfs->max_pfs) {
        nvmx_debug("No PFs of type %d", type);
        goto err;
    }

    pf_list = calloc(pfs->max_pfs, sizeof(*pf_list));
    if (!pf_list) {
        nvmx_error("Cannot allocate PF list");
        goto err;
    }
    *spci_list_sz = snap_get_pf_list(sctx, type, pf_list);

    return pf_list;

err:
    *spci_list_sz = 0;
    return NULL;
}

struct snap_pci *mlnx_snap_get_pf_snap_pci_by_vuid(enum snap_emulation_type type,
                                        const char *rdma_dev, const char *vuid)
{
    struct snap_pci **pf_list;
    struct snap_pci *pci_func = NULL;
    int pf_list_sz;
    int i;

    if (!vuid)
            return NULL;

    pf_list = mlnx_snap_get_snap_pci_list(type, rdma_dev, &pf_list_sz);
    if (!pf_list)
        return NULL;

    for (i = 0; i < pf_list_sz; i++) {
        if (!pf_list[i]->plugged)
            continue;

        if (!strncmp(pf_list[i]->vuid, vuid,
                     sizeof(pf_list[i]->vuid))) {
            pci_func = pf_list[i];
            break;
        }
    }

    if (!pci_func && vuid)
            nvmx_error("PF with VUID %s cannot be found on %s", vuid, rdma_dev);

    free(pf_list);
    return pci_func;
}

struct snap_pci *mlnx_snap_get_snap_pci(enum snap_emulation_type type,
                                        const char *rdma_dev,
                                        const char *pci_bdf,
                                        int pf_index, int vf_index)
{
    struct snap_pci **pf_list;
    struct snap_pci *pci_func = NULL;
    int pf_list_sz;
    int i, j;

    pf_list = mlnx_snap_get_snap_pci_list(type, rdma_dev, &pf_list_sz);
    if (!pf_list)
        return NULL;

    if (!pci_bdf && pf_index < 0) {
        nvmx_error("At least one of 'pci_bdf'/'pf_index' must be defined");
        free(pf_list);
        return NULL;
    } else if (pci_bdf && pf_index >= 0) {
        nvmx_error("At most one of 'pci_bdf'/'pf_index' must be defined");
        free(pf_list);
        return NULL;
    } else if (pci_bdf) {
        for (i = 0; i < pf_list_sz; i++) {
            if (!pf_list[i]->plugged)
                continue;

            if (!strncmp(pf_list[i]->pci_number, pci_bdf,
                         sizeof(pf_list[i]->pci_number))) {
                pci_func = pf_list[i];
                break;
            }

            /* check vfs */
            for (j = 0; j < pf_list[i]->num_vfs; j++) {
                if (!strncmp(pf_list[i]->vfs[j].pci_number, pci_bdf,
                            sizeof(pf_list[i]->vfs[j].pci_number))) {
                    pci_func = &pf_list[i]->vfs[j];
                    goto found;
                }
            }
        }
    } else {
        if (pf_index >= pf_list_sz) {
            nvmx_error("PF %d exceeds limit (%lu)", pf_index,
                    pf_list_sz);
        } else if (vf_index < 0) {
                pci_func = pf_list[pf_index];
        } else if (vf_index >= pf_list[pf_index]->num_vfs) {
            nvmx_error("VF %d exceeds limit (%lu)", vf_index,
                       pf_list[pf_index]->num_vfs);
        } else {
            pci_func = &pf_list[pf_index]->vfs[vf_index];
        }
    }
found:
    free(pf_list);

    if (!pci_func && pci_bdf)
            nvmx_error("%s cannot be found on %s", pci_bdf, rdma_dev);

    return pci_func;
}

void mlnx_snap_pci_manager_clear()
{
    struct mlnx_snap_emu_manager *emu_manager;

    while ((emu_manager = LIST_FIRST(&mlnx_snap_emu_manager_list)) != NULL) {
        LIST_REMOVE(emu_manager, entry);
        if (emu_manager->sctx)
            snap_close(emu_manager->sctx);
        else
            ibv_close_device(emu_manager->ibctx);
        free(emu_manager);
    }
}

int mlnx_snap_pci_manager_init()
{
    struct ibv_device **list;
    struct mlnx_snap_emu_manager *emu_manager;
    struct snap_context *sctx;
    struct ibv_context *ibctx;
    int i, dev_count;
    bool found_emu_managers = false;

    list = ibv_get_device_list(&dev_count);
    if (!list) {
        nvmx_error("failed to open IB device list");
        goto err;
    }

    /* Verify RDMA device exist in ibv list */
    for (i = 0; i < dev_count; i++) {
        emu_manager = calloc(1, sizeof(*emu_manager));
        if (!emu_manager)
            goto clear_emu_manager_list;

        sctx = snap_open(list[i]);
        if (sctx) {
            found_emu_managers = true;
            ibctx = sctx->context;
            snap_nvme_pci_functions_cleanup(sctx);
            snap_virtio_blk_pci_functions_cleanup(sctx);
        } else {
            ibctx = ibv_open_device(list[i]);
            if (!ibctx) {
                free(emu_manager);
                goto clear_emu_manager_list;
            }
        }

        emu_manager->sctx = sctx;
        emu_manager->ibctx = ibctx;
        emu_manager->ibdev = list[i];
        TAILQ_INIT(&emu_manager->nsid_list);
        LIST_INSERT_HEAD(&mlnx_snap_emu_manager_list, emu_manager, entry);
    }
    if (!found_emu_managers) {
        nvmx_error("No emulation managers detected");
        goto clear_emu_manager_list;
    }

    ibv_free_device_list(list);
    return 0;

clear_emu_manager_list:
    mlnx_snap_pci_manager_clear();
    ibv_free_device_list(list);
err:
    return -EINVAL;
}

static struct mlnx_snap_emu_manager *to_emu_manager(const char *rdma_dev)
{
    struct mlnx_snap_emu_manager *emu_manager, *found = NULL;

    LIST_FOREACH(emu_manager, &mlnx_snap_emu_manager_list, entry) {
        if (!strncmp(rdma_dev, ibv_get_device_name(emu_manager->ibdev), 16)) {
            found = emu_manager;
            break;
        }
    }

    return found;
}

uint32_t mlnx_snap_acquire_namespace_id(const char *rdma_dev)
{
    uint32_t nsid;
    struct nsid_list_entry *elem, *prev, *map;
    struct mlnx_snap_emu_manager *emu_manager;

    emu_manager = to_emu_manager(rdma_dev);
    if (!emu_manager) {
        nvmx_error("Failed to find emulation manager for %s", rdma_dev);
        goto err;
    }

    /* Find the minimal empty nsid "slot" */
    nsid = 1;
    if (!TAILQ_EMPTY(&emu_manager->nsid_list)) {
        TAILQ_FOREACH(elem, &emu_manager->nsid_list, entry) {
            if (elem->nsid == nsid) {
                nsid++;
                continue;
            } else
                break;
        }
    }

    if (nsid > emu_manager->sctx->nvme_caps.max_nvme_nsid) {
        nvmx_error("There are no valid NSIDs on emu manager");
        goto err;
    }

    map = calloc(1, sizeof(*map));
    if (!map) {
        nvmx_error("Failed to allocate nsid map object");
        goto err;
    }
    map->nsid = nsid;

    /* Insert new element at the right place to keep list ordered */
    prev = NULL;
    TAILQ_FOREACH(elem, &emu_manager->nsid_list, entry) {
        if (elem->nsid < nsid) {
            prev = elem;
            continue;
        } else if (elem->nsid == nsid) {
            nvmx_error("Requested NSID %lu is already mapped", nsid);
            goto free_map;
        } else
            break;
    }
    if (prev)
        TAILQ_INSERT_AFTER(&emu_manager->nsid_list, prev, map, entry);
    else
        TAILQ_INSERT_HEAD(&emu_manager->nsid_list, map, entry);

    nvmx_debug("NSID %lu is acquired", nsid);
    return nsid;

free_map:
    free(map);
err:
    return 0; //invliad nsid
}

void mlnx_snap_release_namespace_id(const char *rdma_dev, uint32_t nsid)
{
    struct nsid_list_entry *elem, *found = NULL;
    struct mlnx_snap_emu_manager *emu_manager;

    emu_manager = to_emu_manager(rdma_dev);
    if (!emu_manager) {
        nvmx_error("Failed to find emulation manager for %s", rdma_dev);
        return;
    }

    TAILQ_FOREACH(elem, &emu_manager->nsid_list, entry) {
        if (elem->nsid != nsid)
            continue;

        found = elem;
        break;
    }

    if (found) {
        nvmx_debug("NSID %lu is released", found->nsid);
        TAILQ_REMOVE(&emu_manager->nsid_list, found, entry);
        free(found);
    } else {
        nvmx_warn("Trying to umnap non-existing src nsid %lu", nsid);
    }

}
#else
int mlnx_snap_pci_manager_init()
{
    return 0;
}
void mlnx_snap_pci_manager_clear()
{
}

struct snap_pci **mlnx_snap_get_snap_pci_list(enum snap_emulation_type type,
                                              const char *rdma_dev,
                                              int *spci_list_sz)
{
    return NULL;
}

struct snap_context *mlnx_snap_get_snap_context(const char *rdma_dev)
{
    return NULL;
}
struct snap_pci *mlnx_snap_get_snap_pci(enum snap_emulation_type type,
                                        const char *rdma_dev,
                                        const char *pci_bdf,
                                        int pf_index, int vf_index)
{
    return NULL;
}

uint32_t mlnx_snap_acquire_namespace_id(const char *rdma_dev)
{
    return 0;
}

void mlnx_snap_release_namespace_id(const char *rdma_dev, uint32_t nsid)
{
}
#endif
