#include "config.h"
#include "nvme_emu_ep.h"
#include "nvme_emu_io_driver.h"
#include "nvme_posix_ctrl.h"
#include "nvme_mem_ctrl.h"
#include "nvmf_emu_ctrl.h"
#include "nvme_spdk_ctrl.h"

void nvme_emu_ep_destroy_backend(nvme_emu_ep_t *ep)
{
    switch (ep->ep_type) {
    case NVME_EMU_EP_TYPE_NVMF_RDMA:
        nvmf_emu_delete_ctrl(ep->ctrl);
        break;
    case NVME_EMU_EP_TYPE_POSIX:
        nvme_emu_delete_posix_ctrl(ep->ctrl);
        break;
    case NVME_EMU_EP_TYPE_MEM:
        nvme_emu_delete_mem_ctrl(ep->ctrl);
        break;
    case NVME_EMU_EP_TYPE_SPDK:
        nvme_spdk_delete_ctrl(ep->ctrl);
        break;
    default:
        nvmx_error("%d ep is not supported", ep->ep_type);
        break;
    }
}

int nvme_emu_ep_create_backend(nvme_emu_ep_t *ep)
{
    int rc = -1;

    switch (ep->ep_type) {
    case NVME_EMU_EP_TYPE_NVMF_RDMA:
        rc = nvmf_emu_create_ctrl(ep);
        break;
    case NVME_EMU_EP_TYPE_POSIX:
        rc = nvme_emu_create_posix_ctrl(ep);
        break;
    case NVME_EMU_EP_TYPE_MEM:
        rc = nvme_emu_create_mem_ctrl(ep);
        break;
    case NVME_EMU_EP_TYPE_SPDK:
        rc = nvme_spdk_create_ctrl(ep);
        break;
    default:
        nvmx_error("%d ep is not supported", ep->ep_type);
        break;
    }

    return rc;
}

/*
 * Must be called with ep_lock held
 */
void nvme_emu_ep_delete_sq(nvme_emu_ep_t *ep, uint16_t sqid)
{

    nvmx_assertv_always(ep->sqs[sqid].state == NVME_EMU_EP_QUEUE_START,
                        "destroying non-valid SQ");

    nvmx_debug("queue %d: destroying SQ for ctrl 0x%p", sqid, ep->ctrl);

    ep->sqs[sqid].depth = 0;
    ep->sqs[sqid].state = NVME_EMU_EP_QUEUE_NEW;
    ep->sqs[sqid].cqid = 0;
    ep->num_created_io_squeues--;

    ep->ep_ops.delete_sq(ep, sqid);
}


/*
 * Must be called with ep_lock held
 */
void nvme_emu_ep_stop_sqs(nvme_emu_ep_t *ep)
{
    nvme_emu_driver_t *driver = ep->io_driver;
    int i;

    for (i = 1; i < driver->num_queues; i++) {
        if (ep->sqs[i].state == NVME_EMU_EP_QUEUE_START)
            nvme_emu_ep_delete_sq(ep, i);
    }
}

/*
 * Must be called with ep_lock held
 */
void nvme_emu_ep_delete_cq(nvme_emu_ep_t *ep, uint16_t cqid)
{
    nvmx_assertv_always(ep->cqs[cqid].state == NVME_EMU_EP_QUEUE_START,
                        "destroying non-valid CQ");

    nvmx_debug("queue %d: destroying CQ for ctrl 0x%p", cqid, ep->ctrl);

    ep->cqs[cqid].depth = 0;
    ep->cqs[cqid].state = NVME_EMU_EP_QUEUE_NEW;
    ep->num_created_io_cqueues--;

    ep->ep_ops.delete_cq(ep, cqid);
}

/*
 * Must be called with ep_lock held
 */
void nvme_emu_ep_stop_cqs(nvme_emu_ep_t *ep)
{
    nvme_emu_driver_t *driver = ep->io_driver;
    int i;

    for (i = 1; i < driver->num_queues; i++) {
        if (ep->cqs[i].state == NVME_EMU_EP_QUEUE_START)
            nvme_emu_ep_delete_cq(ep, i);
    }
}

/*
 * Must be called with ep_lock held
 */
void nvme_emu_ep_stop_backend(nvme_emu_ep_t *ep, bool clear_namespaces)
{

    nvmx_assertv_always(ep->num_created_io_cqueues == 0,
                        "stopping ep with valid CQs");

    nvmx_assertv_always(ep->num_created_io_squeues == 0,
                        "stopping ep with valid SQs");

    switch (ep->ep_type) {
    case NVME_EMU_EP_TYPE_NVMF_RDMA:
        nvmf_emu_stop_ctrl(ep, clear_namespaces);
        break;
    case NVME_EMU_EP_TYPE_SPDK:
        nvme_spdk_stop_ctrl(ep);
        break;
    case NVME_EMU_EP_TYPE_POSIX:
        nvme_stop_posix_ctrl(ep);
        break;
    case NVME_EMU_EP_TYPE_MEM:
        nvme_stop_mem_ctrl(ep);
        break;
    default:
        nvmx_error("%d ep is not supported", ep->ep_type);
        break;
    }
}

/*
 * Must be called with ep_lock held
 */
int nvme_emu_ep_start_backend(nvme_emu_ep_t *ep)
{
    int rc = -1;

    switch (ep->ep_type) {
    case NVME_EMU_EP_TYPE_NVMF_RDMA:
        rc = nvmf_emu_start_ctrl(ep);
        break;
    case NVME_EMU_EP_TYPE_SPDK:
        rc = nvme_spdk_start_ctrl(ep);
        break;
    case NVME_EMU_EP_TYPE_POSIX:
        rc = nvme_start_posix_ctrl(ep);
        break;
    case NVME_EMU_EP_TYPE_MEM:
        rc = nvme_start_mem_ctrl(ep);
        break;
    default:
        nvmx_error("%d ep is not supported", ep->ep_type);
        break;
    }

    return rc;
}

/*
 * Must be called with ep_lock held
 */
void nvme_emu_ep_deactivate_cq(nvme_emu_ep_t *ep, uint16_t cqid)
{

    nvmx_assertv_always(ep->cqs[cqid].state == NVME_EMU_EP_QUEUE_ACTIVE,
                        "deactivating non active CQ");

    nvmx_debug("queue %d: deactivating CQ for ctrl 0x%p", cqid, ep->ctrl);

    ep->ep_ops.deactivate_cq(ep, cqid);
    ep->cqs[cqid].state = NVME_EMU_EP_QUEUE_START;
}

/*
 * Must be called with ep_lock held
 */
static int nvme_emu_ep_activate_cq(nvme_emu_ep_t *ep, uint16_t cqid)
{
    int ret;

    nvmx_assertv_always(ep->cqs[cqid].state == NVME_EMU_EP_QUEUE_START,
                        "activating non started CQ");

    nvmx_debug("queue %d: activating CQ for ctrl 0x%p", cqid, ep->ctrl);

    ret = ep->ep_ops.activate_cq(ep, cqid);
    if (!ret)
        ep->cqs[cqid].state = NVME_EMU_EP_QUEUE_ACTIVE;

    return ret;
}

/*
 * Must be called with ep_lock held
 */
void nvme_emu_ep_deactivate_sq(nvme_emu_ep_t *ep, uint16_t sqid)
{

    nvmx_assertv_always(ep->sqs[sqid].state == NVME_EMU_EP_QUEUE_ACTIVE,
                        "deactivating non active SQ");

    nvmx_debug("queue %d: deactivating SQ for ctrl 0x%p", sqid, ep->ctrl);

    ep->ep_ops.deactivate_sq(ep, sqid);
    ep->sqs[sqid].state = NVME_EMU_EP_QUEUE_START;
}

/*
 * Must be called with ep_lock held
 */
static int nvme_emu_ep_activate_sq(nvme_emu_ep_t *ep, uint16_t sqid)
{
    int ret;
    nvme_emu_ep_queue_t *sq = &ep->sqs[sqid];

    nvmx_assertv_always(sq->state == NVME_EMU_EP_QUEUE_START,
                        "activating non started SQ");

    if (ep->cqs[sq->cqid].state != NVME_EMU_EP_QUEUE_ACTIVE) {
        nvmx_error("queue %d: paired CQ %d is not activated", sqid, sq->cqid);
        return -1;
    }

    nvmx_debug("queue %d: activating SQ for ctrl 0x%p", sqid, ep->ctrl);

    ret = ep->ep_ops.activate_sq(ep, sqid);
    if (!ret)
        sq->state = NVME_EMU_EP_QUEUE_ACTIVE;

    return ret;
}

/*
 * Must be called with ep_lock held
 */
void nvme_emu_ep_deactivate_backend(nvme_emu_ep_t *ep)
{
    nvme_emu_driver_t *driver = ep->io_driver;
    int i;

    for (i = 1; i < driver->num_queues; i++) {
        if (ep->sqs[i].state == NVME_EMU_EP_QUEUE_ACTIVE)
            nvme_emu_ep_deactivate_sq(ep, i);
    }

    for (i = 1; i < driver->num_queues; i++) {
        if (ep->cqs[i].state == NVME_EMU_EP_QUEUE_ACTIVE)
            nvme_emu_ep_deactivate_cq(ep, i);
    }
}

/*
 * Must be called with ep_lock held
 */
int nvme_emu_ep_activate_backend(nvme_emu_ep_t *ep, int *activated_cqs,
                                 int *activated_sqs)
{
    nvme_emu_driver_t *driver = ep->io_driver;
    int i, ret = 0;

    if (activated_cqs)
        *activated_cqs = 0;
    if (activated_sqs)
        *activated_sqs = 0;

    for (i = 1; i < driver->num_queues; i++) {
        int curr_ret;

        if (ep->cqs[i].state == NVME_EMU_EP_QUEUE_START) {
            curr_ret = nvme_emu_ep_activate_cq(ep, i);
            if (!curr_ret && activated_cqs)
                (*activated_cqs)++;
            ret |= curr_ret;
        }
    }

    for (i = 1; i < driver->num_queues; i++) {
        int curr_ret;

        if (ep->sqs[i].state == NVME_EMU_EP_QUEUE_START) {
            curr_ret = nvme_emu_ep_activate_sq(ep, i);
            if (!curr_ret && activated_sqs)
                (*activated_sqs)++;
            ret |= curr_ret;
        }
    }

    return ret;
}

void nvme_emu_ep_fail_backend(nvme_emu_ep_t *ep)
{
    nvme_driver_fail_ep(ep->io_driver, ep);
}

int nvme_emu_ep_attach_volume(nvme_emu_ep_t *ep, uint32_t nsid,
                              uint64_t size_b, uint8_t block_order,
                              uint16_t backend_metadata,
                              nvme_id_ns_descriptor_t id_descs[], void *priv_ns)
{
    return nvme_driver_attach_ep_volume(ep->io_driver, nsid, size_b, block_order,
                                        backend_metadata, id_descs, priv_ns);
}

void nvme_emu_ep_detach_volume(nvme_emu_ep_t *ep, uint32_t nsid)
{
    nvme_driver_detach_ep_volume(ep->io_driver, nsid);
}

/*
 * Must be called with ep_lock held
 */
int nvme_emu_ep_create_sq(nvme_emu_ep_t *ep, uint16_t sqid, uint16_t depth,
                          uint16_t cqid)
{
    int ret;

    nvmx_assertv_always(ep->sqs[sqid].state == NVME_EMU_EP_QUEUE_NEW,
                        "creating already valid SQ");

    nvmx_debug("queue %d: creating SQ of size %u", sqid, depth);

    ret = ep->ep_ops.create_sq(ep, sqid, depth, cqid);
    if (ret)
        return ret;

    ep->sqs[sqid].depth = depth;
    ep->sqs[sqid].state = NVME_EMU_EP_QUEUE_START;
    ep->sqs[sqid].type = NVME_EMU_EP_SQ;
    ep->sqs[sqid].cqid = cqid;
    ep->num_created_io_squeues++;

    return 0;
}

/*
 * Must be called with ep_lock held
 */
int nvme_emu_ep_create_cq(nvme_emu_ep_t *ep, uint16_t cqid, uint16_t depth)
{
    int ret;

    nvmx_assertv_always(ep->cqs[cqid].state == NVME_EMU_EP_QUEUE_NEW,
                        "creating already valid CQ");

    nvmx_debug("queue %d: creating CQ of size %u", cqid, depth);

    ret = ep->ep_ops.create_cq(ep, cqid, depth);
    if (ret)
        return ret;

    ep->cqs[cqid].depth = depth;
    ep->cqs[cqid].state = NVME_EMU_EP_QUEUE_START;
    ep->cqs[cqid].type = NVME_EMU_EP_CQ;
    ep->num_created_io_cqueues++;

    return 0;
}

/*
 * Must be called with ep_lock held
 *
 * Try your best to create new ep CQs.
 * In case some operation fails, continue trying starting next CQ.
 * The destruction function will make sure we don't leak.
 *
 * returns the gap between driver's cq's and ep cq's.
 */
int nvme_emu_ep_start_cqs(nvme_emu_ep_t *ep)
{
    nvme_emu_driver_t *driver = ep->io_driver;
    int i;

    for (i = 1; i < driver->num_queues; i++) {
        if (driver->cqs[i].depth != 0 &&
            ep->cqs[i].state == NVME_EMU_EP_QUEUE_NEW)
            nvme_emu_ep_create_cq(ep, i, driver->cqs[i].depth);
    }

    return driver->num_created_io_cqueues - ep->num_created_io_cqueues;
}

/*
 * Must be called with ep_lock held
 *
 * Try your best to create new ep SQs.
 * In case some operation fails, continue trying starting next SQ.
 * The destruction function will make sure we don't leak.
 *
 * returns the gap between driver's sq's and ep sq's.
 */
int nvme_emu_ep_start_sqs(nvme_emu_ep_t *ep)
{
    nvme_emu_driver_t *driver = ep->io_driver;
    int i;

    for (i = 1; i < driver->num_queues; i++) {
        if (driver->sqs[i].depth != 0 &&
            ep->sqs[i].state == NVME_EMU_EP_QUEUE_NEW)
            nvme_emu_ep_create_sq(ep, i, driver->sqs[i].depth, driver->sqs[i].cqid);
    }

    return driver->num_created_io_squeues - ep->num_created_io_squeues;
}
