#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <semaphore.h>
#include <signal.h>
#include <string.h>
#include <stdarg.h>
#include <sys/utsname.h>
#include <unistd.h>
#include <stdbool.h>
#include <assert.h>
#include <infiniband/verbs.h>

#include "config.h"
#include "nvme_emu_log.h"
#include "nvme_emu.h"
#include "debug.h"

#define NVMX_LOGFILE_PATH "NVME_SNAP_LOGFILE_PATH"
#define NVMX_LOG_LEVEL "NVME_SNAP_LOG_LEVEL"
#define NVMX_DEFAULT_LOGFILE_PATH_FOLDER "/var/log/mlnx_snap/"
#define NVMX_DEFAULT_LOGFILE_NAME_PREFIX "mlnx_snap_log"
#define NVMX_DEFAULT_LOGFILE_SUFFIX ".log"
#define NVMX_LOGFILE_PATH_MAX_LEN 255

#define LOG_ERROR(_fmt, ...) \
    do { \
        fprintf(stderr, "%s:%d ERROR " _fmt "\n", __FILE__, __LINE__, ## __VA_ARGS__); \
        fflush(stderr); \
    } while(0)

static struct logger_context {
    pthread_mutex_t lock;
    sem_t sem_progress;
    pthread_t rotate_thread;
    FILE* fp;
    char filename[NVMX_LOGFILE_PATH_MAX_LEN+1];
    bool logged_to_file;
    enum NvmxEmuLogLevel log_level;
} logger_ctx;

static void print_fw_info()
{
    struct ibv_device **dev_list;
    struct ibv_context *ctx;
    struct ibv_device_attr_ex device_attr;
    struct ibv_port_attr port_attr;
    int port, num_devices;

    dev_list = ibv_get_device_list(&num_devices);
    if (!dev_list) {
        nvmx_warn("Failed to get IB devices list");
        goto cleanup;
    } else if (!num_devices) {
        nvmx_warn("Failed to get RDMA devices");
        goto cleanup;
    }

    ctx = ibv_open_device(dev_list[0]);
    if (!ctx) {
        nvmx_warn("Failed to open device");
        goto cleanup;
    }
    if (ibv_query_device_ex(ctx, NULL, &device_attr)) {
        nvmx_warn("Failed to query device props, error = %s",strerror(errno));
        goto close_ctx;
    }
    if (strlen(device_attr.orig_attr.fw_ver)) {
        nvmx_info("fw_ver:\t%s", device_attr.orig_attr.fw_ver);
    }
    else {
        nvmx_warn("fw_ver:failed to locate fw version");
    }
    for (port = 1; port <= device_attr.orig_attr.phys_port_cnt; ++port) {
        if (ibv_query_port(ctx, port, &port_attr)) {
            nvmx_warn("Failed to query port %u props", port);
            continue;
        }
        nvmx_debug("port:\t%d state:\t%s (%d)", port,
                   ibv_port_state_str(port_attr.state), port_attr.state);
    }

close_ctx:
    if (ibv_close_device(ctx)) {
        nvmx_warn("Failed to close device");
    }
cleanup:
    ibv_free_device_list(dev_list);
}

static inline void print_time(FILE *out)
{
    long mic_s;
    time_t ltime;
    char s_time[26];
    struct timespec spec;
    struct tm ltm;

    clock_gettime(CLOCK_REALTIME, &spec);
    time(&ltime);

    mic_s = spec.tv_nsec / 1.0e3; // Convert nanoseconds to microseconds
    strftime(s_time, 26, "%Y-%m-%d %H:%M:%S", localtime_r(&ltime, &ltm));
    fprintf(out, "[%s.%06ld] ", s_time, mic_s);
}

static int open_log_file_nolock();
static void close_log_file_nolock();
static int reload_log_file();
static int close_logger_flag = 0;

void nvmx_logger(const char *file_name, unsigned int line_num,
                enum NvmxEmuLogLevel level, const char *level_c,
                const char *format, ...)
{
    va_list args = {};
    int oldstate;

    /*
     * logger might be used before it was initialized - for example
     * when the logger uses json functions to get configuration
     */
    assert(logger_ctx.fp && "Tried to log without opening the logfile\n");

    if (level < logger_ctx.log_level)
        return;

    va_start(args, format);

    pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &oldstate);
    pthread_mutex_lock(&logger_ctx.lock);
    print_time(logger_ctx.fp);
    fprintf(logger_ctx.fp, "%s::%s:%d:: ", level_c, file_name, line_num);
    vfprintf(logger_ctx.fp, format, args);
    fprintf(logger_ctx.fp, "\n");
    fflush(logger_ctx.fp);
    pthread_mutex_unlock(&logger_ctx.lock);
    pthread_setcancelstate(oldstate, NULL);

    va_end(args);
}

void logger_handle_sig(int sig)
{
    sem_post(&logger_ctx.sem_progress);
}

static void *logrotate()
{
    while(1) {
        sem_wait(&logger_ctx.sem_progress);
        if (close_logger_flag == 1)
            pthread_exit(NULL);
        reload_log_file();
    }
}

static enum NvmxEmuLogLevel log_level_str_to_enum(char *log_level)
{
    if (strcmp(log_level, "DBG") == 0)
        return NVMX_LOG_VERBOSE_LEVEL_DBG;
    if (strcmp(log_level, "INFO") == 0)
        return NVMX_LOG_VERBOSE_LEVEL_INFO;
    if (strcmp(log_level, "WARN") == 0)
        return NVMX_LOG_VERBOSE_LEVEL_WARN;
    if (strcmp(log_level, "ERR") == 0)
        return NVMX_LOG_VERBOSE_LEVEL_ERR;

    if (strcmp(log_level, "MAX") == 0)
        return NVMX_LOG_VERBOSE_LEVEL_MAX;

    LOG_ERROR("ERROR:: Invalid logging verbosity requested: %s",
            log_level);

    return NVMX_LOG_VERBOSE_LEVEL_MAX;
}

/*
 * configures the logger context, calling func should
 * lock before calling
 */
static void _set_logfile_config()
{
    char *log_level;
    char *logfile_path_env;
    char s_time[26];
    time_t ltime;
    bool use_default_path = false;

    logfile_path_env = getenv(NVMX_LOGFILE_PATH);
    if (logfile_path_env) {
        if (strlen(logfile_path_env) > NVMX_LOGFILE_PATH_MAX_LEN) {
            LOG_ERROR("Logfile path = %s is longer than" \
                    "maximum length (%d). Using default path",
                    logfile_path_env, NVMX_LOGFILE_PATH_MAX_LEN);
            use_default_path = true;
        } else {
            snprintf(logger_ctx.filename, sizeof(logger_ctx.filename),
                        "%s", logfile_path_env);
        }
    }
    else {
       use_default_path = true;
    }

    if (use_default_path) {
        time(&ltime);
        strftime(s_time, 26, "%Y-%m-%d_%H-%M-%S", localtime(&ltime));
        snprintf(logger_ctx.filename, sizeof(logger_ctx.filename), "%s%s_%s_%d%s",
                 NVMX_DEFAULT_LOGFILE_PATH_FOLDER, s_time,
                 NVMX_DEFAULT_LOGFILE_NAME_PREFIX, getpid(),
                 NVMX_DEFAULT_LOGFILE_SUFFIX);
    }

    if (getenv(NVMX_LOG_LEVEL)) {
        log_level = getenv(NVMX_LOG_LEVEL);
    } else {
        log_level = NVMX_DEFAULT_LOG_LEVEL;
    }

    logger_ctx.log_level = log_level_str_to_enum(log_level);
}

static void print_header() {
    struct utsname os_info;

    nvmx_info("------------NVMe SNAP Application Log-------------");

    if (uname(&os_info) != 0) {
        nvmx_warn("failed to get os info");
    }
    else {
        nvmx_info("system name = %s", os_info.sysname);
        nvmx_info("node name   = %s", os_info.nodename);
        nvmx_info("release     = %s", os_info.release);
        nvmx_info("version     = %s", os_info.version);
        nvmx_info("machine     = %s", os_info.machine);
    }
    print_fw_info();
}

/*
 * Opens the log file defined by logger_ctx.filename
 * handle cases where it should be stdout/stderr
 *
 * calling function should take lock
 */
static int open_log_file_nolock()
{
    logger_ctx.logged_to_file = false;

    if (strcmp(logger_ctx.filename, "stderr") == 0) {
        logger_ctx.fp = stderr;
        return 0;
    }
    if (strcmp(logger_ctx.filename, "stdout") == 0) {
        logger_ctx.fp = stdout;
        return 0;
    }

    logger_ctx.fp = fopen(logger_ctx.filename, "a");
    if (!logger_ctx.fp) {
        LOG_ERROR("failed to open log file: %s, "
                "error: %s setting to stderr",
                logger_ctx.filename, strerror(errno));
        logger_ctx.fp = stderr;
        return 1;
    }
    logger_ctx.logged_to_file = true;
    return 0;
}

static int open_log_file()
{
    int rc;

    pthread_mutex_lock(&logger_ctx.lock);
    rc = open_log_file_nolock();
    pthread_mutex_unlock(&logger_ctx.lock);
    print_header();
    return rc;
}

static void close_log_file_nolock()
{
    if (logger_ctx.logged_to_file) {
        fflush(logger_ctx.fp);
        if (fclose(logger_ctx.fp))
            LOG_ERROR("failed to close log file");
    }
    logger_ctx.logged_to_file = false;
    return;
}

static void close_log_file()
{
    pthread_mutex_lock(&logger_ctx.lock);
    close_log_file_nolock();
    pthread_mutex_unlock(&logger_ctx.lock);
    return;
}

static int reload_log_file()
{
    int rc;

    pthread_mutex_lock(&logger_ctx.lock);
    close_log_file_nolock();
    rc = open_log_file_nolock();
    pthread_mutex_unlock(&logger_ctx.lock);
    print_header();
    return rc;
}

int nvme_init_logger()
{
    struct sigaction act;
    int ret;

    if (sem_init(&logger_ctx.sem_progress, 0, 0)) {
        LOG_ERROR("failed to init logger semaphore");
        return 1;
    }

    memset(&act, 0, sizeof(act));
    act.sa_handler = logger_handle_sig;
    sigaction(SIGHUP, &act, 0);

    _set_logfile_config();
    if (pthread_mutex_init(&logger_ctx.lock, PTHREAD_PROCESS_PRIVATE)) {
        LOG_ERROR("failed to init logger mutex");
        goto destroy_sem;
    }
    if (open_log_file()) {
        LOG_ERROR("failed to open log file");
        goto destroy_mutex;
    }
    if (pthread_create(&logger_ctx.rotate_thread, NULL, &logrotate, NULL)) {
        LOG_ERROR("failed to create logrotate thread");
        goto out;
    }
    ret = backtrace_init(NULL, true);
    if (ret < 0)
        nvmx_warn("failed to init backtracing, continue anyway...");

    return 0;

out:
    close_log_file();
destroy_mutex:
    pthread_mutex_destroy(&logger_ctx.lock);
destroy_sem:
    sem_destroy(&logger_ctx.sem_progress);
    return 1;
}

int nvme_close_logger() {
    int rc = 0;

    close_logger_flag = 1;
    sem_post(&logger_ctx.sem_progress);
    if (pthread_join(logger_ctx.rotate_thread, NULL)) {
        LOG_ERROR("failed to join rotate thread");
        rc = 1;
    }
    close_log_file();
    if (sem_destroy(&logger_ctx.sem_progress)) {
        LOG_ERROR("failed to destroy sem rotate log");
        rc = 1;
    }
    if (pthread_mutex_destroy(&logger_ctx.lock)) {
        LOG_ERROR("failed to destroy logger lock");
        rc = 1;
    }

    return rc;
}
