#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "config.h"
#include "nvme_emu.h"
#include "nvme_regs.h"

static void nvme_reg_cap_dump(uint64_t cap, char *dump);
static void nvme_reg_vs_dump(uint64_t vs, char *dump);
static void nvme_reg_cc_dump(uint64_t cc, char *dump);
static void nvme_reg_csts_dump(uint64_t csts, char *dump);
static void nvme_reg_aqa_dump(uint64_t aqa, char *dump);

nvme_register_t nvme_regs[] = {
    { SNAP_NVME_REG_CAP,    8, SNAP_NVME_REG_RO, "CAP", "Controller Capabilities", nvme_reg_cap_dump },
    { SNAP_NVME_REG_VS,     4, SNAP_NVME_REG_RO, "VS",  "Controller Version", nvme_reg_vs_dump },
    { SNAP_NVME_REG_INTMS,  4, SNAP_NVME_REG_RW1S, "INTMS", "Interrupt Mask Set" },
    { SNAP_NVME_REG_INTMC,  4, SNAP_NVME_REG_RW1C, "INTMC", "Interrupt Mask Clear" },
    { SNAP_NVME_REG_CC,     4, SNAP_NVME_REG_RW,   "CC",    "Controller Configuration", nvme_reg_cc_dump },
    { SNAP_NVME_REG_CSTS,   4, SNAP_NVME_REG_RO|SNAP_NVME_REG_RW1C, "CSTS", "Controller Status", nvme_reg_csts_dump },
    { SNAP_NVME_REG_NSSR,   4, SNAP_NVME_REG_RW,   "NSSR", "NVM Subsystem Reset" },

    /* Adming Queue */
    { SNAP_NVME_REG_AQA,    4, SNAP_NVME_REG_RW,   "AQA",  "Admin Queue Attributes", nvme_reg_aqa_dump },
    { SNAP_NVME_REG_ASQ,    8, SNAP_NVME_REG_RW,   "ASQ",  "Admin Submission Queue Base Address" },
    { SNAP_NVME_REG_ACQ,    8, SNAP_NVME_REG_RW,   "ACQ",  "Admin Completion Queue Base Address" },

    /* Opttional registers */
    { SNAP_NVME_REG_CMBLOC, 4, SNAP_NVME_REG_RO,   "CMBLOC", "Controller Memory Buffer Location" },
    { SNAP_NVME_REG_CMBSZ,  4, SNAP_NVME_REG_RO,   "CMBSZ",  "Controller Memory Buffer Size" },
    { SNAP_NVME_REG_BPINFO, 4, SNAP_NVME_REG_RO,   "BPINFO", "Boot Partition Information" },
    { SNAP_NVME_REG_BPRSEL, 4, SNAP_NVME_REG_RW,   "BPRSEL", "Boot Partition Select" },
    { SNAP_NVME_REG_BPMBL,  4, SNAP_NVME_REG_RW,   "BPMBL",  "Boot Partition Memory Buffer Location" },
    { SNAP_NVME_REG_LAST,   0, 0, 0, 0 }
};

uint64_t nvme_reg_get(nvme_register_t *r, void *bar)
{
    char *p;

    p = (char *)bar + r->reg_base;
    return r->reg_size == 4 ? *(uint32_t *)p : *(uint64_t *)p;
}

void nvme_reg_set(nvme_register_t *r, void *bar, uint64_t v)
{
    char *p;

    p = (char *)bar + r->reg_base;

    if (r->reg_size == 4) {
        *(uint32_t *)p = v;
    } else {
        *(uint64_t *)p = v;
    }
}

void nvme_reg_dump(nvme_register_t *r, void *bar, bool user_mode)
{
    uint64_t val;
    char dump_priv[SNAP_NVME_REG_MAX_DUMP_FUNC_LEN] = {0};

    val = nvme_reg_get(r, bar);

    if (r->reg_dump_func)
        r->reg_dump_func(val, dump_priv);

    if (user_mode)
        printf("%-6s [%s, 0x%02x..0x%02x]: 0x%llx {%s}\n",
               r->name, r->desc,
               r->reg_base, r->reg_base + r->reg_size,
               (unsigned long long)val,
               dump_priv);
    else
        nvmx_info("%-6s [%s, 0x%02x..0x%02x]: 0x%llx {%s}",
                  r->name, r->desc,
                  r->reg_base, r->reg_base + r->reg_size,
                  (unsigned long long)val,
                  dump_priv);
}

void nvme_bar_dump(void *bar, unsigned len)
{
   int i;

   for (i = 0; 
           nvme_regs[i].reg_base != SNAP_NVME_REG_LAST && nvme_regs[i].reg_base < len;
           i++) {
       nvme_reg_dump(&nvme_regs[i], bar, true);
   }
}

int nvme_reg_cap_mqes(uint64_t cap)
{
    return cap & 0xffff;
}

int nvme_reg_cap_cqr(uint64_t cap)
{
    return (cap >> 16) & 0x1;
}

int nvme_reg_cap_ams(uint64_t cap)
{
    return (cap >> 17) & 0x3;
}

int nvme_reg_cap_to(uint64_t cap)
{
    return (cap >> 24) & 0xff;
}

int nvme_reg_cap_dstrd(uint64_t cap)
{
    return (cap >> 32) & 0xf;
}

int nvme_reg_cap_nssrs(uint64_t cap)
{
    return (cap >> 33) & 0x1;
}

int nvme_reg_cap_css(uint64_t cap)
{
    return (cap >> 37) & 0xff;
}

int nvme_reg_cap_bps(uint64_t cap)
{
    return (cap >> 45) & 0x1;
}

int nvme_reg_cap_mpsmin(uint64_t cap)
{
    return (cap >> 48) & 0xf;
}

int nvme_reg_cap_mpsmax(uint64_t cap)
{
    return (cap >> 52) & 0xf;
}

static void nvme_reg_cap_dump(uint64_t cap, char *dump)
{
    snprintf(dump, SNAP_NVME_REG_MAX_DUMP_FUNC_LEN,
             "MQES:%d CQR:%d AMS:%d TO:%d DSTRD:%d NSSRS:%d CSS:%d BPS:%d MPSMIN:%d MPSMAX:%d",
             nvme_reg_cap_mqes(cap),
             nvme_reg_cap_cqr(cap),
             nvme_reg_cap_ams(cap),
             nvme_reg_cap_to(cap),
             nvme_reg_cap_dstrd(cap),
             nvme_reg_cap_nssrs(cap),
             nvme_reg_cap_css(cap),
             nvme_reg_cap_bps(cap),
             nvme_reg_cap_mpsmin(cap),
             nvme_reg_cap_mpsmax(cap));
}

int nvme_reg_vs_ter(uint32_t vs)
{
    return vs & 0xff;
}

int nvme_reg_vs_mnr(uint32_t vs)
{
    return (vs >> 8) & 0xff;
}

int nvme_reg_vs_mjr(uint32_t vs)
{
    return (vs >> 16) & 0xffff;
}

static void nvme_reg_vs_dump(uint64_t vs, char *dump)
{
    snprintf(dump, SNAP_NVME_REG_MAX_DUMP_FUNC_LEN,
             "%d.%d.%d",
             nvme_reg_vs_mjr(vs),
             nvme_reg_vs_mnr(vs),
             nvme_reg_vs_ter(vs));
}

/* controller configuration */
int nvme_reg_cc_en(uint32_t cc)
{
    return cc & 0x1;
}

int nvme_reg_cc_css(uint32_t cc)
{
    return (cc >> 4) & 0x7;
}

int nvme_reg_cc_mps(uint32_t cc)
{
    return (cc >> 7) & 0xf;
}

int nvme_reg_cc_ams(uint32_t cc)
{
    return (cc >> 11) & 0x7;
}

int nvme_reg_cc_shn(uint32_t cc)
{
    return (cc >> 14) & 0x3;
}

int nvme_reg_cc_iosqes(uint32_t cc)
{
    return (cc >> 16) & 0xf;
}

int nvme_reg_cc_iocqes(uint32_t cc)
{
    return (cc >> 20) & 0xf;
}

static void nvme_reg_cc_dump(uint64_t cc, char *dump)
{
    snprintf(dump, SNAP_NVME_REG_MAX_DUMP_FUNC_LEN,
             "EN:%d CSS:%d MPS:%d AMS:%d SHN:%d IOSQES:%d IOCQES:%d",
             nvme_reg_cc_en(cc),
             nvme_reg_cc_css(cc),
             nvme_reg_cc_mps(cc),
             nvme_reg_cc_ams(cc),
             nvme_reg_cc_shn(cc),
             nvme_reg_cc_iosqes(cc),
             nvme_reg_cc_iocqes(cc));
}

int nvme_reg_csts_rdy(uint32_t csts)
{
    return (csts >> CSTS_RDY_SHIFT) & CSTS_RDY_MASK;
}

int nvme_reg_csts_cfs(uint32_t csts)
{
    return (csts >> CSTS_CFS_SHIFT) & CSTS_CFS_MASK;
}

int nvme_reg_csts_shst(uint32_t csts)
{
    return (csts >> CSTS_SHST_SHIFT) & CSTS_SHST_MASK;
}

int nvme_reg_csts_nssro(uint32_t csts)
{
    return (csts >> CSTS_NSSRO_SHIFT) & CSTS_NSSRO_MASK;
}

int nvme_reg_csts_pp(uint32_t csts)
{
    return (csts >> CSTS_PP_SHIFT) & CSTS_PP_MASK;
}

static void nvme_reg_csts_dump(uint64_t csts, char *dump)
{
    snprintf(dump, SNAP_NVME_REG_MAX_DUMP_FUNC_LEN,
             "RDY:%d CFS:%d SHST:%d NSSRO:%d PP:%d",
             nvme_reg_csts_rdy(csts),
             nvme_reg_csts_cfs(csts),
             nvme_reg_csts_shst(csts),
             nvme_reg_csts_nssro(csts),
             nvme_reg_csts_pp(csts));
}

int nvme_reg_aqa_asqs(uint32_t aqa)
{
    return aqa & 0xfff;
}

int nvme_reg_aqa_acqs(uint32_t aqa)
{
    return (aqa >> 16) & 0xfff;
}

static void nvme_reg_aqa_dump(uint64_t aqa, char *dump)
{
    snprintf(dump, SNAP_NVME_REG_MAX_DUMP_FUNC_LEN,
             "ASQS:%d ACQS:%d",
             nvme_reg_aqa_asqs(aqa),
             nvme_reg_aqa_acqs(aqa));
}

int nvme_bar_init_modify(nvme_bar_write_func_t bar_writer,
                         nvme_bar_instance_t *bar, void *ucontext)
{
    nvme_register_t *csts_reg = &nvme_regs[SNAP_NVME_REG_CSTS_IDX];
    int err;

    nvmx_assertv_always(!memcmp(&bar->prev, &bar->curr, sizeof(nvme_bar_t)),
                        "bar was already modified");

    if (!nvme_reg_cc_en(bar->curr.cc)) {
        /*
         * Controller is not enabled, so we can safely unset the CFS bit
         * and all other CSTS bits.
         */
        bar->prev.csts = bar->curr.csts = 0;
        err = bar_writer(ucontext, &bar->curr.csts,
                         csts_reg->reg_base, csts_reg->reg_size);
        if (err) {
            nvmx_error("Failed to reset CSTS bar register");
            return NVMX_ERR;
        }
        return NVMX_OK;
    }

    if (!nvme_reg_csts_rdy(bar->curr.csts)) {
        /* Driver is already running and enabled controller. So pretend that
         * we just got EN=1
         */
        nvmx_debug("CC.EN=1 and controller is not ready, scheduling enable controller");
        bar->prev.cc &= ~0x1;
    } else {
        /* Controller is already up and running. But since we are (re)starting all
         * state will be lost. It means that driver is going to reset controller
         * when admin or io command timeouts. Raising CFS will help the driver
         * to do the right thing.
         */
        bar->curr.csts |= NVME_CSTS_FAILED;
        bar->prev.csts = bar->curr.csts;
        /* since raising CFS is optional, don't check error code */
        bar_writer(ucontext, &bar->curr.csts, csts_reg->reg_base, csts_reg->reg_size);
        nvmx_debug("CC.EN=1 and controller was running, raising CFS");
    }

    /* There was a shutdown request which did not complete. Make sure we complete it */
    if (nvme_reg_cc_shn(bar->curr.cc) && !nvme_reg_csts_shst(bar->curr.csts)) {
        nvmx_info("CC.SHN=1 and shutdown is not completed, scheduling shutdown");
        bar->prev.cc &= ~(0x3 << 14);
    }

    return NVMX_OK;
}

int nvme_bar_init(nvme_bar_read_func_t bar_reader, nvme_bar_instance_t *bar,
		  void *ucontext)
{
    int err;
    int i;
    nvme_register_t *reg;
    const int init_bar_regs[4] = {
        SNAP_NVME_REG_CAP_IDX,
        SNAP_NVME_REG_VS_IDX,
        SNAP_NVME_REG_CC_IDX,
        SNAP_NVME_REG_CSTS_IDX
    };
    const int init_bar_regs_sz = sizeof(init_bar_regs) /
                                 sizeof(init_bar_regs[0]);

    memset(bar, 0, sizeof(*bar));

    bar->ucontext = ucontext;
    for (i = 0; i < init_bar_regs_sz; i++) {
        reg = &nvme_regs[init_bar_regs[i]];
        err = bar_reader(ucontext, (char*)&bar->curr + reg->reg_base,
                         reg->reg_base, reg->reg_size);
        if (err) {
            nvmx_error("Failed to read initial value of %s", reg->name);
            return NVMX_ERR;
        }
        /* Dump initial state, it makes log analysis much easier */
        nvme_reg_dump(&nvme_regs[init_bar_regs[i]], &bar->curr, false);
    }

    nvmx_assertv_always(err == NVMX_OK, "bar_reader failed");
    nvmx_assertv_always(bar->curr.cap, "cap");
    nvmx_assertv_always(nvme_reg_cap_mqes(bar->curr.cap) > 0, "cap_mqes");
    nvmx_assertv_always(nvme_reg_cap_cqr(bar->curr.cap) == 1, "cap_cqr");
    nvmx_assertv_always(nvme_reg_cap_nssrs(bar->curr.cap) == 0, "cap_nssrs");
    nvmx_assertv_always(nvme_reg_cap_css(bar->curr.cap) == 1, "cap_css");
    nvmx_assertv_always(nvme_reg_cap_bps(bar->curr.cap) == 0, "cap_bps");

    nvmx_assertv_always(bar->curr.vs, "vs");
    nvmx_assertv_always(nvme_reg_vs_mjr(bar->curr.vs) == 1, "vs_mjr");

    memcpy(&bar->prev, &bar->curr, sizeof(nvme_bar_t));
    return NVMX_OK;
}

int nvme_bar_update(nvme_bar_instance_t *bar, nvme_bar_read_func_t bar_reader,
		    nvme_reg_mod_cb_func_t cb)
{
    uint64_t cur_val, prev_val;
    int i;

    /* we do not support cmb and boot */
    bar_reader(bar->ucontext, &bar->curr, 0, SNAP_NVME_REG_CMBLOC);
    /* do sanity check */
    /* TODO: write back fatal status */
    nvmx_assertv_always(bar->curr.cap == bar->prev.cap, "read-only cap");
    nvmx_assertv_always(bar->curr.vs == bar->prev.vs, "read-only vs");
    //do not check until we handle bar
    //nvmx_assertv_always(cur_bar.csts == prev_bar.csts, "read-only csts");

    /* nothing is written to regs that we do not support */
    nvmx_assertv_always(bar->curr.intms == 0, "intms not supported");
    nvmx_assertv_always(bar->curr.intmc == 0, "intmc not supported");
    nvmx_assertv_always(bar->curr.nssrc == 0, "nssrc is not supported");

    /* Controller Config changes should be handled last */ 
    for (i = SNAP_NVME_REG_ACQ_IDX; i >= SNAP_NVME_REG_CC_IDX; i--) {
        cur_val  = nvme_reg_get(&nvme_regs[i], &bar->curr);
        prev_val = nvme_reg_get(&nvme_regs[i], &bar->prev);
        if (cur_val != prev_val) {
            nvme_reg_dump(&nvme_regs[i], &bar->curr, false);
            cb(&bar->curr, &nvme_regs[i], cur_val, prev_val);
            nvme_reg_set(&nvme_regs[i], &bar->prev, cur_val);
        }
    }

    /* just memcopy 80 bytes */
    memcpy(&bar->prev, &bar->curr, sizeof(nvme_bar_t));
    return NVMX_OK;
}
