#include <stdint.h>

#include "config.h"
#include "nvme_emu.h"
#include "nvme_sgl.h"

enum SGL_STATUS {
    SGL_OK,
    SGL_ERR,
    SGL_EOF,
    SGL_LAST_DATA,
    SGL_DATA_IN_PROGRESS,
    SGL_LAST_DATA_IN_PROGRESS,
    SGL_FRAGMENT_IN_PROGRESS,
    SGL_WRITEV_IN_PROGRESS,
};

/*
 * read/write from/to sgl
 */
static inline void nvme_emu_dma_rw_sgl(nvme_queue_ops_t *ops, void *buf,
                                       const nvme_sgl_desc_t *sgl, size_t len,
                                       int dir)
{
    uint32_t rkey = 0;
    uint64_t addr = 0;

    switch (sgl->type) {
    case NVME_SGL_TYPE_KEYED_DATA_BLOCK:
        rkey = le32_to_cpu(sgl->keyed_data_block.key);
        addr = le64_to_cpu(sgl->keyed_data_block.address);
        break;
    case NVME_SGL_TYPE_DATA_BLOCK:
        addr = le64_to_cpu(sgl->data_block.address);
        break;
    default:
        nvmx_assertv_always(0, "unsupported type %d\n", sgl->type);
    }

    switch (dir) {
        case SGL_DIR_WRITE:
            nvme_emu_dma_write(ops, buf, addr, len, rkey);
            return;
        case SGL_DIR_READ:
            nvme_emu_dma_read(ops, buf, addr, len, rkey);
            return;
        default:
            nvmx_fatal("invalid direction %d", dir);
    }
}

void nvme_sgl_rw(nvme_queue_ops_t *ops, const nvme_sgl_desc_t *sgl, void *buf,
                 size_t len, int dir)
{
    nvmx_debug("got an sgl request to 0x%x\n", dir);

    nvmx_assertv_always(sgl->type == NVME_SGL_TYPE_KEYED_DATA_BLOCK,
                        "unsupported type %d\n", sgl->type);

    nvmx_assertv_always(sgl->subtype == NVME_SGL_SUBTYPE_ADDRESS,
                        "unsupported type %d\n", sgl->type);

    nvmx_assertv_always(len <= 8192, "len %d is too big\n", (int)len);

    nvme_emu_dma_rw_sgl(ops, buf, sgl, len, dir);
}

static void nvme_sgl_desc_dump(const nvme_sgl_desc_t *desc)
{
    switch (desc->type) {
    case NVME_SGL_TYPE_KEYED_DATA_BLOCK:
        nvmx_debug("SGL_KEYED_DATA_BLOCK: addr %p len %u key %x subtype %u",
                le64_to_cpu(desc->keyed_data_block.address),
                le32_to_cpu(desc->keyed_data_block.length),
                le32_to_cpu(desc->keyed_data_block.key),
                desc->subtype);
        break;
    case NVME_SGL_TYPE_DATA_BLOCK:
        nvmx_debug("SGL_DATA_BLOCK: addr %p len %u subtype %u",
                le64_to_cpu(desc->data_block.address),
                le32_to_cpu(desc->data_block.length),
                desc->subtype);
        break;
    case NVME_SGL_TYPE_SEGMENT:
        nvmx_debug("SGL_SEGMENT: addr %p len %u subtype %u",
                le64_to_cpu(desc->segment.address),
                le32_to_cpu(desc->segment.length),
                desc->subtype);
        break;
    case NVME_SGL_TYPE_LAST_SEGMENT:
        nvmx_debug("SGL_LAST_SEGMENT: addr %p len %u subtype %u",
                le64_to_cpu(desc->last_segment.address),
                le32_to_cpu(desc->last_segment.length),
                desc->subtype);
        break;
    case NVME_SGL_TYPE_BIT_BUCKET:
        nvmx_debug("SGL_BIT_BUCKET: len %u subtype %u",
                le32_to_cpu(desc->bit_bucket.length),
                desc->subtype);
        break;
    default:
        nvmx_debug("UNSUPPORTED DESCRIPTOR: type %u subtype %u",
                desc->type,
                desc->subtype);
    }
}

static enum SGL_STATUS sgl_segment_init(nvme_async_req_t *req, const nvme_sgl_desc_t *desc)
{
    uint32_t length;
    uint64_t host_addr;
    sgl_segment *segment = &req->sgl.segment;

    length = le32_to_cpu(desc->segment.length);
    host_addr = le64_to_cpu(desc->segment.address);

    if (unlikely(length == 0 || (length & 0x0FU))) {
        req->err_code = NVME_SC_SGL_INVALID_SEGMENT_DESCRIPTOR;
        return SGL_ERR;
    }

    if (unlikely(host_addr > 0xFFFFFFFFFFFFFFFFULL - length)) {
        req->err_code = NVME_SC_SGL_INVALID_DATA_LENGTH;
        return SGL_ERR;
    }

    memset(segment, 0, sizeof(sgl_segment));

    segment->host_addr = host_addr;
    segment->nr_total = length / 16U;
    segment->last_segment = desc->type == NVME_SGL_TYPE_LAST_SEGMENT;

    return SGL_OK;
}

static enum SGL_STATUS sgl_process_data_block(nvme_async_req_t *req, const nvme_sgl_desc_t *desc, bool last_descriptor)
{
    uint32_t length;
    uint64_t host_addr;
    uint32_t bytes_to_process;
    enum SGL_STATUS rc;
    enum nvme_emu_status status;

    length = le32_to_cpu(desc->data_block.length);
    host_addr = le64_to_cpu(desc->data_block.address);

    if (unlikely(host_addr > 0xFFFFFFFFFFFFFFFFULL - length)) {
        req->err_code = NVME_SC_SGL_INVALID_DATA_LENGTH;
        return SGL_ERR;
    }

    if (unlikely(length == 0)) {
        return SGL_OK;
    }

    req->sgl.operation_context.data_length += length;

    rc = SGL_OK;

    if (req->sgl.operation_context.iov) {
        if (unlikely(req->iov_cnt == req->iov_max)) {
            req->err_code = NVME_SC_INTERNAL_DEV_ERROR;
            return SGL_ERR;
        }

        req->iov[req->iov_cnt].iov_len = length;
        req->iov[req->iov_cnt].iov_base = (void *)host_addr;

        nvmx_debug("REQ %p, iov[%u] = { %p, %u }", req, req->iov_cnt, (void *)host_addr, length);

        req->iov_cnt++;
        rc = (req->sgl.operation_context.data_length >= req->len) ? SGL_LAST_DATA : SGL_OK;
    }

    if (req->sgl.operation_context.data) {
        bytes_to_process = nvmx_min(length, req->len - req->sgl.operation_context.bytes_processed);

        req->dma_cmd.op = req->sgl.operation_context.write ? NVME_REQ_DMA_FROM_HOST : NVME_REQ_DMA_TO_HOST;
        req->dma_cmd.raddr = host_addr;
        req->dma_cmd.laddr = req->io_buf + req->sgl.operation_context.bytes_processed;
        req->dma_cmd.len = bytes_to_process;
        req->sgl.operation_context.bytes_processed += bytes_to_process;
        req->sgl.operation_context.fragment = 0;

        if (unlikely(req->sgl.operation_context.bytes_processed > req->data_len)) {
            req->err_code = NVME_SC_SGL_INVALID_DATA_LENGTH;
            return SGL_ERR;
        }

        if (unlikely(req->sgl.operation_context.bytes_processed == req->len || last_descriptor)) {
            req->sgl.operation_context.eof = 1;
            rc = SGL_LAST_DATA_IN_PROGRESS;
        }
        else
            rc = SGL_DATA_IN_PROGRESS;

        if(req->is_tcp_rx_zcopy && (req->sgl.operation_context.eof == 0)) {
            nvmx_debug("sgl_process_data_block - Exit on non-last SGE");
            return SGL_WRITEV_IN_PROGRESS;
        }
        else {
            nvmx_debug("sgl_process_data_block: REQ %p, Start DMA operation %p", req, (void *)host_addr);
            status = nvme_emu_request_submit(req);
        }

        if (unlikely(status != NVMX_INPROGRESS)) {
            req->err_code = NVME_SC_DATA_TRAS_ERROR;
            return SGL_ERR;
        }
    }

    return rc;
}

static enum SGL_STATUS sgl_load_next_fragment(nvme_async_req_t *req)
{
    enum nvme_emu_status rc;

    const size_t max_fragment_len = req->aux_len / sizeof(nvme_sgl_desc_t);
    const size_t nr_left = req->sgl.segment.nr_total - req->sgl.segment.nr_processed;
    const size_t fragment_len = max_fragment_len < nr_left ? max_fragment_len : nr_left;
    const size_t bytes_load = fragment_len * sizeof(nvme_sgl_desc_t);
    const size_t offset = req->sgl.segment.nr_processed * sizeof(nvme_sgl_desc_t);

    if (unlikely(fragment_len == 0)) {
        req->err_code = NVME_SC_INTERNAL_DEV_ERROR;
        return SGL_ERR;
    }

    req->sgl.segment.fragment_len = fragment_len;

    req->dma_cmd.op = NVME_REQ_DMA_FROM_HOST;
    req->dma_cmd.len = bytes_load;
    req->dma_cmd.raddr = req->sgl.segment.host_addr + offset;
    req->dma_cmd.laddr = req->aux_buf;
    req->sgl.operation_context.fragment = 1;

    nvmx_debug("REQ %p, DMA read from host %p", req, (void *)(req->sgl.segment.host_addr + offset));

    rc = nvme_emu_request_submit(req);

    if (unlikely(rc != NVMX_INPROGRESS)) {
        req->err_code = NVME_SC_DATA_TRAS_ERROR;
        return SGL_ERR;
    }

    return SGL_FRAGMENT_IN_PROGRESS;
}

static enum SGL_STATUS sgl_process_desriptor(nvme_async_req_t *req, const nvme_sgl_desc_t *desc,
                                    bool last_descriptor, bool last_segment)
{
    enum SGL_STATUS rc;

    nvme_sgl_desc_dump(desc);

    switch (desc->type) {
    case NVME_SGL_TYPE_DATA_BLOCK:
        rc = sgl_process_data_block(req, desc, last_descriptor);
        break;
    case NVME_SGL_TYPE_SEGMENT:
    case NVME_SGL_TYPE_LAST_SEGMENT:

        if (unlikely(!last_descriptor)) {
            req->err_code = NVME_SC_SGL_INVALID_DESCRIPTORS_NUMBER;
            rc = SGL_ERR;
        }
        else if (unlikely(last_segment)) {
            req->err_code = NVME_SC_SGL_INVALID_SEGMENT_DESCRIPTOR;
            rc = SGL_ERR;
        }
        else {
            rc = sgl_segment_init(req, desc);
            if (likely(rc == SGL_OK))
                rc = sgl_load_next_fragment(req);
        }
        break;
    default:
        req->err_code = NVME_SC_SGL_INVALID_DESCRIPTOR_TYPE;
        rc = SGL_ERR;
        break;
    }

    return rc;
}

static enum SGL_STATUS sgl_process_fragment(nvme_async_req_t *req)
{
    size_t i;
    enum SGL_STATUS rc;
    sgl_segment *segment = &req->sgl.segment;

    segment->fragment = (nvme_sgl_desc_t *)req->aux_buf;

    for (i = 0; i < segment->fragment_len; i++) {

        bool last_segment = segment->last_segment;
        bool last_descriptor = segment->nr_processed == segment->nr_total - 1;
        segment->nr_processed++;
        rc = sgl_process_desriptor(req, segment->fragment++, last_descriptor, last_segment);
        switch (rc) {
        case SGL_FRAGMENT_IN_PROGRESS:
        case SGL_LAST_DATA_IN_PROGRESS:
            return SGL_OK;
        case SGL_LAST_DATA:
            return SGL_EOF;
        case SGL_ERR:
            return SGL_ERR;
        case SGL_OK:
        case SGL_EOF:
        case SGL_DATA_IN_PROGRESS:
            break;
        case SGL_WRITEV_IN_PROGRESS:
            sgl_load_next_fragment(req);
            continue; 
        }
    }

    if (req->sgl.segment.nr_processed != req->sgl.segment.nr_total)
        rc = sgl_load_next_fragment(req);
    else
        rc = SGL_EOF;

    return rc;
}

static inline void sgl_complete(nvme_async_req_t *req)
{
    if (unlikely(req->err_code == NVME_SC_SUCCESS && req->sgl.operation_context.data_length < req->len))
        req->err_code = NVME_SC_SGL_INVALID_DATA_LENGTH;
    
    req->comp_cb(req->comp_cb_user, req->err_code == NVME_SC_SUCCESS ? NVMX_OK : NVMX_ERR);
}

static void sgl_data_ready(void *user, enum nvme_emu_status status)
{
    enum SGL_STATUS rc;
    nvme_async_req_t *req = user;

    nvmx_debug("REQ %p, data ready", req);

    if (unlikely(req->err_code != NVME_SC_SUCCESS)) {
        req->comp_cb(req->comp_cb_user, NVMX_ERR);
        return;
    }

    if (unlikely(status != NVMX_OK)) {
        req->err_code = NVME_SC_DATA_TRAS_ERROR;
        sgl_complete(req);
        return;
    }

    if (req->sgl.operation_context.fragment) {
        rc = sgl_process_fragment(req);
        if (rc == SGL_ERR || rc == SGL_EOF)
            sgl_complete(req);
    }
    else {
        if (req->sgl.operation_context.eof)
            sgl_complete(req);
    }
}

static void sgl_operation(nvme_async_req_t *req, bool iov, bool data)
{
    enum SGL_STATUS rc;

    memset(&req->sgl.operation_context, 0, sizeof(req->sgl.operation_context));
    req->iov_cnt = 0;
    req->err_code = NVME_SC_SUCCESS;
    req->sgl.operation_context.iov = iov;
    req->sgl.operation_context.data = data;
    if (req->dma_cmd.op == NVME_REQ_DMA_FROM_HOST)
        req->sgl.operation_context.write = 1;

    req->dma_cmd.comp_cb = sgl_data_ready;
    req->dma_cmd.comp_cb_user = req;
    req->dma_cmd.rkey = 0;

    rc = sgl_process_desriptor(req, &req->sgl.sgl, true, false);

    if (rc != SGL_DATA_IN_PROGRESS &&
        rc != SGL_LAST_DATA_IN_PROGRESS &&
        rc != SGL_FRAGMENT_IN_PROGRESS)
        sgl_complete(req);
}

int nvme_sgl_rw_nb(nvme_async_req_t *req)
{
    sgl_operation(req, false, true);
    return NVMX_INPROGRESS;
}

void nvme_sgl_rw_iov_nb(nvme_async_req_t *req)
{
    sgl_operation(req, true, false);
}

void nvme_sgl_rw_iov_data_nb(nvme_async_req_t *req)
{
    sgl_operation(req, true, true);
}
