#define __GNU_SOURCE
#include <stdio.h>
#include <stdint.h>
#include <fcntl.h>
#include <unistd.h>
#include <dirent.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <spdk/rpc.h>
#include <spdk/env.h>
#include <spdk/json.h>
#include <spdk/jsonrpc.h>
#include <spdk/bdev_module.h>
#include <infiniband/verbs.h>
#include <linux/virtio_blk.h>
#include <linux/virtio_config.h>
#include <snap.h>
#include <snap_virtio_blk_ctrl.h>

#include "config.h"
#ifdef HAVE_SPDK_UTIL_H
#include <spdk/util.h>
#endif
#include "spdk/event.h"

#include "mlnx_snap_pci_manager.h"
#include "spdk_emu_mgr.h"
#include "spdk_io_mgr.h"
#include "nvme_spdk_subsystem.h"
#include "nvme_controller.h"
#include "virtio_blk_controller.h"
#include "virtio_blk_regs.h"
#include "virtio_fs_controller.h"
#include "virtio_fs_regs.h"
#include "memp.h"

#include <snap_env.h>

#define MEM_POOL_SIZE "MEM_POOL_SIZE"
#define DEFAULT_MEM_POOL_SIZE 0
SNAP_ENV_REG_ENV_VARIABLE(MEM_POOL_SIZE, DEFAULT_MEM_POOL_SIZE);

#define STORAGE_ADMIN_SCRIPT_PATH_ENV "MLNX_SNAP_STORAGE_ADMIN_PATH"
#define NETWORK_ADMIN_SCRIPT_PATH_ENV "MLNX_SNAP_NETWORK_ADMIN_PATH"
#define VIRTIO_BLK_EMU_MAX_NUM_MSIX 31
#define VIRTIO_BLK_CTRL_DEFAULT_SUSPEND_TMO 10
#define NVME_REGS_DEFAULT_QUEUE_SIZE 64
#ifndef VIRTIO_F_SR_IOV
#define VIRTIO_F_SR_IOV 37
#endif

#define SNAP_PCI_CLASS_CODE_NVME_CTRL 0x010802
#define SNAP_PCI_CLASS_CODE_MASS_STORAGE_CTRL 0x018000
#define SNAP_PCI_VENDOR_ID_MELLANOX 0x15B3
#define SNAP_PCI_VENDOR_ID_REDHAT 0x1AF4
#define SNAP_PCI_REVISION_ID_VIRTIO_BLK 0x1
#define SNAP_PCI_DEVICE_ID_VIRTIO_BLK 0x1042
#define SNAP_PCI_DEVICE_ID_BLUEFIELD_NVME 0x6001
#define SNAP_PCI_DEVICE_ID_BLUEFIELD_VIRTIO_BLK 0x6002
#define SNAP_PCI_NUM_MSIX_DEFAULT 2
#define SNAP_PCI_DEVICE_ID_VIRTIO_BLK_TRANSITIONAL 0x1001
#define SNAP_PCI_REVISION_ID_VIRTIO_BLK_TRANSITIONAL 0x0

#define SNAP_VIRTIO_BLK_VIRTIO_DEVICE_ID 2

#define INTENSIVE_POLLING_TIMEOUT 1
#define HOTUNPLUG_TIMEOUT 10
/*
 * virtio-v1.1-cs01
 * Device ID  Virtio Device
 * 26         file system device
 * 0x105A (0x1040 + 0x1A)
 */
#define SNAP_PCI_DEVICE_ID_VIRTIO_FS 0x105A
#define SNAP_PCI_SSID_VIRTIO_FS 0x6004
// 15b3  Mellanox Technologies
#define SNAP_PCI_SSVID_VIRTIO_FS SNAP_PCI_VENDOR_ID_MELLANOX

static memp_t *memp = NULL;

static int spdk_emu_ctx_io_threads_create(struct spdk_emu_ctx *ctrl_ctx);


static void spdk_emu_ctrl_destroy(struct spdk_emu_ctx *ctx,
                                  void (*done_cb)(void *arg),
                                  void *done_cb_arg);
static int spdk_emu_progress(void *arg);

void *spdk_get_snap_blacklist_method(nvme_ctrl_t *ctrl);
int spdk_handle_snap_blacklist_method(void *ctx, nvme_ctrl_t *ctrl);

static void spdk_emu_vf_change_handler(struct spdk_emu_ctx *sctx);

pthread_mutex_t spdk_emu_list_lock = PTHREAD_MUTEX_INITIALIZER;

struct spdk_emu_list_head spdk_emu_list =
                              LIST_HEAD_INITIALIZER(spdk_emu_list);
bool spdk_emu_rpc_finished = false;

int spdk_emu_init()
{
    long long size = snap_env_getenv(MEM_POOL_SIZE);

    if (!size)
        return 0;

    if (size == -EINVAL)
        return -EINVAL;

    void *buffer = spdk_dma_malloc(size, 4096, NULL);

    if (!buffer)
        return -ENOMEM;

    // Application thread adds one
    memp = memp_init(buffer, size, spdk_io_mgr_get_num_threads() + 1);
    if (!memp)
        goto free_buffer;

    return 0;

free_buffer:
    free(buffer);
    return -ENOMEM;
}

void spdk_emu_clear()
{
    if (memp) {
        void *buffer = memp_get_buffer(memp);
        memp_destroy(memp);
        if (buffer)
            spdk_dma_free(buffer);
    }
}

struct spdk_emu_controller_vblk_create_attr {
    char *emu_manager;
    int   pf_id;
    int   vf_id;
    char *pci_bdf;
    char *bdev_type;
    char *bdev;
    uint32_t num_queues;
    uint32_t queue_depth;
    uint32_t size_max;
    uint32_t seg_max;
    char *serial;
    bool force_in_order;
    bool suspended;
    bool recover;
    char *mem;
    bool force_recover;
    char *vuid;
    bool admin_q;
};

struct spdk_emu_controller_vblk_delete_attr {
    char *name;
    char *vuid;
    bool force;
};

struct spdk_emu_controller_vfs_create_attr {
    char *emu_manager;
    int   pf_id;
    int   vf_id;
    char *pci_bdf;
    char *dev_type;
    uint32_t num_queues;
    uint32_t queue_depth;
    char *tag;
    bool force_in_order;
    bool suspended;
    bool recover;
};

struct spdk_emu_controller_vfs_delete_attr {
    char *name;
    bool force;
};

struct spdk_emu_controller_vblk_get_iostat_attr {
    char *name;
};

struct spdk_emu_ctx *
spdk_emu_ctx_find_by_pci_id(const char *emu_manager,
                            int pf_id, int vf_id,
                            enum spdk_emu_protocol protocol)
{
    enum snap_pci_type  type;
    struct spdk_emu_ctx *ctx;

    if (protocol == SPDK_EMU_PROTOCOL_NVME)
        type = vf_id < 0 ? SNAP_NVME_PF : SNAP_NVME_VF;
    else if (protocol == SPDK_EMU_PROTOCOL_VIRTIO_BLK)
        type = vf_id < 0 ? SNAP_VIRTIO_BLK_PF : SNAP_VIRTIO_BLK_VF;
    else if (protocol == SPDK_EMU_PROTOCOL_VIRTIO_FS)
        type = vf_id < 0 ? SNAP_VIRTIO_FS_PF : SNAP_VIRTIO_FS_VF;
    else
        return NULL;

    LIST_FOREACH(ctx, &spdk_emu_list, entry) {
        if (strncmp(ctx->emu_manager, emu_manager,
                    SPDK_EMU_MANAGER_NAME_MAXLEN) ||
            ctx->spci->type != type)
            continue;

        if (vf_id < 0) { /*use pf_id case, only need to match pf_id */
            if (ctx->spci->id == pf_id)
                return ctx;
        } else { /*use vf_id case, need to match pf_id and vf_id both */
            if (ctx->spci->parent->id == pf_id && ctx->spci->id == vf_id)
                return ctx;
        }
    }

    return NULL;
}

struct spdk_emu_ctx *
spdk_emu_ctx_find_by_pci_bdf(const char *emu_manager, const char *pci_bdf)
{
    struct spdk_emu_ctx *ctx;

    LIST_FOREACH(ctx, &spdk_emu_list, entry) {
        if (strncmp(ctx->emu_manager, emu_manager,
                    SPDK_EMU_MANAGER_NAME_MAXLEN) == 0 &&
            strncmp(ctx->spci->pci_number, pci_bdf,
                    sizeof(ctx->spci->pci_number)) == 0)
            return ctx;
    }
    return NULL;
}

struct spdk_emu_ctx *spdk_emu_ctx_find_by_emu_name(const char *emu_name)
{
    struct spdk_emu_ctx *ctx;

    LIST_FOREACH(ctx, &spdk_emu_list, entry) {
        if (!strncmp(ctx->emu_name, emu_name, SPDK_EMU_NAME_MAXLEN))
            return ctx;
    }
    return NULL;
}

struct spdk_emu_ctx *
spdk_emu_ctx_find_by_nqn_cntlid(const char *subnqn, int cntlid)
{
    int id;
    struct spdk_emu_ctx *ctx;
    struct nvme_subsystem *subsys;

    subsys = subsystem_nvme_find(subnqn);
    if (!subsys) {
        nvmx_error("subnqn: %s is not exist", subnqn);
        return NULL;
    }

    LIST_FOREACH(ctx, &spdk_emu_list, entry) {
        id = snap_config_get_cntlid(ctx->sconfig);
        if (ctx->nvme_subsys == subsys && id == cntlid)
            return ctx;
    }

    return NULL;
}

struct spdk_emu_ctx *
spdk_emu_ctx_find_by_vuid(const char *vuid)
{
    struct spdk_emu_ctx *ctx;

    LIST_FOREACH(ctx, &spdk_emu_list, entry) {
           if (strncmp(ctx->spci->vuid, vuid, 128) == 0)
            return ctx;
    }

    return NULL;
}

/* For backward compatible consideration, we support two approach
 * to find controller: by controller name or by nqn and controller
 * id pair. this function can be removed when only support to find
 * controller by nqn and controller id pair. */
struct spdk_emu_ctx *
spdk_emu_ctx_find(const char *subnqn, int cntlid, const char *name)
{
    if (cntlid != -1)
        return spdk_emu_ctx_find_by_nqn_cntlid(subnqn, cntlid);
    else
        return spdk_emu_ctx_find_by_emu_name(name);
}

void spdk_emu_ctx_fini_cb_rpc(void *arg)
{
    struct spdk_jsonrpc_request *request = arg;
    struct spdk_json_write_ctx *w;

    if (spdk_emu_rpc_finished) {
        nvmx_warn("spdk rpc is closed. Not returning response for rpc request");
    } else {
        /* Return RPC completion */
        w = spdk_jsonrpc_begin_result(request);
        spdk_json_write_bool(w, true);
        spdk_jsonrpc_end_result(request, w);
    }
}

static inline int nvme_ctrl_progress_admin_wrap(void *arg)
{
    nvme_ctrl_t *ctrl = arg;

    return nvme_ctrl_progress_admin(ctrl);
}

static inline void nvme_ctrl_progress_mmio_wrap(void *arg)
{
    nvme_ctrl_t *ctrl = arg;

    nvme_ctrl_progress_mmio(ctrl);
}

static inline int nvme_ctrl_progress_all_io_wrap(void *arg)
{
    nvme_ctrl_t *ctrl = arg;

    return nvme_ctrl_progress_all_io(ctrl);
}

static inline int nvme_ctrl_progress_io_thread_wrap(void *arg, int thread_id)
{
    nvme_ctrl_t *ctrl = arg;

    return nvme_ctrl_progress_io(ctrl, thread_id);
}

static void nvme_ctrl_suspend_wrap(void *arg)
{
    nvme_ctrl_t *ctrl = arg;

    nvme_ctrl_suspend(ctrl);
}

static bool nvme_ctrl_is_suspended_wrap(void *arg)
{
    nvme_ctrl_t *ctrl = arg;

    return nvme_ctrl_is_suspended(ctrl);
}

static int nvme_ctrl_resume_wrap(void *arg)
{
    return -1;
}

struct spdk_emu_ctx_ctrl_ops {
    const char *prefix;
    int (*progress)(void *ctrl);
    void (*progress_mmio)(void *ctrl);
    int (*progress_io)(void *ctrl);
    int (*progress_io_thread)(void *ctrl, int thread_id);
    /* stop accepting new requests and complete all outstaning
     * requests. The function is async */
    void (*suspend)(void *ctrl);
    /* true if controller has completed suspend */
    bool (*is_suspended)(void *ctrl);
    /* reverse of suspend */
    int (*resume)(void *ctrl);
};

static const struct spdk_emu_ctx_ctrl_ops spdk_emu_ctx_ctrl_ops_nvme = {
    .prefix = NVME_EMU_NAME_PREFIX,

    .progress = nvme_ctrl_progress_admin_wrap,
    .progress_mmio = nvme_ctrl_progress_mmio_wrap,
    .progress_io = nvme_ctrl_progress_all_io_wrap,
    .progress_io_thread = nvme_ctrl_progress_io_thread_wrap,
    .suspend = nvme_ctrl_suspend_wrap,
    .is_suspended = nvme_ctrl_is_suspended_wrap,
    .resume = nvme_ctrl_resume_wrap
};

static const struct spdk_emu_ctx_ctrl_ops spdk_emu_ctx_ctrl_ops_virtio_blk = {
    .prefix = VBLK_EMU_NAME_PREFIX,

    .progress = NULL,
    .progress_mmio = virtio_blk_ctrl_progress,
    .progress_io = virtio_blk_ctrl_progress_all_io,
    .progress_io_thread = virtio_blk_ctrl_progress_io,
    .suspend = virtio_blk_ctrl_suspend,
    .is_suspended = virtio_blk_ctrl_is_suspended,
    .resume = virtio_blk_ctrl_resume
};

static const struct spdk_emu_ctx_ctrl_ops spdk_emu_ctx_ctrl_ops_virtio_fs = {
    .prefix = "VfsEmu",

    .progress = NULL,
    .progress_mmio = virtio_fs_ctrl_progress,
    .progress_io = virtio_fs_ctrl_progress_all_io,
    .progress_io_thread = virtio_fs_ctrl_progress_io,
    .suspend = virtio_fs_ctrl_suspend,
    .is_suspended = virtio_fs_ctrl_is_suspended,
    .resume = virtio_fs_ctrl_resume
};

static void spdk_emu_ctx_destroy_all_vf_ctxs(struct spdk_emu_ctx *pf_ctx,
                                             spdk_emu_mgr_fini_cb_t fini_cb,
                                             void *fini_cb_arg)
{
    struct spdk_emu_ctx *ctx;
    struct snap_pci *pf = pf_ctx->spci;
    uint16_t num_vfs = pf_ctx->num_vfs;
    int i;

    pthread_mutex_lock(&spdk_emu_list_lock);
    /* Iterate ove all VFs, and close any open controllers */
    for (i = 0; i < num_vfs; i++) {
        ctx = spdk_emu_ctx_find_by_pci_bdf(pf_ctx->emu_manager,
                                            pf->vfs[i].pci_number);
        if (!ctx) {
            /* In case VF controller is not opened, run callback directly */
            fini_cb(fini_cb_arg);
            continue;
        }

        nvmx_info("Deleting VF controller %s", ctx->emu_name);
        if (ctx->protocol == SPDK_EMU_PROTOCOL_NVME)
            nvme_subsys_ctrl_del(ctx);
        LIST_REMOVE(ctx, entry);
        spdk_emu_ctx_destroy(ctx, fini_cb, fini_cb_arg);
    }
    pthread_mutex_unlock(&spdk_emu_list_lock);
}

static void spdk_emu_vf_change_event_end(void *arg)
{
    struct spdk_emu_ctx *sctx = arg;

    sctx->num_vfs--;

    /* Only rescan after all VF controllers were closed properly */
    if (sctx->num_vfs == 0) {
        sctx->vfs_deletion_in_progress = false;
        spdk_emu_vf_change_handler(sctx);
    }
};

static void spdk_emu_vf_change_event_begin(void *ctx, int pf_id, int new_numvfs)
{
    struct spdk_emu_ctx *sctx = ctx;

    nvmx_info("Num of VFs changed to %d on pf %s", new_numvfs,
              sctx->spci->pci_number);

    sctx->num_vfs_next = new_numvfs;
    spdk_emu_vf_change_handler(sctx);
}


static void spdk_emu_vf_change_handler(struct spdk_emu_ctx *sctx)
{
    /*
     * Case 1: We are already in deletion process.
     * only cache the latest numvfs value for later,
     * and this change will be handled once deletion process
     * is completed.
     */
    if (sctx->vfs_deletion_in_progress)
        return;

    /*
     * Case 2: No change in numvfs is actually detected.
     * We cannot distinguish today (polling mode) if value rapid change
     * pattern was X-->0-->X, so we just ignore it.
     * TODO: handle such case when libsnap event mechanism is available.
     */
    if (sctx->num_vfs != 0 && sctx->num_vfs == sctx->num_vfs_next)
        return;

    /*
     * Case 3: A change in numvfs is detected, active VFs present.
     * According to SR-IOV spec, need to first destroy all previous VFs
     * before assigning new resources.
     * We first handle the deletion process, and when it is finished we
     * may call handler again for new VFs rescan (Case 4).
     */
    if (sctx->num_vfs != 0) {
        sctx->vfs_deletion_in_progress = true;
        nvmx_debug("Destroying old %d VFs controllers", sctx->num_vfs);
        spdk_emu_ctx_destroy_all_vf_ctxs(sctx, spdk_emu_vf_change_event_end,
                                         sctx);
        return;
    }

    /*
     * Case 4: A change in numvfs is detected, NO active VFs present.
     * We might reach here either directly, or right after Case 3 is completed.
     */
    nvmx_assertv_always(sctx->num_vfs == 0, "non-zero numvfs when rescanning");
    nvmx_info("Rescanning VFs for PF %s", sctx->spci->pci_number);
    if (snap_rescan_vfs(sctx->spci, sctx->num_vfs_next))
        nvmx_warn("Failed to rescan VFs");
    sctx->num_vfs = sctx->num_vfs_next;
}

void spdk_emu_device_detach_timeout(void *arg)
{
    struct spdk_emu_ctx *ctx = arg;
    struct spdk_emu_device_hotunplug_ctx *hotunplug_ctx = ctx->hotunplug_ctx;
    struct snap_pci *spci = hotunplug_ctx->spci;
    struct spdk_jsonrpc_request *request = hotunplug_ctx->request;

    nvmx_error("Timed out detaching PCI function %s. Note: Function is still in the process of detaching.", spci->pci_number);
    if (request)
        spdk_jsonrpc_send_error_response(request,
                         SPDK_JSONRPC_ERROR_INTERNAL_ERROR,
                         "Failed to detach PCI function");

    /* If we timed out on detach but received POWER_OFF state later,
     * make sure we dont try responding on this request again.
     */
    hotunplug_ctx->request = NULL;
    hotunplug_ctx->timeout_once = true;
}

void spdk_emu_device_detach_end(void *arg)
{
    int ret;
    struct spdk_json_write_ctx *w;
    struct spdk_emu_device_hotunplug_ctx *hotunplug_ctx = arg;
    struct snap_pci *spci = hotunplug_ctx->spci;
    struct spdk_jsonrpc_request *request = hotunplug_ctx->request;

    ret = snap_hotunplug_pf(spci);
    if (ret) {
        SPDK_ERRLOG("Failed to detach PCI function %s - %s\n",
                    spci->pci_number, strerror(ret));
        goto err;
    }

    SPDK_NOTICELOG("PCI function %s is detached\n", spci->pci_number);

    if (request) {
        w = spdk_jsonrpc_begin_result(request);
        spdk_json_write_bool(w, true);
        spdk_jsonrpc_end_result(request, w);
    }

    free(hotunplug_ctx);
    return;

err:
    if (request) {
        spdk_jsonrpc_send_error_response(request,
                         SPDK_JSONRPC_ERROR_INTERNAL_ERROR,
                         "Failed to detach PCI function");
    }
    free(hotunplug_ctx);
}

void spdk_emu_ctrl_destroy_device_detach_cb(void *arg)
{
    struct spdk_emu_ctx *ctx = arg;
    struct spdk_emu_device_hotunplug_ctx *hotunplug_ctx;
    struct snap_pci *spci;

    // A controller is still open delete it
    if (ctx) {
        hotunplug_ctx = ctx->hotunplug_ctx;
        spci = hotunplug_ctx->spci;
        pthread_mutex_lock(&spdk_emu_list_lock);
        LIST_REMOVE(ctx, entry);
        if (spci->type == SNAP_NVME_PF)
           nvme_subsys_ctrl_del(ctx);
        pthread_mutex_unlock(&spdk_emu_list_lock);
        spdk_emu_ctx_destroy(ctx, spdk_emu_device_detach_end, hotunplug_ctx);
    }
}

static void spdk_emu_hotunplug_device_begin(void *arg)
{
    spdk_thread_send_msg(spdk_get_thread(), spdk_emu_ctrl_destroy_device_detach_cb,
                         arg);
}

static void spdk_emu_hotunplug_timeout(void *arg)
{
    spdk_thread_send_msg(spdk_get_thread(), spdk_emu_device_detach_timeout,
                         arg);
}

static int spdk_emu_ctrl_create(struct spdk_emu_ctx *ctx,
                                const struct spdk_emu_ctx_create_attr *attr)
{
    nvme_ctrl_init_param_t nvme_init_attr = {};
    struct virtio_blk_ctrl_init_attr vblk_init_attr = {};
    struct virtio_fs_ctrl_init_attr vfs_init_attr = {};
    struct spdk_emu_controller_nvme_create_attr *nvme_attr;
    struct spdk_emu_controller_vblk_create_attr *vblk_attr;
    struct spdk_emu_controller_vfs_create_attr *vfs_attr;

    switch (ctx->protocol) {
    case SPDK_EMU_PROTOCOL_NVME:
        nvme_attr = attr->priv;
        ctx->sconfig = nvme_attr->sconfig;
        nvme_init_attr.sconfig = ctx->sconfig;
        nvme_init_attr.nthreads = spdk_env_get_core_count();
        nvme_init_attr.vf_change_ctx = ctx;
        nvme_init_attr.vf_change_cb = spdk_emu_vf_change_event_begin;
        nvme_init_attr.rpc_method_is_supported_cb = spdk_get_snap_blacklist_method;
        nvme_init_attr.rpc_method_handle_cb = spdk_handle_snap_blacklist_method;
        nvme_init_attr.memp = memp;
        nvme_init_attr.hotunplug_device_cb = spdk_emu_hotunplug_device_begin;
        nvme_init_attr.hotunplug_device_ctx = ctx;
        nvme_init_attr.hotunplug_timeout_cb = spdk_emu_hotunplug_timeout;
        ctx->ctrl = nvme_ctrl_init(&nvme_init_attr);
        if (!ctx->ctrl)
            goto err;

        ctx->ctrl_ops = &spdk_emu_ctx_ctrl_ops_nvme;
        break;
    case SPDK_EMU_PROTOCOL_VIRTIO_BLK:
        vblk_attr = attr->priv;
        vblk_init_attr.emu_manager_name = attr->emu_manager;
        if (attr->spci->type == SNAP_VIRTIO_BLK_VF) {
            vblk_init_attr.pf_id = attr->spci->parent->id;
            vblk_init_attr.vf_id = attr->spci->id;
        } else {
            vblk_init_attr.pf_id = attr->spci->id;
            vblk_init_attr.vf_id = -1;
        }
        vblk_init_attr.bdev_type = vblk_attr->bdev_type;
        vblk_init_attr.bdev_name = vblk_attr->bdev;
        vblk_init_attr.nthreads = spdk_env_get_core_count();
        vblk_init_attr.num_queues = vblk_attr->num_queues;
        /* Virtio blk needs at least 3 descriptors: header, data and status.
         * VIRTIO_F_RING_PACKED is not set so queue_size must be a power of 2
         * from Virtio spec 4.1.4.3.2
         */
        if ((vblk_attr->queue_depth > 0 && vblk_attr->queue_depth < 3) ||
                vblk_attr->queue_depth  & (vblk_attr->queue_depth -1)) {
            SPDK_ERRLOG("Virtio blk queue depth must be a power of 2 and at least 3\n");
            goto err;
        }
        vblk_init_attr.queue_depth = vblk_attr->queue_depth;
        vblk_init_attr.size_max = vblk_attr->size_max;
        vblk_init_attr.seg_max = vblk_attr->seg_max;
        vblk_init_attr.serial = vblk_attr->serial;
        vblk_init_attr.vf_change_cb = spdk_emu_vf_change_event_begin;
        vblk_init_attr.vf_change_cb_arg = ctx;
        vblk_init_attr.force_in_order = vblk_attr->force_in_order;
        vblk_init_attr.suspended = vblk_attr->suspended;
        vblk_init_attr.recover = vblk_attr->recover;
        vblk_init_attr.force_recover = vblk_attr->force_recover;
        vblk_init_attr.use_mem_pool = vblk_attr->mem &&
                                        !strcmp(vblk_attr->mem, "pool");
        vblk_init_attr.memp = memp;
        vblk_init_attr.admin_q = vblk_attr->admin_q;
        vblk_init_attr.hotunplug_device_cb = spdk_emu_hotunplug_device_begin;
        vblk_init_attr.hotunplug_device_ctx = ctx;
        vblk_init_attr.hotunplug_timeout_cb = spdk_emu_hotunplug_timeout;
        ctx->ctrl = virtio_blk_ctrl_init(&vblk_init_attr);
        if (!ctx->ctrl)
            goto err;

        ctx->ctrl_ops = &spdk_emu_ctx_ctrl_ops_virtio_blk;
        break;
    case SPDK_EMU_PROTOCOL_VIRTIO_FS:
        vfs_attr = attr->priv;
        vfs_init_attr.emu_manager_name = attr->emu_manager;
        if (attr->spci->type == SNAP_VIRTIO_FS_VF) {
            vfs_init_attr.pf_id = attr->spci->parent->id;
            vfs_init_attr.vf_id = attr->spci->id;
        } else {
            vfs_init_attr.pf_id = attr->spci->id;
            vfs_init_attr.vf_id = -1;
        }
        vfs_init_attr.dev_type = vfs_attr->dev_type;
        vfs_init_attr.nthreads = spdk_env_get_core_count();
        vfs_init_attr.num_queues = vfs_attr->num_queues;
        vfs_init_attr.queue_depth = vfs_attr->queue_depth;
        vfs_init_attr.tag = vfs_attr->tag;
        vfs_init_attr.vf_change_cb = spdk_emu_vf_change_event_begin;
        vfs_init_attr.vf_change_cb_arg = ctx;
        vfs_init_attr.force_in_order = vfs_attr->force_in_order;
        vfs_init_attr.suspended = vfs_attr->suspended;
        vfs_init_attr.recover = vfs_attr->recover;
        ctx->ctrl = virtio_fs_ctrl_init(&vfs_init_attr);
        if (!ctx->ctrl)
            goto err;

        ctx->ctrl_ops = &spdk_emu_ctx_ctrl_ops_virtio_fs;
        break;
    default:
        SPDK_ERRLOG("Unknown controller type %d\n", ctx->protocol);
        goto err;
    }
    return 0;

err:
    SPDK_ERRLOG("failed to initialize controller of type %d\n",
                ctx->protocol);
    return -1;
}

static void spdk_emu_ctx_stop_pollers(struct spdk_emu_ctx *ctx);

static inline int spdk_emu_progress_intensive(void *arg)
{
    struct spdk_emu_ctx *ctx = arg;
    int n;

    n = ctx->ctrl_ops->progress(ctx->ctrl);
#ifdef HAVE_SPDK_POLLER_PAUSE
    if (n > 0)
        ctx->intensive_polling_end = spdk_get_ticks() + INTENSIVE_POLLING_TIMEOUT * spdk_get_ticks_hz();
    else if (ctx->intensive_polling && ctx->intensive_polling_end <= spdk_get_ticks()) {
        ctx->intensive_polling = false;
        spdk_poller_pause(ctx->admin_intensive_poller);
    }
#endif

    return n;
}

static inline int spdk_emu_progress(void *arg)
{
    struct spdk_emu_ctx *ctx = arg;
    int n = 0;

    if (ctx->ctrl_ops->progress) {
        n = ctx->ctrl_ops->progress(ctx->ctrl);
#ifdef HAVE_SPDK_POLLER_PAUSE
        if (n > 0) {
            ctx->intensive_polling_end = spdk_get_ticks() + INTENSIVE_POLLING_TIMEOUT * spdk_get_ticks_hz();
            if (!ctx->intensive_polling) {
                ctx->intensive_polling = true;
                spdk_poller_resume(ctx->admin_intensive_poller);
            }
        }
#endif
    }

    ctx->ctrl_ops->progress_mmio(ctx->ctrl);

    /* suspend must be initiated by us */
    if (ctx->should_stop && ctx->ctrl_ops->is_suspended(ctx->ctrl)) {
        ctx->should_stop = false;
        spdk_emu_ctx_stop_pollers(ctx);
    }

    return n;
}

static inline int spdk_emu_progress_io(void *arg)
{
    struct spdk_emu_ctx *ctx = arg;

    return ctx->ctrl_ops->progress_io(ctx->ctrl);
}

static inline int spdk_emu_progress_io_thread(void *arg)
{
    struct spdk_emu_io_thread *thread = arg;
    struct spdk_emu_ctx *ctrl_ctx = thread->ctrl_ctx;

    return ctrl_ctx->ctrl_ops->progress_io_thread(ctrl_ctx->ctrl, thread->id);
}

static bool spdk_emu_ctrl_has_mt(struct spdk_emu_ctx *ctx)
{
    return ctx->ctrl_ops->progress_io_thread &&
           spdk_env_get_core_count() > 1;
}

static void spdk_emu_ctx_destroy_end(void *arg)
{
    struct spdk_emu_ctx *ctx = arg;

    SPDK_NOTICELOG("Controller %s was destroyed\n", ctx->emu_name);

    /*
     * Before we finish, we must call fini_cb, since our caller may
     * expect to finish some operation once deletion is done.
     * Examples for such operations:
     *   1. send RPC completion.
     *   2. Call spdk_app_stop().
     */
    if (ctx->fini_cb)
        ctx->fini_cb(ctx->fini_cb_arg);
    free(ctx);
}

static void spdk_emu_ctx_destroy_mt_end(void *arg)
{
    struct spdk_emu_io_thread *thread = arg;
    struct spdk_emu_ctx *ctx = thread->ctrl_ctx;

    ctx->num_io_threads--;

    /* Only after all threads are done, it is safe to free resources */
    if (!ctx->num_io_threads) {
        free(ctx->io_threads);
        spdk_poller_unregister(&ctx->admin_poller);
#ifdef HAVE_SPDK_POLLER_PAUSE
        spdk_poller_unregister(&ctx->admin_intensive_poller);
#endif
        /*
         * After IO pollers are all quiesced, it is now safe to
         * begin controller destruction
         */
        spdk_emu_ctrl_destroy(ctx, spdk_emu_ctx_destroy_end, ctx);
    }
}

static void spdk_emu_thread_unregister_poller(void *arg)
{
    struct spdk_emu_io_thread *thread = arg;

    spdk_poller_unregister(&thread->spdk_poller);
    spdk_thread_send_msg(thread->spdk_thread_creator,
                         spdk_emu_ctx_destroy_mt_end, thread);
}

static void spdk_emu_thread_register_poller(void *arg)
{
    struct spdk_emu_io_thread *thread = arg;

    thread->spdk_poller = spdk_poller_register(spdk_emu_progress_io_thread,
                                               thread, 0);
    if (!thread->spdk_poller)
        SPDK_ERRLOG("failed to register SPDK poller\n");
}

static void spdk_emu_ctx_destroy_mt_begin(struct spdk_emu_ctx *ctx)
{
    int i;
    struct spdk_emu_io_thread *thread;

    for (i = 0; i < ctx->num_io_threads; i++) {
        thread = &ctx->io_threads[i];

        spdk_thread_send_msg(thread->spdk_thread,
                             spdk_emu_thread_unregister_poller, thread);
    }
}

static void spdk_emu_ctrl_destroy_st_end(void *arg)
{
    struct spdk_emu_ctx *ctx = arg;

    spdk_thread_send_msg(spdk_get_thread(), spdk_emu_ctx_destroy_end, ctx);
}

static void spdk_emu_ctx_stop_pollers(struct spdk_emu_ctx *ctx)
{
    if (spdk_emu_ctrl_has_mt(ctx)) {
        spdk_emu_ctx_destroy_mt_begin(ctx);
    } else {
        spdk_poller_unregister(&ctx->io_poller);
        spdk_poller_unregister(&ctx->admin_poller);
#ifdef HAVE_SPDK_POLLER_PAUSE
        spdk_poller_unregister(&ctx->admin_intensive_poller);
#endif
        spdk_emu_ctrl_destroy(ctx, spdk_emu_ctrl_destroy_st_end, ctx);
    }
}

void spdk_emu_ctx_destroy(struct spdk_emu_ctx *ctx,
                          spdk_emu_mgr_fini_cb_t fini_cb, void *fini_cb_arg)
{
    ctx->fini_cb = fini_cb;
    ctx->fini_cb_arg = fini_cb_arg;

    /* Before stopping all io threads give controller a chance to
     * finish all outstanding io requests. There is no need to use send_msg
     * because admin poller is running on the ctx_ctreate/destroy thread
    */
    ctx->should_stop = true;
    ctx->ctrl_ops->suspend(ctx->ctrl);

    if (ctx->ctrl_ops->is_suspended(ctx->ctrl)) {
        ctx->should_stop = false;
        spdk_emu_ctx_stop_pollers(ctx);
    }
}

struct spdk_emu_ctx *
spdk_emu_ctx_create(const struct spdk_emu_ctx_create_attr *attr)
{
    struct spdk_emu_ctx *ctx;
    int err;

    ctx = calloc(1, sizeof(*ctx));
    if (!ctx)
        goto err;

    ctx->protocol = attr->protocol;
    err = spdk_emu_ctrl_create(ctx, attr);
    if (err)
        goto free_ctx;

    ctx->admin_poller = spdk_poller_register(spdk_emu_progress,
                                             ctx, 100000);
    if (!ctx->admin_poller) {
        SPDK_ERRLOG("failed to start controller\n");
        goto ctrl_destroy;
    }

#ifdef HAVE_SPDK_POLLER_PAUSE
    ctx->admin_intensive_poller = spdk_poller_register(spdk_emu_progress_intensive,
                                             ctx, 0);
    if (!ctx->admin_intensive_poller) {
        SPDK_ERRLOG("failed to start controller\n");
        goto unregister_adm_poller;
    }
    spdk_poller_pause(ctx->admin_intensive_poller);
#endif

    if (spdk_emu_ctrl_has_mt(ctx)) {
        err = spdk_emu_ctx_io_threads_create(ctx);
        if (err) {
            SPDK_ERRLOG("failed to start IO threads\n");
            goto unregister_adm_int_poller;
        }
    } else {
        ctx->io_poller = spdk_poller_register(spdk_emu_progress_io,
                                              ctx, 0);
        if (!ctx->io_poller) {
            SPDK_ERRLOG("failed to start general IO poller\n");
            goto unregister_adm_int_poller;
        }
    }

    ctx->spci = attr->spci;
    strncpy(ctx->emu_manager, attr->emu_manager, SPDK_EMU_MANAGER_NAME_MAXLEN - 1);
    if (attr->spci->parent)
        snprintf(ctx->emu_name, SPDK_EMU_NAME_MAXLEN,
                "%s%dpf%dvf%d", ctx->ctrl_ops->prefix,
                rdma_dev_name_to_id(attr->emu_manager), attr->spci->parent->id,
                attr->spci->id);
    else
        snprintf(ctx->emu_name, SPDK_EMU_NAME_MAXLEN,
                "%s%dpf%d", ctx->ctrl_ops->prefix,
                rdma_dev_name_to_id(attr->emu_manager), attr->spci->id);

    return ctx;

unregister_adm_int_poller:
#ifdef HAVE_SPDK_POLLER_PAUSE
    spdk_poller_unregister(&ctx->admin_intensive_poller);
unregister_adm_poller:
#endif
    spdk_poller_unregister(&ctx->admin_poller);
ctrl_destroy:
    spdk_emu_ctrl_destroy(ctx, NULL, NULL);
free_ctx:
    free(ctx);
err:
    return NULL;
}

static int spdk_emu_ctx_io_threads_create(struct spdk_emu_ctx *ctrl_ctx)
{
    int i;
    struct spdk_emu_io_thread *thread;

    ctrl_ctx->num_io_threads = spdk_io_mgr_get_num_threads();
    ctrl_ctx->io_threads = calloc(ctrl_ctx->num_io_threads,
                                  sizeof(*ctrl_ctx->io_threads));
    if (!ctrl_ctx->io_threads)
        return -ENOMEM;

    for (i = 0; i < ctrl_ctx->num_io_threads; i++) {
        thread = &ctrl_ctx->io_threads[i];
        thread->spdk_thread = spdk_io_mgr_get_thread(i);
        thread->id = i;
        thread->ctrl_ctx = ctrl_ctx;
        thread->spdk_thread_creator = spdk_get_thread();
        spdk_thread_send_msg(thread->spdk_thread,
                             spdk_emu_thread_register_poller, thread);
    }

    return 0;
}

static void spdk_emu_ctrl_destroy(struct spdk_emu_ctx *ctx,
                                  void (*done_cb)(void *arg),
                                  void *done_cb_arg)
{
    ctx->ctrl_ops = NULL;

    switch (ctx->protocol) {
    case SPDK_EMU_PROTOCOL_NVME:
        nvme_ctrl_stop(ctx->ctrl);
        nvme_ctrl_destroy(ctx->ctrl);
        nvme_ctrl_config_destroy(ctx->sconfig);
	/* NVMe destruction is synchronous, can immediately call done_cb */
        if (done_cb)
            done_cb(done_cb_arg);
        break;
    case SPDK_EMU_PROTOCOL_VIRTIO_BLK:
        virtio_blk_ctrl_destroy(ctx->ctrl, done_cb, done_cb_arg);
        break;
    case SPDK_EMU_PROTOCOL_VIRTIO_FS:
        virtio_fs_ctrl_destroy(ctx->ctrl);
        //TODO: Add virtio fs async destruction flow
        if (done_cb)
            done_cb(done_cb_arg);
        break;
    default:
        SPDK_ERRLOG("Unknown controller type %d\n", ctx->protocol);
    }
}

static void spdk_emu_rpc_controller_list(struct spdk_jsonrpc_request *request,
                                         const struct spdk_json_val *params)
{
    struct spdk_json_write_ctx *w;
    struct spdk_emu_ctx *ctx;

    if (params) {
        spdk_jsonrpc_send_error_response(request,
                    SPDK_JSONRPC_ERROR_INVALID_PARAMS,
                    "No parameters required");
        return;
    }

    pthread_mutex_lock(&spdk_emu_list_lock);
    w = spdk_jsonrpc_begin_result(request);

    spdk_json_write_array_begin(w);
    LIST_FOREACH(ctx, &spdk_emu_list, entry) {
        spdk_json_write_object_begin(w);
        if (SPDK_EMU_PROTOCOL_NVME == ctx->protocol) {
            struct nvme_ctrl_caps nvme_caps = {};

            nvme_ctrl_query_caps((nvme_ctrl_t *)ctx->ctrl, &nvme_caps);
            spdk_json_write_named_string(w, "subnqn", ctx->nvme_subsys->subnqn);
            spdk_json_write_named_int32(w, "cntlid",
                          snap_config_get_cntlid(ctx->sconfig));
            spdk_json_write_named_string(w, "version", nvme_caps.version);
            spdk_json_write_named_bool(w, "offload", nvme_caps.offload);
            spdk_json_write_named_bool(w, "mempool", nvme_caps.mempool);
            spdk_json_write_named_uint32(w, "max_nsid", nvme_caps.max_nsid);
            spdk_json_write_named_uint32(w, "max_namespaces",
                                         nvme_caps.max_namespaces);
        } else if (SPDK_EMU_PROTOCOL_VIRTIO_BLK == ctx->protocol) {
            spdk_json_write_named_bool(w, "mempool",
                          virtio_blk_ctrl_dma_pool_enabled(ctx->ctrl));
        }
        spdk_json_write_named_string(w, "name", ctx->emu_name);
        spdk_json_write_named_string(w, "emulation_manager", ctx->emu_manager);
        spdk_json_write_named_string(w, "type",
                                     spdk_emu_protocol_to_string(ctx->protocol));
        spdk_json_write_named_uint32(w, "pci_index", ctx->spci->id);
        spdk_json_write_named_string(w, "pci_bdf", ctx->spci->pci_number);
        spdk_json_write_object_end(w);
    }
    spdk_json_write_array_end(w);

    spdk_jsonrpc_end_result(request, w);
    pthread_mutex_unlock(&spdk_emu_list_lock);
    return;
}

SPDK_RPC_REGISTER("controller_list", spdk_emu_rpc_controller_list,
                  SPDK_RPC_RUNTIME)

static void
spdk_emu_pcifunc_to_json(struct spdk_json_write_ctx *w,
                         struct ibv_device *ib_dev, struct snap_pci *pci)
{
    struct spdk_emu_ctx *ctx;

    spdk_json_write_named_string(w, "emulation_manager",
                                 ibv_get_device_name(ib_dev));
    spdk_json_write_named_bool(w, "hotplugged", pci->hotplugged);
    switch (pci->type) {
        case SNAP_NVME_PF:
        case SNAP_NVME_VF:
            spdk_json_write_named_string(w, "emulation_type",
                                         "nvme");
            break;
        case SNAP_VIRTIO_NET_PF:
        case SNAP_VIRTIO_NET_VF:
            spdk_json_write_named_string(w, "emulation_type",
                                         "virtio_net");
            break;
        case SNAP_VIRTIO_BLK_PF:
        case SNAP_VIRTIO_BLK_VF:
            spdk_json_write_named_string(w, "emulation_type",
                                         "virtio_blk");
            break;
        case SNAP_VIRTIO_FS_PF:
        case SNAP_VIRTIO_FS_VF:
            spdk_json_write_named_string(w, "emulation_type",
                                         "virtio_fs");
            break;
        default:
            spdk_json_write_named_string(w, "emulation_type",
                                         "unknown");
    }

    switch (pci->type) {
        case SNAP_NVME_PF:
        case SNAP_VIRTIO_BLK_PF:
        case SNAP_VIRTIO_FS_PF:
        case SNAP_VIRTIO_NET_PF:
            spdk_json_write_named_uint32(w, "pf_index", pci->id);
            spdk_json_write_named_string(w, "VUID", pci->vuid);
            break;
        case SNAP_NVME_VF:
        case SNAP_VIRTIO_BLK_VF:
        case SNAP_VIRTIO_FS_VF:
        case SNAP_VIRTIO_NET_VF:
            spdk_json_write_named_uint32(w, "pf_index", pci->parent->id);
            spdk_json_write_named_uint32(w, "vf_index", pci->id);
            spdk_json_write_named_string(w, "VUID", pci->vuid);
            break;
        default:
            spdk_json_write_named_string(w, "pci_type",
                    "unknown");
    }

    spdk_json_write_named_string(w, "pci_bdf", pci->pci_number);

    spdk_json_write_named_object_begin(w, "controller");
    pthread_mutex_lock(&spdk_emu_list_lock);
    LIST_FOREACH(ctx, &spdk_emu_list, entry) {
        if (!strncmp(ctx->spci->vuid, pci->vuid, 128)
            && !strcmp(ctx->emu_manager, ibv_get_device_name(ib_dev))) {
            if (SPDK_EMU_PROTOCOL_NVME == ctx->protocol) {
                spdk_json_write_named_string(w, "subnqn",
                                     ctx->nvme_subsys->subnqn);
                spdk_json_write_named_int32(w, "cntlid",
                              snap_config_get_cntlid(ctx->sconfig));
            }
            spdk_json_write_named_string(w, "name", ctx->emu_name);
        }
    }
    pthread_mutex_unlock(&spdk_emu_list_lock);
    spdk_json_write_object_end(w);
}

static void
spdk_emu_rpc_emulation_functions_list(struct spdk_jsonrpc_request *request,
                                      const struct spdk_json_val *params)
{
    struct spdk_json_write_ctx *w;
    struct ibv_device **ibv_list;
    struct snap_pci **spci_list;
    int ibv_count, spci_count, i, j, k;
    enum snap_emulation_type type;

    if (params) {
        spdk_jsonrpc_send_error_response(request,
                    SPDK_JSONRPC_ERROR_INVALID_PARAMS,
                    "No parameters required");
        return;
    }

    w = spdk_jsonrpc_begin_result(request);
    spdk_json_write_array_begin(w);

    ibv_list = ibv_get_device_list(&ibv_count);
    if (!ibv_list) {
        SPDK_ERRLOG("failed to open IB device list\n");
        goto err;
    }

    for (i = 0; i < ibv_count; i++) {
        for (type = SNAP_NVME; type <= SNAP_VIRTIO_FS; type <<= 1) {
            spci_list = mlnx_snap_get_snap_pci_list(type,
                                              ibv_get_device_name(ibv_list[i]),
                                              &spci_count);
            if (!spci_list)
                continue;

            for (j = 0; j < spci_count; j++) {
                if (!spci_list[j]->plugged)
                    continue;

                spdk_json_write_object_begin(w);
                spdk_emu_pcifunc_to_json(w, ibv_list[i], spci_list[j]);
                spdk_json_write_named_uint32(w, "num_vfs",
                                             spci_list[j]->num_vfs);
                if (spci_list[j]->num_vfs) {
                    spdk_json_write_named_array_begin(w, "virtual_functions");
                    for (k = 0; k < spci_list[j]->num_vfs; k++) {
                        spdk_json_write_object_begin(w);
                        spdk_emu_pcifunc_to_json(w, ibv_list[i],
                                                 &spci_list[j]->vfs[k]);
                        spdk_json_write_object_end(w);
                    }
                    spdk_json_write_array_end(w);
                }
                spdk_json_write_object_end(w);
            }

            free(spci_list);
        }

    }

    free(ibv_list);

    spdk_json_write_array_end(w);
    spdk_jsonrpc_end_result(request, w);
    return;

err:
    spdk_jsonrpc_send_error_response(request,
                                     SPDK_JSONRPC_ERROR_INVALID_REQUEST,
                                     "Request failed");
}

SPDK_RPC_REGISTER("emulation_functions_list",
                  spdk_emu_rpc_emulation_functions_list, SPDK_RPC_RUNTIME)

static void
spdk_emu_rpc_emulation_managers_list(struct spdk_jsonrpc_request *request,
                                     const struct spdk_json_val *params)
{
    struct spdk_json_write_ctx *w;
    struct ibv_device **ibv_list;
    const char *rdma_device;
    struct snap_context *sctx;
    int ibv_count, i;

    if (params) {
        spdk_jsonrpc_send_error_response(request,
                    SPDK_JSONRPC_ERROR_INVALID_PARAMS,
                    "No parameters required");
        return;
    }

    w = spdk_jsonrpc_begin_result(request);
    spdk_json_write_array_begin(w);

    ibv_list = ibv_get_device_list(&ibv_count);
    if (!ibv_list) {
        SPDK_ERRLOG("failed to open IB device list\n");
        goto err;
    }

    for (i = 0; i < ibv_count; i++) {
        rdma_device = ibv_get_device_name(ibv_list[i]);

        sctx = mlnx_snap_get_snap_context(rdma_device);
        if (!sctx)
            continue;

        spdk_json_write_object_begin(w);
        spdk_json_write_named_string(w, "emulation_manager", rdma_device);
        spdk_json_write_named_bool(w, "hotplug_support",
                                   sctx->hotplug_supported);
        spdk_json_write_named_array_begin(w, "supported_types");
        if (sctx->emulation_caps & SNAP_NVME)
            spdk_json_write_string(w, "nvme");
        if (sctx->emulation_caps & SNAP_VIRTIO_BLK)
            spdk_json_write_string(w, "virtio_blk");
        if (sctx->emulation_caps & SNAP_VIRTIO_NET)
            spdk_json_write_string(w, "virtio_net");
        if (sctx->emulation_caps & SNAP_VIRTIO_FS)
            spdk_json_write_string(w, "virtio_fs");
        spdk_json_write_array_end(w);
        spdk_json_write_object_end(w);
    }


    free(ibv_list);

    spdk_json_write_array_end(w);
    spdk_jsonrpc_end_result(request, w);
    return;

err:
    spdk_jsonrpc_send_error_response(request,
                                     SPDK_JSONRPC_ERROR_INVALID_REQUEST,
                                     "Request failed");
}

SPDK_RPC_REGISTER("emulation_managers_list",
                  spdk_emu_rpc_emulation_managers_list, SPDK_RPC_RUNTIME)

struct spdk_emu_device_detach_attr {
    char *emu_manager;
    char *device_type;
    char *pci_bdf;
    int   pci_index;
    bool  force;
    char  *vuid;
    int timeout;
};

static struct snap_context *
spdk_emu_find_hotplug_manager(const char *rdma_device)
{
    struct snap_context *sctx;

    /* Verify RDMA device exist in ibv list */
    sctx = mlnx_snap_get_snap_context(rdma_device);
    if (!sctx) {
        SPDK_ERRLOG("Cannot find snap context on %s\n", rdma_device);
        return NULL;
    } else if (!sctx->hotplug_supported) {
        SPDK_ERRLOG("RDMA device %s is not hotplug manager\n", rdma_device);
        return NULL;
    }

    return sctx;
}


static const struct spdk_json_object_decoder
spdk_emu_device_detach_decoder[] = {
    {
        "emulation_manager",
        offsetof(struct spdk_emu_device_detach_attr, emu_manager),
        spdk_json_decode_string
    },
    {
        "device_type",
        offsetof(struct spdk_emu_device_detach_attr, device_type),
        spdk_json_decode_string
    },
    {
        "pci_bdf",
        offsetof(struct spdk_emu_device_detach_attr, pci_bdf),
        spdk_json_decode_string,
        true
    },
    {
        "pci_index",
        offsetof(struct spdk_emu_device_detach_attr, pci_index),
        spdk_json_decode_int32,
        true
    },
    {
        "force",
        offsetof(struct spdk_emu_device_detach_attr, force),
        spdk_json_decode_bool,
        true
    },
    {
        "vuid",
        offsetof(struct spdk_emu_device_detach_attr, vuid),
        spdk_json_decode_string,
        true
    },
    {
        "timeout",
        offsetof(struct spdk_emu_device_detach_attr, timeout),
        spdk_json_decode_int32,
        true
    },

};

static void
spdk_emu_rpc_device_detach(struct spdk_jsonrpc_request *request,
                           const struct spdk_json_val *params)
{
    struct spdk_emu_device_detach_attr *attr = NULL;
    struct snap_context *sctx;
    struct snap_pci *spci;
    struct spdk_emu_ctx *ctx;
    enum spdk_emu_protocol protocol;
    enum snap_emulation_type type;
    struct spdk_emu_device_hotunplug_ctx *hotunplug_ctx;
    int ret;

    attr = calloc(1, sizeof(*attr));
    if (!attr)
        goto invalid;

    // Set invalid index, to identify when value was not decoded
    attr->pci_index = -1;

    if (spdk_json_decode_object(params,
            spdk_emu_device_detach_decoder,
            SPDK_COUNTOF(spdk_emu_device_detach_decoder),
            attr)) {
        SPDK_ERRLOG("Failed to decode parameters\n");
        goto free_attr;
    }

    if (!strncmp(attr->device_type, "nvme", 4))
        type = SNAP_NVME;
    else if (!strncmp(attr->device_type, "virtio_blk", 10))
        type = SNAP_VIRTIO_BLK;
    else if (!strncmp(attr->device_type, "virtio_fs", 9))
        type = SNAP_VIRTIO_FS;
    else {
        SPDK_ERRLOG("Invalid device type: %s\n", attr->device_type);
        goto free_attr_strings;
    }

    sctx = spdk_emu_find_hotplug_manager(attr->emu_manager);
    if (!sctx) {
        SPDK_ERRLOG("Cannot find hotplug snap context\n");
        goto free_attr_strings;
    }

    spci = mlnx_snap_get_pf_snap_pci_by_vuid(type, attr->emu_manager, attr->vuid);
    if (!attr->vuid && !spci)
         spci = mlnx_snap_get_snap_pci(type, attr->emu_manager,
                                       attr->pci_bdf, attr->pci_index, -1);
    if (!spci || !spci->plugged) {
        if (attr->pci_index > -1)
            SPDK_ERRLOG("PCI ID %d is not plugged\n", attr->pci_index);
        else if (attr->pci_bdf)
            SPDK_ERRLOG("PCI BDF %s is not plugged\n", attr->pci_bdf);
        else
            SPDK_ERRLOG("VUID %s is not plugged\n", attr->vuid);
        goto free_attr_strings;
    }
    if (!spci->hotplugged) {
        nvmx_error("PCI function %s is not hotpluggable", spci->pci_number);
        goto free_attr_strings;
    }

    if (spci->type == SNAP_NVME_PF)
        protocol = SPDK_EMU_PROTOCOL_NVME;
    else if (spci->type == SNAP_VIRTIO_BLK_PF)
        protocol = SPDK_EMU_PROTOCOL_VIRTIO_BLK;
    else
        protocol = SPDK_EMU_PROTOCOL_VIRTIO_FS;

    ctx = spdk_emu_ctx_find_by_pci_id(attr->emu_manager, spci->id, -1, protocol);
    if (!ctx) {
        SPDK_ERRLOG("No controller attached, cannot detach PCI function %s\n",
                    spci->pci_number);
        goto free_attr_strings;
    }

    if (ctx->hotunplug_ctx) {
        if (ctx->hotunplug_ctx->timeout_once && attr->force) {
            /*
             * reach to here meaning we tried hotunplug once without --force flag
             * but it timeouted, and we are doing the second try with --force flag.
             */
            ctx->hotunplug_ctx->request = request;
            spdk_emu_ctrl_destroy_device_detach_cb(ctx);
            goto out;
        }

        nvmx_error("Cannot detach PCI function %s previous detach still in progress.", spci->pci_number);
        goto free_attr_strings;
    }

    hotunplug_ctx = calloc(1, sizeof(*hotunplug_ctx));
    if (!hotunplug_ctx) {
        SPDK_ERRLOG("Failed to allocate hotunplug context\n");
        goto free_attr_strings;
    }
    hotunplug_ctx->spci = spci;
    hotunplug_ctx->request = request;

    ctx->hotunplug_ctx = hotunplug_ctx;

    if (sctx->hotplug.pci_hotplug_state_change) {
        // Graceful hotunplug - wait for POWER_OFF
        if (spci->type == SNAP_NVME_PF) {
            struct nvme_ctrl *ctrl = ctx->ctrl;

            ret = snap_nvme_ctrl_hotunplug(ctrl->dev);
            if (ret) {
                nvmx_error("Failed to hotunplug emulation function");
                goto free_hotunplug_ctx;
            }

            if (time(&ctrl->hotunplug_timer_expire) == (time_t)(-1)) {
                nvmx_error("Failed to get system time");
                goto free_hotunplug_ctx;
            }
            ctrl->hotunplug_timer_expire += (attr->timeout ? attr->timeout : HOTUNPLUG_TIMEOUT);
            ctrl->force_hotunplug_device = attr->force;

        } else if (spci->type == SNAP_VIRTIO_BLK_PF) {
            struct virtio_blk_ctrl *ctrl = ctx->ctrl;
            struct snap_virtio_ctrl *sctrl = &ctrl->sctrl->common;

            /*
             * If io was sent after ctrl was suspended we won't get POWER_OFF
             * from host. Driver would wait for io to complete before detaching and
             * be blocked. To avoid this resume suspended ctrl before hotunplug.
             * Always resume - if ctrl was not suspended it will not resume,
             * if ctrl was in suspending state it will resume ince suspended.
             */
            ret = virtio_blk_ctrl_resume(ctrl);
            if (ret)
                nvmx_warn("Failed to resume suspended ctrl before hotunplug\n");

            ret = snap_virtio_ctrl_hotunplug(sctrl);
            if (ret)
                goto free_hotunplug_ctx;

            if (time(&ctrl->hotunplug_timer_expire) == (time_t)(-1)) {
                nvmx_error("Failed to get system time");
                goto free_hotunplug_ctx;
            }
            // TODO: once we can respond on an rpc request that timed out from rpc timeout
            // without snap crush. Remove hotunplug timeout and use rpc timeout instead.
            ctrl->hotunplug_timer_expire += (attr->timeout ? attr->timeout : HOTUNPLUG_TIMEOUT);
            ctrl->force_hotunplug_device = attr->force;
        } else {
            // Graceful hotunplug is not supported for virtio_fs. Surprise hotunplug
            spdk_emu_ctrl_destroy_device_detach_cb(ctx);
        }
    } else {
        // Surprise hotunplug - don't wait for POWER_OFF
        nvmx_debug("Hotunplug using surprise unplug - cap for graceful unplug is false");
        spdk_emu_ctrl_destroy_device_detach_cb(ctx);
    }

out:
    free(attr->emu_manager);
    free(attr->device_type);
    free(attr->pci_bdf);
    free(attr->vuid);
    free(attr);
    return;

free_hotunplug_ctx:
    free(hotunplug_ctx);
    ctx->hotunplug_ctx = NULL;
free_attr_strings:
    free(attr->emu_manager);
    free(attr->device_type);
    free(attr->pci_bdf);
    free(attr->vuid);
free_attr:
    free(attr);
invalid:
    spdk_jsonrpc_send_error_response(request,
                     SPDK_JSONRPC_ERROR_INVALID_PARAMS,
                     "Invalid parameters");
}

SPDK_RPC_REGISTER("emulation_device_detach",
                  spdk_emu_rpc_device_detach, SPDK_RPC_RUNTIME)

struct spdk_emu_ctx *
spdk_emu_hotunplug_create_controller(void  *arg, struct ibv_device *ibv_dev)
{
    struct spdk_emu_ctx *ctx;
    struct spdk_emu_ctx_create_attr emu_attr = {};
    struct spdk_emu_controller_vblk_create_attr *create_attr = NULL;
    struct snap_pci *spci = arg;
    char *emu_manager;
    struct spdk_emu_device_hotunplug_ctx *hotunplug_ctx;
    struct virtio_blk_ctrl *ctrl;

    create_attr = calloc(1, sizeof(*create_attr));
    if (!create_attr)
            goto error;

    emu_manager = strndup(ibv_get_device_name(ibv_dev), SPDK_EMU_MANAGER_NAME_MAXLEN - 1);
    if (!emu_manager){
            free(create_attr);
            goto error;
    }

    pthread_mutex_lock(&spdk_emu_list_lock);
    create_attr->bdev_type = "none";
    create_attr->num_queues = 0;
    create_attr->pf_id = spci->id;
    create_attr->recover = true;
    create_attr->emu_manager = emu_manager;
    create_attr->force_recover = true;
    create_attr->vf_id = -1;

    emu_attr.priv = create_attr;
    emu_attr.protocol = SPDK_EMU_PROTOCOL_VIRTIO_BLK;
    emu_attr.emu_manager = emu_manager;
    emu_attr.spci = spci;

    ctx = spdk_emu_ctx_create(&emu_attr);
    if (!ctx) {
            pthread_mutex_unlock(&spdk_emu_list_lock);
            free(create_attr);
            free(emu_manager);
            goto error;
    }

    LIST_INSERT_HEAD(&spdk_emu_list, ctx, entry);
    pthread_mutex_unlock(&spdk_emu_list_lock);

    hotunplug_ctx = calloc(1, sizeof(*hotunplug_ctx));
    if (!hotunplug_ctx) {
        SPDK_ERRLOG("Failed to allocate hotunplug context\n");
        return NULL;
    }

    hotunplug_ctx->spci = spci;
    hotunplug_ctx->request = NULL;
    ctx->hotunplug_ctx = hotunplug_ctx;

    ctrl = ctx->ctrl;
    ctrl->hotunplug_timer_expire = -1;

    free(create_attr);
    free(emu_manager);

    nvmx_info("Created dummy controller for hotplug function which failed to be unplugged before snap started");
    return ctx;

error:
    nvmx_error("Failed to create dummy controller for hotunplug");
    return NULL;
}

struct spdk_emu_controller_vblk_bdev_detach_attr {
    char *ctrl_name;
};

static const struct spdk_json_object_decoder
spdk_emu_rpc_controller_vblk_bdev_detach_decoder[] = {
    {
        "name",
        offsetof(struct spdk_emu_controller_vblk_bdev_detach_attr, ctrl_name),
        spdk_json_decode_string
    },
};

static void
spdk_emu_controller_vblk_bdev_detach_end(void *arg, bool success)
{
    struct spdk_jsonrpc_request *request = arg;
    struct spdk_json_write_ctx *w;

    if (success) {
        SPDK_NOTICELOG("Bdev was detached from controller successfully\n");
        w = spdk_jsonrpc_begin_result(request);
        spdk_json_write_string(w, "Bdev detached");
        spdk_jsonrpc_end_result(request, w);
    } else {
        spdk_jsonrpc_send_error_response(request,
                     SPDK_JSONRPC_ERROR_INTERNAL_ERROR,
                     "Invalid parameters");
   }

}

static void
spdk_emu_rpc_controller_vblk_bdev_detach(struct spdk_jsonrpc_request *request,
                                         const struct spdk_json_val *params)
{
    struct spdk_emu_controller_vblk_bdev_detach_attr *attr = NULL;
    struct spdk_emu_ctx *ctx;

    attr = calloc(1, sizeof(*attr));
    if (!attr)
        goto invalid;

    if (spdk_json_decode_object(params,
            spdk_emu_rpc_controller_vblk_bdev_detach_decoder,
            SPDK_COUNTOF(spdk_emu_rpc_controller_vblk_bdev_detach_decoder),
            attr)) {
        SPDK_ERRLOG("Failed to decode parameters\n");
        goto free_attr;
    }

    ctx = spdk_emu_ctx_find_by_emu_name(attr->ctrl_name);
    if (!ctx) {
        SPDK_ERRLOG("emulation %s cannot be found\n", attr->ctrl_name);
        goto free_attr_strings;
    }

    virtio_blk_ctrl_bdev_detach(ctx->ctrl, request,
                    spdk_emu_controller_vblk_bdev_detach_end);

    free(attr->ctrl_name);
    free(attr);
    return;

free_attr_strings:
    free(attr->ctrl_name);
free_attr:
    free(attr);
invalid:
    spdk_jsonrpc_send_error_response(request,
                     SPDK_JSONRPC_ERROR_INVALID_PARAMS,
                     "Invalid parameters");
}

SPDK_RPC_REGISTER("controller_virtio_blk_bdev_detach",
                  spdk_emu_rpc_controller_vblk_bdev_detach, SPDK_RPC_RUNTIME)

struct spdk_emu_controller_vblk_get_debugstat_attr {
    char *ctrl_name;
};

static const struct spdk_json_object_decoder
spdk_emu_rpc_controller_vblk_get_debugstat_decoder[] = {
    {
        "name",
        offsetof(struct spdk_emu_controller_vblk_get_debugstat_attr, ctrl_name),
        spdk_json_decode_string,
        true
    },
};

static void
spdk_emu_ctrl_debugstat_to_json(struct spdk_json_write_ctx *w,
                              struct snap_virtio_ctrl_debugstat *ctrl_debugstat)
{
    struct snap_virtio_queue_debugstat *q_debugstat;
    int i;

    spdk_json_write_named_object_begin(w, "global");
    spdk_json_write_named_uint32(w, "network_error",
                                 ctrl_debugstat->network_error);
    spdk_json_write_named_uint32(w, "bad_descriptor_error",
                                 ctrl_debugstat->bad_descriptor_error);
    spdk_json_write_named_uint32(w, "invalid_buffer",
                                 ctrl_debugstat->invalid_buffer);
    spdk_json_write_named_uint32(w, "desc_list_exceed_limit",
                                 ctrl_debugstat->desc_list_exceed_limit);
    spdk_json_write_named_uint32(w, "internal_error",
                                 ctrl_debugstat->internal_error);
    spdk_json_write_object_end(w);

    spdk_json_write_named_array_begin(w, "queues");
    for (i = 0; i < ctrl_debugstat->num_queues; i++) {
        q_debugstat = &ctrl_debugstat->queues[i];

        spdk_json_write_object_begin(w);
        spdk_json_write_named_uint32(w, "qid", q_debugstat->qid);
        spdk_json_write_named_uint32(w, "hw_available_index",
                                     q_debugstat->hw_available_index);
        spdk_json_write_named_uint32(w, "sw_available_index",
                                     q_debugstat->sw_available_index);
        spdk_json_write_named_uint32(w, "hw_used_index",
                                     q_debugstat->hw_used_index);
        spdk_json_write_named_uint32(w, "sw_used_index",
                                     q_debugstat->sw_used_index);
        spdk_json_write_named_uint32(w, "hw_received_descs",
                                     q_debugstat->hw_received_descs);
        spdk_json_write_named_uint32(w, "hw_completed_descs",
                                     q_debugstat->hw_completed_descs);
        spdk_json_write_object_end(w);
    }
    spdk_json_write_array_end(w);
}

static void
spdk_emu_rpc_controller_vblk_get_debugstat(struct spdk_jsonrpc_request *request,
                                           const struct spdk_json_val *params)
{
    struct spdk_emu_controller_vblk_get_debugstat_attr *attr = NULL;
    struct spdk_emu_ctx *ctx;
    struct spdk_json_write_ctx *w;
    struct snap_context *sctx;

    attr = calloc(1, sizeof(*attr));
    if (!attr) {
        SPDK_ERRLOG("alloc memory for `attr` failed\n");
        goto err;
    }

    if (params && spdk_json_decode_object(params,
            spdk_emu_rpc_controller_vblk_get_debugstat_decoder,
            SPDK_COUNTOF(spdk_emu_rpc_controller_vblk_get_debugstat_decoder),
            attr)) {
        SPDK_ERRLOG("Failed to decode parameters\n");
        goto free_attr;
    }

    if (params && !spdk_emu_ctx_find_by_emu_name(attr->ctrl_name)) {
        SPDK_ERRLOG("emulation %s cannot be found\n", attr->ctrl_name);
        goto free_attr_strings;
    }

    w = spdk_jsonrpc_begin_result(request);
    spdk_json_write_object_begin(w);
    spdk_json_write_named_array_begin(w, "controllers");

    pthread_mutex_lock(&spdk_emu_list_lock);
    LIST_FOREACH(ctx, &spdk_emu_list, entry) {
        struct snap_virtio_ctrl_debugstat *ctrl_debugstat;

        if (params && strncmp(ctx->emu_name, attr->ctrl_name,
                              SPDK_EMU_NAME_MAXLEN))
            continue;

        if (ctx->protocol != SPDK_EMU_PROTOCOL_VIRTIO_BLK) {
            if (params)
                SPDK_ERRLOG("Given controller name is of wrong type (%s)\n",
                            spdk_emu_protocol_to_string(ctx->protocol));
            continue;
        }

        sctx = mlnx_snap_get_snap_context(ctx->emu_manager);
        if (!sctx)
            continue;

        ctrl_debugstat = calloc(1, sizeof(*ctrl_debugstat));
        if (!ctrl_debugstat) {
            SPDK_ERRLOG("alloc memory for `ctrl_debugstat` failed\n");
            continue;
        }

        ctrl_debugstat->queues =
            calloc(sctx->virtio_blk_caps.max_emulated_virtqs,
                   sizeof(struct snap_virtio_queue_debugstat));
        if (!ctrl_debugstat->queues) {
            SPDK_ERRLOG("alloc memory for `attr` failed\n");
            free(ctrl_debugstat);
            continue;
        }

        if (virtio_blk_ctrl_get_debugstat(ctx->ctrl, ctrl_debugstat)) {
            SPDK_ERRLOG("get debugstat failed for emu %s\n", ctx->emu_name);
            free(ctrl_debugstat->queues);
            free(ctrl_debugstat);
            continue;
        }

        spdk_json_write_object_begin(w);
        spdk_json_write_named_string(w, "name", ctx->emu_name);
        spdk_emu_ctrl_debugstat_to_json(w, ctrl_debugstat);
        spdk_json_write_object_end(w);

        free(ctrl_debugstat->queues);
        free(ctrl_debugstat);
    }
    pthread_mutex_unlock(&spdk_emu_list_lock);

    spdk_json_write_array_end(w);
    spdk_json_write_object_end(w);
    spdk_jsonrpc_end_result(request, w);

    free(attr->ctrl_name);
    free(attr);
    return;

free_attr_strings:
    free(attr->ctrl_name);
free_attr:
    free(attr);
err:
    spdk_jsonrpc_send_error_response(request,
                     SPDK_JSONRPC_ERROR_INVALID_PARAMS,
                     "Invalid parameters");
}

SPDK_RPC_REGISTER("controller_virtio_blk_get_debugstat",
                  spdk_emu_rpc_controller_vblk_get_debugstat, SPDK_RPC_RUNTIME)

struct spdk_emu_controller_vblk_bdev_attach_attr {
    char *ctrl_name;
    char *bdev_type;
    char *bdev;
    uint32_t size_max;
    uint32_t seg_max;
    char *serial;
};

static const struct spdk_json_object_decoder
spdk_emu_rpc_controller_vblk_bdev_attach_decoder[] = {
    {
        "name",
        offsetof(struct spdk_emu_controller_vblk_bdev_attach_attr, ctrl_name),
        spdk_json_decode_string
    },
    {
        "bdev_type",
        offsetof(struct spdk_emu_controller_vblk_bdev_attach_attr, bdev_type),
        spdk_json_decode_string
    },
    {
        "bdev",
        offsetof(struct spdk_emu_controller_vblk_bdev_attach_attr, bdev),
        spdk_json_decode_string
    },
    {
        "size_max",
        offsetof(struct spdk_emu_controller_vblk_bdev_attach_attr, size_max),
        spdk_json_decode_uint32,
        true
    },
    {
        "seg_max",
        offsetof(struct spdk_emu_controller_vblk_bdev_attach_attr, seg_max),
        spdk_json_decode_uint32,
        true
    },
    {
        "serial",
        offsetof(struct spdk_emu_controller_vblk_bdev_attach_attr, serial),
        spdk_json_decode_string,
        true
    },
};

static void
spdk_emu_rpc_controller_vblk_bdev_attach(struct spdk_jsonrpc_request *request,
                                         const struct spdk_json_val *params)
{
    struct spdk_emu_controller_vblk_bdev_attach_attr *attr = NULL;
    struct spdk_emu_ctx *ctx;
    struct spdk_json_write_ctx *w;

    attr = calloc(1, sizeof(*attr));
    if (!attr)
        goto invalid;

    if (spdk_json_decode_object(params,
            spdk_emu_rpc_controller_vblk_bdev_attach_decoder,
            SPDK_COUNTOF(spdk_emu_rpc_controller_vblk_bdev_attach_decoder),
            attr)) {
        SPDK_ERRLOG("Failed to decode parameters\n");
        goto free_attr;
    }

    ctx = spdk_emu_ctx_find_by_emu_name(attr->ctrl_name);
    if (!ctx) {
        SPDK_ERRLOG("emulation %s cannot be found\n", attr->ctrl_name);
        goto free_attr_strings;
    }

    if (strncmp(attr->bdev_type, "spdk", 4)) {
        SPDK_ERRLOG("Bdev type '%s' is not supported\n", attr->bdev_type);
        goto free_attr_strings;
    }

    if (virtio_blk_ctrl_bdev_attach(ctx->ctrl, attr->bdev_type, attr->bdev,
                                    attr->size_max, attr->seg_max, attr->serial))
        goto free_attr_strings;

    nvmx_info("Bdev %s was attached to controller %s successfully",
              attr->bdev, attr->ctrl_name);

    w = spdk_jsonrpc_begin_result(request);
    spdk_json_write_string(w, ctx->emu_name);
    spdk_jsonrpc_end_result(request, w);

    free(attr->bdev_type);
    free(attr->bdev);
    free(attr->ctrl_name);
    free(attr->serial);
    free(attr);
    return;

free_attr_strings:
    free(attr->bdev_type);
    free(attr->bdev);
    free(attr->ctrl_name);
    free(attr->serial);
free_attr:
    free(attr);
invalid:
    spdk_jsonrpc_send_error_response(request,
                     SPDK_JSONRPC_ERROR_INVALID_PARAMS,
                     "Invalid parameters");
}

SPDK_RPC_REGISTER("controller_virtio_blk_bdev_attach",
                  spdk_emu_rpc_controller_vblk_bdev_attach, SPDK_RPC_RUNTIME)

struct spdk_emu_controller_vblk_bdev_list_attr {
    char *ctrl_name;
};

static const struct spdk_json_object_decoder
spdk_emu_rpc_controller_vblk_bdev_list_decoder[] = {
    {
        "name",
        offsetof(struct spdk_emu_controller_vblk_bdev_list_attr, ctrl_name),
        spdk_json_decode_string
    },
};

static void
spdk_emu_rpc_controller_vblk_bdev_list(struct spdk_jsonrpc_request *request,
                                       const struct spdk_json_val *params)
{
    struct spdk_emu_controller_vblk_bdev_list_attr *attr = NULL;
    struct spdk_emu_ctx *ctx;
    struct spdk_json_write_ctx *w;
    struct virtio_blk_ctrl *ctrl;

    attr = calloc(1, sizeof(*attr));
    if (!attr)
        goto invalid;

    if (spdk_json_decode_object(params,
            spdk_emu_rpc_controller_vblk_bdev_list_decoder,
            SPDK_COUNTOF(spdk_emu_rpc_controller_vblk_bdev_list_decoder),
            attr)) {
        SPDK_ERRLOG("Failed to decode parameters\n");
        goto free_attr;
    }

    pthread_mutex_lock(&spdk_emu_list_lock);
    ctx = spdk_emu_ctx_find_by_emu_name(attr->ctrl_name);
    if (!ctx) {
        pthread_mutex_unlock(&spdk_emu_list_lock);
        SPDK_ERRLOG("emulation %s cannot be found\n", attr->ctrl_name);
        goto free_attr_strings;
    }

    ctrl = ctx->ctrl;

    w = spdk_jsonrpc_begin_result(request);
    spdk_json_write_object_begin(w);
    spdk_json_write_named_string(w, "name", ctx->emu_name);
    spdk_json_write_named_object_begin(w, "bdev");
    if (ctrl->bctx.desc) {  /* Have a real bdev attached case */
        spdk_json_write_named_string(w, "bdev", spdk_bdev_get_name(ctrl->bctx.bdev));
        spdk_json_write_named_string(w, "bdev_type", "spdk");
        spdk_json_write_named_uint32(w, "block_size",
                                     spdk_bdev_get_block_size(ctrl->bctx.bdev));
        spdk_json_write_named_uint64(w, "num_blocks",
                                     spdk_bdev_get_num_blocks(ctrl->bctx.bdev));
        spdk_json_write_named_bool(w, "snap-direct",
                                   virtio_blk_ctrl_is_zcopy(ctrl));
    }
    spdk_json_write_object_end(w);
    spdk_json_write_object_end(w);
    spdk_jsonrpc_end_result(request, w);

    pthread_mutex_unlock(&spdk_emu_list_lock);
    free(attr->ctrl_name);
    free(attr);
    return;

free_attr_strings:
    free(attr->ctrl_name);
free_attr:
    free(attr);
invalid:
    spdk_jsonrpc_send_error_response(request,
                     SPDK_JSONRPC_ERROR_INVALID_PARAMS,
                     "Invalid parameters");
}

SPDK_RPC_REGISTER("controller_virtio_blk_bdev_list",
                  spdk_emu_rpc_controller_vblk_bdev_list, SPDK_RPC_RUNTIME)

/* VirtIO Blk: TODO: move to virtio blk 'subsystem' */

static const struct spdk_json_object_decoder
spdk_emu_rpc_controller_vblk_create_decoder[] = {
    {
        "emulation_manager",
        offsetof(struct spdk_emu_controller_vblk_create_attr, emu_manager),
        spdk_json_decode_string
    },
    {
        "pci_bdf",
        offsetof(struct spdk_emu_controller_vblk_create_attr, pci_bdf),
        spdk_json_decode_string,
        true
    },
    {
        "pf_id",
        offsetof(struct spdk_emu_controller_vblk_create_attr, pf_id),
        spdk_json_decode_int32,
        true
    },
    {
        "vf_id",
        offsetof(struct spdk_emu_controller_vblk_create_attr, vf_id),
        spdk_json_decode_int32,
        true
    },
    {
        "bdev_type",
        offsetof(struct spdk_emu_controller_vblk_create_attr, bdev_type),
        spdk_json_decode_string
    },
    {
        "bdev",
        offsetof(struct spdk_emu_controller_vblk_create_attr, bdev),
        spdk_json_decode_string,
        true
    },
    {
        "num_queues",
        offsetof(struct spdk_emu_controller_vblk_create_attr, num_queues),
        spdk_json_decode_uint32,
        true
    },
    {
        "queue_depth",
        offsetof(struct spdk_emu_controller_vblk_create_attr, queue_depth),
        spdk_json_decode_uint32,
        true
    },
    {
        "size_max",
        offsetof(struct spdk_emu_controller_vblk_create_attr, size_max),
        spdk_json_decode_uint32,
        true
    },
    {
        "seg_max",
        offsetof(struct spdk_emu_controller_vblk_create_attr, seg_max),
        spdk_json_decode_uint32,
        true
    },
    {
        "serial",
        offsetof(struct spdk_emu_controller_vblk_create_attr, serial),
        spdk_json_decode_string,
        true
    },
    {
        "force_in_order",
        offsetof(struct spdk_emu_controller_vblk_create_attr, force_in_order),
        spdk_json_decode_bool,
        true
    },
    {
        "suspend",
        offsetof(struct spdk_emu_controller_vblk_create_attr, suspended),
        spdk_json_decode_bool,
        true
    },
    {   "recover",
        offsetof(struct spdk_emu_controller_vblk_create_attr, recover),
        spdk_json_decode_bool,
        true
    },
    {
        "mem",
        offsetof(struct spdk_emu_controller_vblk_create_attr, mem),
        spdk_json_decode_string,
        true
    },
    {
        "vuid",
        offsetof(struct spdk_emu_controller_vblk_create_attr, vuid),
        spdk_json_decode_string,
        true
    },
    {
        "admin_q",
        offsetof(struct spdk_emu_controller_vblk_create_attr, admin_q),
        spdk_json_decode_bool,
        true
    }
};

static void
spdk_emu_rpc_controller_vblk_create(struct spdk_jsonrpc_request *request,
                                    const struct spdk_json_val *params)
{
    struct spdk_emu_controller_vblk_create_attr *attr = NULL;
    struct spdk_emu_ctx_create_attr emu_attr = {};
    struct spdk_emu_ctx *ctx;
    struct spdk_json_write_ctx *w;
    struct snap_pci *spci;
    struct snap_virtio_blk_ctrl *pf_ctrl;

    attr = calloc(1, sizeof(*attr));
    if (!attr)
        goto invalid;

    // Set invalid index, to identify when value was not decoded
    attr->pf_id = -1;
    attr->vf_id = -1;

    if (spdk_json_decode_object(params,
            spdk_emu_rpc_controller_vblk_create_decoder,
            SPDK_COUNTOF(spdk_emu_rpc_controller_vblk_create_decoder),
            attr)) {
        SPDK_ERRLOG("Failed to decode parameters\n");
        goto free_attr;
    }

    if (attr->mem && strcmp(attr->mem, "static") && strcmp(attr->mem, "pool")) {
        SPDK_ERRLOG("Memory model `%s` is not supported\n", attr->mem);
        goto free_attr_strings;
    }

    spci = mlnx_snap_get_pf_snap_pci_by_vuid(SNAP_VIRTIO_BLK, attr->emu_manager, attr->vuid);
    if (!spci && !attr->vuid)
        spci = mlnx_snap_get_snap_pci(SNAP_VIRTIO_BLK, attr->emu_manager,
                                      attr->pci_bdf, attr->pf_id, attr->vf_id);
    if (!spci)
        goto free_attr_strings;

    if (strncmp(attr->bdev_type, "spdk", 4) &&
        strncmp(attr->bdev_type, "none", 4)) {
        SPDK_ERRLOG("Bdev type '%s' is not supported\n", attr->bdev_type);
        goto free_attr_strings;
    } else if (!strncmp(attr->bdev_type, "none", 4) && attr->bdev) {
        SPDK_ERRLOG("Bdev type '%s' requires no \"bdev\" attribute\n",
                    attr->bdev_type);
        goto free_attr_strings;
    } else if (!strncmp(attr->bdev_type, "spdk", 4) && !attr->bdev) {
        SPDK_ERRLOG("Bdev type '%s' requires \"bdev\" attribute\n",
                    attr->bdev_type);
        goto free_attr_strings;
    }

    pthread_mutex_lock(&spdk_emu_list_lock);

    if (attr->pci_bdf) {
        ctx = spdk_emu_ctx_find_by_pci_bdf(attr->emu_manager,
                           spci->pci_number);
    } else {
        ctx = spdk_emu_ctx_find_by_pci_id(attr->emu_manager,
                           attr->pf_id, attr->vf_id,
                           SPDK_EMU_PROTOCOL_VIRTIO_BLK);
    }

    if (ctx) {
        pthread_mutex_unlock(&spdk_emu_list_lock);
        SPDK_ERRLOG("PCI function is already in use by %s\n", ctx->emu_name);
        goto free_attr_strings;
    }

    emu_attr.priv = attr;
    emu_attr.protocol = SPDK_EMU_PROTOCOL_VIRTIO_BLK;
    emu_attr.emu_manager = attr->emu_manager;
    emu_attr.spci = spci;

    ctx = spdk_emu_ctx_create(&emu_attr);
    if (!ctx) {
        pthread_mutex_unlock(&spdk_emu_list_lock);
        goto free_attr_strings;
    }

    LIST_INSERT_HEAD(&spdk_emu_list, ctx, entry);

    if(attr->vf_id >= 0) {
        pf_ctrl = ((struct virtio_blk_ctrl*)spdk_emu_ctx_find_by_pci_id(attr->emu_manager,
                            attr->pf_id, -1, SPDK_EMU_PROTOCOL_VIRTIO_BLK)->ctrl)->sctrl;
        pf_ctrl->vfs_ctrl[attr->vf_id] = ((struct virtio_blk_ctrl*)ctx->ctrl)->sctrl;

    }

    pthread_mutex_unlock(&spdk_emu_list_lock);

    if (spdk_emu_rpc_finished) {
        nvmx_warn("spdk rpc is closed. Not returning response for rpc request");
    } else {
        w = spdk_jsonrpc_begin_result(request);
        spdk_json_write_string(w, ctx->emu_name);
        spdk_jsonrpc_end_result(request, w);
    }

    free(attr->serial);
    free(attr->bdev_type);
    free(attr->bdev);
    free(attr->pci_bdf);
    free(attr->emu_manager);
    free(attr->mem);
    free(attr->vuid);
    free(attr);
    return;

free_attr_strings:
    free(attr->serial);
    free(attr->bdev_type);
    free(attr->bdev);
    free(attr->pci_bdf);
    free(attr->emu_manager);
    free(attr->mem);
    free(attr->vuid);
free_attr:
    free(attr);
invalid:
    if (spdk_emu_rpc_finished) {
        nvmx_warn("spdk rpc is closed. Not returning error response for rpc request");
    } else {
        spdk_jsonrpc_send_error_response(request,
                         SPDK_JSONRPC_ERROR_INVALID_PARAMS,
                         "Invalid parameters");
    }
}

SPDK_RPC_REGISTER("controller_virtio_blk_create",
                  spdk_emu_rpc_controller_vblk_create, SPDK_RPC_RUNTIME)

static const struct spdk_json_object_decoder
spdk_emu_rpc_controller_vblk_delete_decoder[] = {
    {
        "name",
        offsetof(struct spdk_emu_controller_vblk_delete_attr, name),
        spdk_json_decode_string,
        true
    },
    {
        "vuid",
        offsetof(struct spdk_emu_controller_vblk_delete_attr, vuid),
        spdk_json_decode_string,
        true
    },
    {
        "force",
        offsetof(struct spdk_emu_controller_vblk_delete_attr, force),
        spdk_json_decode_bool,
        true
    }
};

static void
spdk_emu_rpc_controller_vblk_delete(struct spdk_jsonrpc_request *request,
                                    const struct spdk_json_val *params)
{
    struct spdk_emu_controller_vblk_delete_attr *attr = NULL;
    struct spdk_emu_ctx *ctx;
    const char *err_msg = "Invalid parameters";
    int err_code = SPDK_JSONRPC_ERROR_INVALID_PARAMS;

    attr = calloc(1, sizeof(*attr));
    if (!attr)
        goto invalid;

    if (spdk_json_decode_object(params,
            spdk_emu_rpc_controller_vblk_delete_decoder,
            SPDK_COUNTOF(spdk_emu_rpc_controller_vblk_delete_decoder),
            attr)) {
        SPDK_ERRLOG("Failed to decode parameters\n");
        goto free_attr;
    }

    /* Find SPDK context matching the requested PF id */
    pthread_mutex_lock(&spdk_emu_list_lock);
    if (attr->name)
        ctx = spdk_emu_ctx_find_by_emu_name(attr->name);
    else
        ctx = spdk_emu_ctx_find_by_vuid(attr->vuid);
    if (!ctx) {
        if (attr->name)
            SPDK_ERRLOG("emulation %s cannot be found\n", attr->name);
        else
            SPDK_ERRLOG("emulation with vuid %s cannot be found\n", attr->vuid);
        pthread_mutex_unlock(&spdk_emu_list_lock);
        goto free_attr_strings;
    }

    if (ctx->protocol != SPDK_EMU_PROTOCOL_VIRTIO_BLK) {
        SPDK_ERRLOG("Given controller name is of wrong type (%s)\n",
                    spdk_emu_protocol_to_string(ctx->protocol));
        pthread_mutex_unlock(&spdk_emu_list_lock);
        goto free_attr_strings;
    }

    if (ctx->spci->plugged && !virtio_blk_ctrl_is_suspended(ctx->ctrl) &&
        !attr->force) {
        SPDK_ERRLOG("emulation %s cannot be closed because virtio driver is "
                    "up\n", attr->name);
        pthread_mutex_unlock(&spdk_emu_list_lock);
        err_msg = "Emulation cannot be deleted";
        err_code = SPDK_JSONRPC_ERROR_INVALID_REQUEST;
        goto free_attr_strings;
    }

    if(ctx->spci && ctx->spci->parent) {
        struct spdk_emu_ctx *pf_emu_ctx = spdk_emu_ctx_find_by_pci_id(ctx->emu_manager,
                ctx->spci->parent->id, -1, SPDK_EMU_PROTOCOL_VIRTIO_BLK);
        struct snap_virtio_blk_ctrl *pf_ctrl = ((struct virtio_blk_ctrl*)pf_emu_ctx->ctrl)->sctrl;

        if (pf_ctrl &&  pf_ctrl->vfs_ctrl) {
            nvmx_info("Disable VF #%d for PF #%d", ctx->spci->id, ctx->spci->parent->id);
            pf_ctrl->vfs_ctrl[ctx->spci->id] = NULL;
        }
    }

    LIST_REMOVE(ctx, entry);
    pthread_mutex_unlock(&spdk_emu_list_lock);
    free(attr->name);
    free(attr);

    spdk_emu_ctx_destroy(ctx, spdk_emu_ctx_fini_cb_rpc, request);
    return;

free_attr_strings:
    free(attr->name);
    free(attr->vuid);
free_attr:
    free(attr);
invalid:
    if (spdk_emu_rpc_finished) {
        nvmx_warn("spdk rpc is closed. Not returning error response for rpc request");
    } else {
        spdk_jsonrpc_send_error_response(request, err_code, err_msg);
    }
}

SPDK_RPC_REGISTER("controller_virtio_blk_delete",
                  spdk_emu_rpc_controller_vblk_delete, SPDK_RPC_RUNTIME)

static const struct spdk_json_object_decoder
spdk_emu_rpc_controller_vfs_create_decoder[] = {
    {
        "emulation_manager",
        offsetof(struct spdk_emu_controller_vfs_create_attr, emu_manager),
        spdk_json_decode_string
    },
    {
        "pci_bdf",
        offsetof(struct spdk_emu_controller_vfs_create_attr, pci_bdf),
        spdk_json_decode_string,
        true
    },
    {
        "pf_id",
        offsetof(struct spdk_emu_controller_vfs_create_attr, pf_id),
        spdk_json_decode_int32,
        true
    },
    {
        "vf_id",
        offsetof(struct spdk_emu_controller_vfs_create_attr, vf_id),
        spdk_json_decode_int32,
        true
    },
    {
        "bdev_type",
        offsetof(struct spdk_emu_controller_vfs_create_attr, dev_type),
        spdk_json_decode_string
    },
    {
        "num_queues",
        offsetof(struct spdk_emu_controller_vfs_create_attr, num_queues),
        spdk_json_decode_uint32,
        true
    },
    {
        "queue_depth",
        offsetof(struct spdk_emu_controller_vfs_create_attr, queue_depth),
        spdk_json_decode_uint32,
        true
    },
    {
        "tag",
        offsetof(struct spdk_emu_controller_vfs_create_attr, tag),
        spdk_json_decode_string
    },
    {
        "force_in_order",
        offsetof(struct spdk_emu_controller_vfs_create_attr, force_in_order),
        spdk_json_decode_bool,
        true
    },
    {
        "suspend",
        offsetof(struct spdk_emu_controller_vfs_create_attr, suspended),
        spdk_json_decode_bool,
        true
    },
    {   "recover",
        offsetof(struct spdk_emu_controller_vfs_create_attr, recover),
        spdk_json_decode_bool,
        true
    }
};

static void
spdk_emu_rpc_controller_vfs_create(struct spdk_jsonrpc_request *request,
                                    const struct spdk_json_val *params)
{
    struct spdk_emu_controller_vfs_create_attr *attr = NULL;
    struct spdk_emu_ctx_create_attr emu_attr = {};
    struct spdk_emu_ctx *ctx;
    struct spdk_json_write_ctx *w;
    struct snap_pci *spci;

    attr = calloc(1, sizeof(*attr));
    if (!attr)
        goto invalid;

    // Set invalid index, to identify when value was not decoded
    attr->pf_id = -1;
    attr->vf_id = -1;

    if (spdk_json_decode_object(params,
            spdk_emu_rpc_controller_vfs_create_decoder,
            SPDK_COUNTOF(spdk_emu_rpc_controller_vfs_create_decoder),
            attr)) {
        SPDK_ERRLOG("Failed to decode parameters\n");
        goto free_attr;
    }

    spci = mlnx_snap_get_snap_pci(SNAP_VIRTIO_FS, attr->emu_manager,
                                  attr->pci_bdf, attr->pf_id, attr->vf_id);
    if (!spci)
        goto free_attr_strings;

    if (strncmp(attr->dev_type, "none", 4)) {
        SPDK_ERRLOG("Dev type '%s' is not supported\n", attr->dev_type);
        goto free_attr_strings;
    }

    pthread_mutex_lock(&spdk_emu_list_lock);

    if (attr->pci_bdf) {
        ctx = spdk_emu_ctx_find_by_pci_bdf(attr->emu_manager,
                           spci->pci_number);
    } else {
        ctx = spdk_emu_ctx_find_by_pci_id(attr->emu_manager,
                           attr->pf_id, attr->vf_id,
                           SPDK_EMU_PROTOCOL_VIRTIO_FS);
    }

    if (ctx) {
        pthread_mutex_unlock(&spdk_emu_list_lock);
        SPDK_ERRLOG("PCI function is already in use by %s\n", ctx->emu_name);
        goto free_attr_strings;
    }

    emu_attr.priv = attr;
    emu_attr.protocol = SPDK_EMU_PROTOCOL_VIRTIO_FS;
    emu_attr.emu_manager = attr->emu_manager;
    emu_attr.spci = spci;
    ctx = spdk_emu_ctx_create(&emu_attr);
    if (!ctx) {
        pthread_mutex_unlock(&spdk_emu_list_lock);
        goto free_attr_strings;
    }

    LIST_INSERT_HEAD(&spdk_emu_list, ctx, entry);
    pthread_mutex_unlock(&spdk_emu_list_lock);

    w = spdk_jsonrpc_begin_result(request);
    spdk_json_write_string(w, ctx->emu_name);
    spdk_jsonrpc_end_result(request, w);

    free(attr->tag);
    free(attr->dev_type);
    free(attr->pci_bdf);
    free(attr->emu_manager);
    free(attr);
    return;

free_attr_strings:
    free(attr->tag);
    free(attr->dev_type);
    free(attr->pci_bdf);
    free(attr->emu_manager);
free_attr:
    free(attr);
invalid:
    spdk_jsonrpc_send_error_response(request,
                     SPDK_JSONRPC_ERROR_INVALID_PARAMS,
                     "Invalid parameters");
}

SPDK_RPC_REGISTER("controller_virtio_fs_create",
                  spdk_emu_rpc_controller_vfs_create, SPDK_RPC_RUNTIME)

static const struct spdk_json_object_decoder
spdk_emu_rpc_controller_vfs_delete_decoder[] = {
    {
        "name",
        offsetof(struct spdk_emu_controller_vfs_delete_attr, name),
        spdk_json_decode_string
    },
    {
        "force",
        offsetof(struct spdk_emu_controller_vfs_delete_attr, force),
        spdk_json_decode_bool,
        true
    }
};

static void
spdk_emu_rpc_controller_vfs_delete(struct spdk_jsonrpc_request *request,
                                    const struct spdk_json_val *params)
{
    struct spdk_emu_controller_vfs_delete_attr *attr = NULL;
    struct spdk_emu_ctx *ctx;
    const char *err_msg = "Invalid parameters";
    int err_code = SPDK_JSONRPC_ERROR_INVALID_PARAMS;

    attr = calloc(1, sizeof(*attr));
    if (!attr)
        goto invalid;

    if (spdk_json_decode_object(params,
            spdk_emu_rpc_controller_vfs_delete_decoder,
            SPDK_COUNTOF(spdk_emu_rpc_controller_vfs_delete_decoder),
            attr)) {
        SPDK_ERRLOG("Failed to decode parameters\n");
        goto free_attr;
    }

    /* Find SPDK context matching the requested PF id */
    pthread_mutex_lock(&spdk_emu_list_lock);
    ctx = spdk_emu_ctx_find_by_emu_name(attr->name);
    if (!ctx) {
        SPDK_ERRLOG("emulation %s cannot be found\n", attr->name);
        pthread_mutex_unlock(&spdk_emu_list_lock);
        goto free_attr_strings;
    }

    if (ctx->protocol != SPDK_EMU_PROTOCOL_VIRTIO_FS) {
        SPDK_ERRLOG("Given controller name is of wrong type (%s)\n",
                    spdk_emu_protocol_to_string(ctx->protocol));
        pthread_mutex_unlock(&spdk_emu_list_lock);
        goto free_attr_strings;
    }

    if (ctx->spci->plugged && !virtio_fs_ctrl_is_suspended(ctx->ctrl) &&
        !attr->force) {
        SPDK_ERRLOG("emulation %s cannot be closed because virtio driver is "
                    "up\n", attr->name);
        pthread_mutex_unlock(&spdk_emu_list_lock);
        err_msg = "Emulation cannot be deleted";
        err_code = SPDK_JSONRPC_ERROR_INVALID_REQUEST;
        goto free_attr_strings;
    }

    LIST_REMOVE(ctx, entry);
    pthread_mutex_unlock(&spdk_emu_list_lock);
    free(attr->name);
    free(attr);

    spdk_emu_ctx_destroy(ctx, spdk_emu_ctx_fini_cb_rpc, request);
    return;

free_attr_strings:
    free(attr->name);
free_attr:
    free(attr);
invalid:
    spdk_jsonrpc_send_error_response(request, err_code, err_msg);
}

SPDK_RPC_REGISTER("controller_virtio_fs_delete",
                  spdk_emu_rpc_controller_vfs_delete, SPDK_RPC_RUNTIME)

struct spdk_emu_controller_vblk_lm_attr {
	char *name;
	char *lm_channel_name;
};

static const struct spdk_json_object_decoder
spdk_emu_rpc_controller_vblk_lm_enable_decoder[] = {
    {
        "name",
        offsetof(struct spdk_emu_controller_vblk_lm_attr, name),
        spdk_json_decode_string
    },
    {
        "lm_channel_name",
        offsetof(struct spdk_emu_controller_vblk_lm_attr, lm_channel_name),
        spdk_json_decode_string
    }
};

static const struct spdk_json_object_decoder
spdk_emu_rpc_controller_vblk_lm_disable_decoder[] = {
    {
        "name",
        offsetof(struct spdk_emu_controller_vblk_lm_attr, name),
        spdk_json_decode_string
    },
};

static void
spdk_emu_rpc_controller_vblk_lm_enable(struct spdk_jsonrpc_request *request,
                                       const struct spdk_json_val *params)
{
    struct spdk_emu_controller_vblk_lm_attr *attr = NULL;
    struct spdk_emu_ctx *ctx;
    const char *err_msg = "Invalid parameters";
    int err_code = SPDK_JSONRPC_ERROR_INVALID_PARAMS;
    struct spdk_json_write_ctx *w;
    int ret;

    attr = calloc(1, sizeof(*attr));
    if (!attr)
        goto invalid;

    if (spdk_json_decode_object(params,
            spdk_emu_rpc_controller_vblk_lm_enable_decoder,
            SPDK_COUNTOF(spdk_emu_rpc_controller_vblk_lm_enable_decoder),
            attr)) {
        SPDK_ERRLOG("Failed to decode parameters\n");
        goto free_attr;
    }

    pthread_mutex_lock(&spdk_emu_list_lock);
    ctx = spdk_emu_ctx_find_by_emu_name(attr->name);
    if (!ctx) {
        SPDK_ERRLOG("emulation %s cannot be found\n", attr->name);
        pthread_mutex_unlock(&spdk_emu_list_lock);
        goto free_attr_strings;
    }

    if (ctx->protocol != SPDK_EMU_PROTOCOL_VIRTIO_BLK) {
        SPDK_ERRLOG("Given controller name is of wrong type (%s)\n",
                    spdk_emu_protocol_to_string(ctx->protocol));
        pthread_mutex_unlock(&spdk_emu_list_lock);
        goto free_attr_strings;
    }

    ret = virtio_blk_ctrl_lm_enable(ctx->ctrl, attr->lm_channel_name);

    pthread_mutex_unlock(&spdk_emu_list_lock);

    if (ret)
        goto free_attr_strings;

    nvmx_info("Enabled live migration channel %s on %s", attr->lm_channel_name, attr->name);

    w = spdk_jsonrpc_begin_result(request);
    spdk_json_write_bool(w, true);
    spdk_jsonrpc_end_result(request, w);

    free(attr->name);
    free(attr->lm_channel_name);
    free(attr);

    return;

free_attr_strings:
    free(attr->name);
    free(attr->lm_channel_name);
free_attr:
    free(attr);
invalid:
    spdk_jsonrpc_send_error_response(request, err_code, err_msg);
}

SPDK_RPC_REGISTER("controller_virtio_blk_lm_enable",
                  spdk_emu_rpc_controller_vblk_lm_enable, SPDK_RPC_RUNTIME)

static void
spdk_emu_rpc_controller_vblk_lm_disable(struct spdk_jsonrpc_request *request,
                                        const struct spdk_json_val *params)
{
    struct spdk_emu_controller_vblk_lm_attr *attr = NULL;
    struct spdk_emu_ctx *ctx;
    const char *err_msg = "Invalid parameters";
    int err_code = SPDK_JSONRPC_ERROR_INVALID_PARAMS;
    struct spdk_json_write_ctx *w;

    attr = calloc(1, sizeof(*attr));
    if (!attr)
        goto invalid;

    if (spdk_json_decode_object(params,
            spdk_emu_rpc_controller_vblk_lm_disable_decoder,
            SPDK_COUNTOF(spdk_emu_rpc_controller_vblk_lm_disable_decoder),
            attr)) {
        SPDK_ERRLOG("Failed to decode parameters\n");
        goto free_attr;
    }

    pthread_mutex_lock(&spdk_emu_list_lock);
    ctx = spdk_emu_ctx_find_by_emu_name(attr->name);
    if (!ctx) {
        SPDK_ERRLOG("emulation %s cannot be found\n", attr->name);
        pthread_mutex_unlock(&spdk_emu_list_lock);
        goto free_attr_strings;
    }

    if (ctx->protocol != SPDK_EMU_PROTOCOL_VIRTIO_BLK) {
        SPDK_ERRLOG("Given controller name is of wrong type (%s)\n",
                    spdk_emu_protocol_to_string(ctx->protocol));
        pthread_mutex_unlock(&spdk_emu_list_lock);
        goto free_attr_strings;
    }

    virtio_blk_ctrl_lm_disable(ctx->ctrl);
    nvmx_info("Disabled live migration on %s", attr->name);

    pthread_mutex_unlock(&spdk_emu_list_lock);

    w = spdk_jsonrpc_begin_result(request);
    spdk_json_write_bool(w, true);
    spdk_jsonrpc_end_result(request, w);

    free(attr->name);
    free(attr);

    return;

free_attr_strings:
    free(attr->name);
free_attr:
    free(attr);
invalid:
    spdk_jsonrpc_send_error_response(request, err_code, err_msg);
}

SPDK_RPC_REGISTER("controller_virtio_blk_lm_disable",
                  spdk_emu_rpc_controller_vblk_lm_disable, SPDK_RPC_RUNTIME)

struct spdk_emu_controller_vblk_suspend_attr {
	char *name;
	uint32_t timeout;
};

static const struct spdk_json_object_decoder
spdk_emu_rpc_controller_vblk_suspend_decoder[] = {
    {
        "name",
        offsetof(struct spdk_emu_controller_vblk_suspend_attr, name),
        spdk_json_decode_string
    },
    {
        "timeout",
        offsetof(struct spdk_emu_controller_vblk_suspend_attr, timeout),
        spdk_json_decode_uint32,
	true
    }
};

struct controller_virtio_blk_suspend_ctx {
    struct spdk_emu_ctx *ctrl_ctx;
    struct spdk_jsonrpc_request *request;
    uint64_t start_time;
    uint64_t timeout_sec;
};

static void spdk_emu_rpc_controller_vblk_suspend_end(void *arg)
{
    struct controller_virtio_blk_suspend_ctx *ctx = arg;
    struct spdk_emu_ctx *ctrl_ctx = ctx->ctrl_ctx;
    struct spdk_jsonrpc_request *request = ctx->request;
    struct spdk_json_write_ctx *w;
    const uint64_t timeout_ticks = ctx->start_time +
                                   ctx->timeout_sec * spdk_get_ticks_hz();

    if (ctrl_ctx->ctrl_ops->is_suspended(ctrl_ctx->ctrl)) {
        w = spdk_jsonrpc_begin_result(request);
        spdk_json_write_bool(w, true);
        spdk_jsonrpc_end_result(request, w);
    } else if (spdk_get_ticks() < timeout_ticks) {
        spdk_thread_send_msg(spdk_get_thread(),
                             spdk_emu_rpc_controller_vblk_suspend_end, ctx);
        return;
    } else {
        spdk_jsonrpc_send_error_response(request,
                                         SPDK_JSONRPC_ERROR_INTERNAL_ERROR,
                                         "Request Timeout");
    }

    free(ctx);
}

static void
spdk_emu_rpc_controller_vblk_suspend(struct spdk_jsonrpc_request *request,
                                     const struct spdk_json_val *params)
{
    struct spdk_emu_controller_vblk_suspend_attr *attr = NULL;
    struct controller_virtio_blk_suspend_ctx *suspend_ctx;
    struct spdk_emu_ctx *ctx;
    const char *err_msg = "Invalid parameters";
    int err_code = SPDK_JSONRPC_ERROR_INVALID_PARAMS;

    attr = calloc(1, sizeof(*attr));
    if (!attr)
        goto invalid;

    attr->timeout = VIRTIO_BLK_CTRL_DEFAULT_SUSPEND_TMO;

    if (spdk_json_decode_object(params,
            spdk_emu_rpc_controller_vblk_suspend_decoder,
            SPDK_COUNTOF(spdk_emu_rpc_controller_vblk_suspend_decoder),
            attr)) {
        SPDK_ERRLOG("Failed to decode parameters\n");
        goto free_attr;
    }

    pthread_mutex_lock(&spdk_emu_list_lock);
    ctx = spdk_emu_ctx_find_by_emu_name(attr->name);
    if (!ctx) {
        SPDK_ERRLOG("emulation %s cannot be found\n", attr->name);
        pthread_mutex_unlock(&spdk_emu_list_lock);
        goto free_attr_strings;
    }

    if (ctx->protocol != SPDK_EMU_PROTOCOL_VIRTIO_BLK) {
        SPDK_ERRLOG("Given controller name is of wrong type (%s)\n",
                    spdk_emu_protocol_to_string(ctx->protocol));
        pthread_mutex_unlock(&spdk_emu_list_lock);
        goto free_attr_strings;
    }

    suspend_ctx = calloc(1, sizeof(*suspend_ctx));
    if (!suspend_ctx) {
        SPDK_ERRLOG("Failed to allocate suspend_ctx\n");
        pthread_mutex_unlock(&spdk_emu_list_lock);
        goto free_attr_strings;
    }

    suspend_ctx->ctrl_ctx = ctx;
    suspend_ctx->request = request;
    suspend_ctx->start_time = spdk_get_ticks();
    suspend_ctx->timeout_sec = attr->timeout;
    ctx->ctrl_ops->suspend(ctx->ctrl);

    pthread_mutex_unlock(&spdk_emu_list_lock);

    spdk_thread_send_msg(spdk_get_thread(),
                         spdk_emu_rpc_controller_vblk_suspend_end,
                         suspend_ctx);
    free(attr->name);
    free(attr);

    return;

free_attr_strings:
    free(attr->name);
free_attr:
    free(attr);
invalid:
    spdk_jsonrpc_send_error_response(request, err_code, err_msg);
}

SPDK_RPC_REGISTER("controller_virtio_blk_suspend",
                  spdk_emu_rpc_controller_vblk_suspend, SPDK_RPC_RUNTIME)

struct spdk_emu_controller_vblk_resume_attr {
	char *name;
};

static const struct spdk_json_object_decoder
spdk_emu_rpc_controller_vblk_resume_decoder[] = {
    {
        "name",
        offsetof(struct spdk_emu_controller_vblk_resume_attr, name),
        spdk_json_decode_string
    }
};

static void
spdk_emu_rpc_controller_vblk_resume(struct spdk_jsonrpc_request *request,
                                     const struct spdk_json_val *params)
{
    struct spdk_emu_controller_vblk_resume_attr *attr = NULL;
    struct spdk_emu_ctx *ctx;
    const char *err_msg = "Invalid parameters";
    int err_code = SPDK_JSONRPC_ERROR_INVALID_PARAMS;
    struct spdk_json_write_ctx *w;

    attr = calloc(1, sizeof(*attr));
    if (!attr)
        goto invalid;

    if (spdk_json_decode_object(params,
            spdk_emu_rpc_controller_vblk_resume_decoder,
            SPDK_COUNTOF(spdk_emu_rpc_controller_vblk_resume_decoder),
            attr)) {
        SPDK_ERRLOG("Failed to decode parameters\n");
        goto free_attr;
    }

    pthread_mutex_lock(&spdk_emu_list_lock);
    ctx = spdk_emu_ctx_find_by_emu_name(attr->name);
    if (!ctx) {
        SPDK_ERRLOG("emulation %s cannot be found\n", attr->name);
        pthread_mutex_unlock(&spdk_emu_list_lock);
        goto free_attr_strings;
    }

    if (ctx->protocol != SPDK_EMU_PROTOCOL_VIRTIO_BLK) {
        SPDK_ERRLOG("Given controller name is of wrong type (%s)\n",
                    spdk_emu_protocol_to_string(ctx->protocol));
        pthread_mutex_unlock(&spdk_emu_list_lock);
        goto free_attr_strings;
    }

    ctx->ctrl_ops->resume(ctx->ctrl);

    pthread_mutex_unlock(&spdk_emu_list_lock);

    w = spdk_jsonrpc_begin_result(request);
    spdk_json_write_bool(w, true);
    spdk_jsonrpc_end_result(request, w);

    free(attr->name);
    free(attr);

    return;

free_attr_strings:
    free(attr->name);
free_attr:
    free(attr);
invalid:
    spdk_jsonrpc_send_error_response(request, err_code, err_msg);
}

SPDK_RPC_REGISTER("controller_virtio_blk_resume",
                  spdk_emu_rpc_controller_vblk_resume, SPDK_RPC_RUNTIME)

struct spdk_emu_controller_vblk_state_attr {
    char *name;
    char *save_file;
    char *restore_file;
};

static const struct spdk_json_object_decoder
spdk_emu_rpc_controller_vblk_state_decoder[] = {
    {
        "name",
        offsetof(struct spdk_emu_controller_vblk_state_attr, name),
        spdk_json_decode_string
    },
    {
        "save",
        offsetof(struct spdk_emu_controller_vblk_state_attr, save_file),
        spdk_json_decode_string,
        true
    },
    {
        "restore",
        offsetof(struct spdk_emu_controller_vblk_state_attr, restore_file),
        spdk_json_decode_string,
        true
    }
};

static void
spdk_emu_rpc_controller_vblk_state(struct spdk_jsonrpc_request *request,
                                   const struct spdk_json_val *params)
{
    const char *err_msg = "Invalid parameters";
    int err_code = SPDK_JSONRPC_ERROR_INVALID_PARAMS;
    struct spdk_emu_controller_vblk_state_attr *attr = NULL;
    struct spdk_emu_ctx *ctx;
    struct spdk_json_write_ctx *w;
    int ret;

    attr = calloc(1, sizeof(*attr));
    if (!attr)
        goto invalid;

    if (spdk_json_decode_object(params,
            spdk_emu_rpc_controller_vblk_state_decoder,
            SPDK_COUNTOF(spdk_emu_rpc_controller_vblk_state_decoder),
            attr)) {
        SPDK_ERRLOG("Failed to decode parameters\n");
        goto free_attr;
    }

    pthread_mutex_lock(&spdk_emu_list_lock);
    ctx = spdk_emu_ctx_find_by_emu_name(attr->name);
    if (!ctx) {
        SPDK_ERRLOG("emulation %s cannot be found\n", attr->name);
        pthread_mutex_unlock(&spdk_emu_list_lock);
        goto free_attr_strings;
    }

    if (ctx->protocol != SPDK_EMU_PROTOCOL_VIRTIO_BLK) {
        SPDK_ERRLOG("Given controller name is of wrong type (%s)\n",
                    spdk_emu_protocol_to_string(ctx->protocol));
        pthread_mutex_unlock(&spdk_emu_list_lock);
        goto free_attr_strings;
    }

    if (attr->save_file) {
        ret = virtio_blk_ctrl_state_save(ctx->ctrl, attr->save_file);
    } else if (attr->restore_file) {
        ret = virtio_blk_ctrl_state_restore(ctx->ctrl, attr->restore_file);
    } else {
        /* TODO: dump state info */
        ret = 0;
    }

    if (ret) {
        pthread_mutex_unlock(&spdk_emu_list_lock);
        goto free_attr_strings;
    }

    w = spdk_jsonrpc_begin_result(request);
    spdk_json_write_bool(w, true);
    spdk_jsonrpc_end_result(request, w);

    pthread_mutex_unlock(&spdk_emu_list_lock);

    free(attr->save_file);
    free(attr->restore_file);
    free(attr->name);
    free(attr);
    return;

free_attr_strings:
    free(attr->save_file);
    free(attr->restore_file);
    free(attr->name);
free_attr:
    free(attr);
invalid:
    spdk_jsonrpc_send_error_response(request, err_code, err_msg);
}

SPDK_RPC_REGISTER("controller_virtio_blk_state",
                  spdk_emu_rpc_controller_vblk_state, SPDK_RPC_RUNTIME)

static const struct spdk_json_object_decoder
spdk_emu_rpc_controller_vblk_get_iostat_decoder[] = {
    {
        "name",
        offsetof(struct spdk_emu_controller_vblk_get_iostat_attr, name),
        spdk_json_decode_string,
        true
    }
};

static void
spdk_emu_rpc_controller_vblk_get_iostat(struct spdk_jsonrpc_request *request,
                                    const struct spdk_json_val *params)
{
    struct spdk_emu_controller_vblk_get_iostat_attr *attr;
    struct spdk_emu_ctx *ctx;
    struct spdk_json_write_ctx *w;
    struct virtio_blk_ctrl *ctrl;
    struct virtio_blk_ctrl_io_stat stat;

    attr = calloc(1, sizeof(*attr));
    if (!attr)
        goto invalid;

    if (params && spdk_json_decode_object(params,
            spdk_emu_rpc_controller_vblk_get_iostat_decoder,
            SPDK_COUNTOF(spdk_emu_rpc_controller_vblk_get_iostat_decoder),
            attr)) {
        SPDK_ERRLOG("Failed to decode parameters\n");
        goto free_attr;
    }

    w = spdk_jsonrpc_begin_result(request);
    spdk_json_write_object_begin(w);
    spdk_json_write_named_array_begin(w, "controllers");

    pthread_mutex_lock(&spdk_emu_list_lock);
    LIST_FOREACH(ctx, &spdk_emu_list, entry) {

        if (ctx->protocol != SPDK_EMU_PROTOCOL_VIRTIO_BLK)
            continue;

        if (attr->name && *attr->name && strncmp(ctx->emu_name, attr->name, SPDK_EMU_NAME_MAXLEN))
            continue;

        ctrl = ctx->ctrl;

        spdk_json_write_object_begin(w);
        spdk_json_write_named_string(w, "name", ctx->emu_name);
        spdk_json_write_named_array_begin(w, "bdevs");

        if (ctrl->bctx.bdev) {
            spdk_json_write_object_begin(w);

            spdk_json_write_named_string(w, "bdev_name", spdk_bdev_get_name(ctrl->bctx.bdev));

            virtio_blk_ctrl_get_stat(ctrl, &stat);

            spdk_json_write_named_uint64(w, "read_ios", stat.read_ios);
            spdk_json_write_named_uint64(w, "completed_read_ios", stat.completed_read_ios);
            spdk_json_write_named_uint64(w, "completed_unordered_read_ios", stat.completed_unord_r_ios);
            spdk_json_write_named_uint64(w, "write_ios", stat.write_ios);
            spdk_json_write_named_uint64(w, "completed_write_ios", stat.completed_write_ios);
            spdk_json_write_named_uint64(w, "completed_unordered_write_ios", stat.completed_unord_w_ios);
            spdk_json_write_named_uint64(w, "flush_ios", stat.flush_ios);
            spdk_json_write_named_uint64(w, "completed_flush_ios", stat.completed_flush_ios);
            spdk_json_write_named_uint64(w, "completed_unordered_flush_ios", stat.completed_unord_f_ios);
            spdk_json_write_named_uint64(w, "err_read_ios", stat.err_read_os);
            spdk_json_write_named_uint64(w, "err_write_ios", stat.err_write_os);
            spdk_json_write_named_uint64(w, "err_flush_ios", stat.err_flush_os);
            spdk_json_write_named_uint64(w, "outstanding_in_ios", stat.outstand_in_ios);
            spdk_json_write_named_uint64(w, "outstanding_in_bdev_ios", stat.outstand_in_bdev_ios);
            spdk_json_write_named_uint64(w, "outstanding_to_host_ios", stat.outstand_to_host_ios);
            spdk_json_write_named_uint64(w, "fatal_ios", stat.fatal_ios);

            spdk_json_write_object_end(w);
        }

        spdk_json_write_array_end(w);

        spdk_json_write_object_end(w);
    }
    pthread_mutex_unlock(&spdk_emu_list_lock);

    spdk_json_write_array_end(w);
    spdk_json_write_object_end(w);
    spdk_jsonrpc_end_result(request, w);

    free(attr->name);
    free(attr);

    return;

free_attr:
    free(attr);
invalid:
    spdk_jsonrpc_send_error_response(request,
                     SPDK_JSONRPC_ERROR_INVALID_PARAMS,
                     "Invalid parameters");
}

SPDK_RPC_REGISTER("controller_virtio_blk_get_iostat",
        spdk_emu_rpc_controller_vblk_get_iostat, SPDK_RPC_RUNTIME)

static void
spdk_emu_mlnx_snap_params_list(struct spdk_jsonrpc_request *request,
                               const struct spdk_json_val *params)
{
    struct spdk_json_write_ctx *w;

    if (params) {
        spdk_jsonrpc_send_error_response(request,
                    SPDK_JSONRPC_ERROR_INVALID_PARAMS,
                    "No parameters required");
        return;
    }

    w = spdk_jsonrpc_begin_result(request);
    spdk_json_write_array_begin(w);

    char buf[64] = {0};
    snap_env_iter_t iter = NULL;
    do {
        iter = snap_env_dump_env_entry(iter, buf, sizeof(buf));
        spdk_json_write_string(w, buf);
    } while (iter);

    spdk_json_write_array_end(w);
    spdk_jsonrpc_end_result(request, w);
  }

SPDK_RPC_REGISTER("mlnx_snap_params_list",
        spdk_emu_mlnx_snap_params_list, SPDK_RPC_RUNTIME)


struct spdk_emu_device_attach_pci_regs {
    uint16_t id;
    uint16_t vid;
    uint16_t revid;
    uint32_t class_code;
    uint16_t ssid;
    uint16_t ssvid;
    uint16_t num_msix;
};

struct spdk_emu_device_attach_virtio_regs {
    char *bdev;
    char *bdev_type;
    uint32_t num_queues;
    bool transitional_device;
};

typedef struct spdk_emu_device_attach_virtio_regs
                spdk_emu_device_attach_virtio_blk_regs;
typedef struct spdk_emu_device_attach_virtio_regs
                spdk_emu_device_attach_virtio_fs_regs;

struct spdk_emu_device_attach_nvme_regs {
    char *version;
};

struct spdk_emu_device_attach_attr {
    char *emu_manager;
    char *device_type;
    uint16_t total_vf;
    uint32_t queue_depth;
    struct spdk_emu_device_attach_pci_regs pci;
    union {
        spdk_emu_device_attach_virtio_blk_regs virtio_blk;
        spdk_emu_device_attach_virtio_fs_regs virtio_fs;
        struct spdk_emu_device_attach_nvme_regs nvme;
    };
};

static void
spdk_emu_hotplug_init_pci(struct snap_hotplug_attr *hp_attr,
                          struct spdk_emu_device_attach_attr *attr)
{
    hp_attr->pci_attr.revision_id = attr->pci.revid;
    hp_attr->pci_attr.num_msix = attr->pci.num_msix ? :
                                 SNAP_PCI_NUM_MSIX_DEFAULT;
    switch (hp_attr->type) {
    case SNAP_NVME:
        hp_attr->pci_attr.class_code = attr->pci.class_code ? :
                                       SNAP_PCI_CLASS_CODE_NVME_CTRL;
        hp_attr->pci_attr.device_id = attr->pci.id ? :
                                      SNAP_PCI_DEVICE_ID_BLUEFIELD_NVME;
        hp_attr->pci_attr.vendor_id = attr->pci.vid ? :
                                      SNAP_PCI_VENDOR_ID_MELLANOX;
        hp_attr->pci_attr.subsystem_id = attr->pci.ssid;
        hp_attr->pci_attr.subsystem_vendor_id = attr->pci.ssvid;
        break;
    case SNAP_VIRTIO_BLK:
        hp_attr->pci_attr.class_code = attr->pci.class_code ? :
                                       SNAP_PCI_CLASS_CODE_MASS_STORAGE_CTRL;
        if (attr->virtio_blk.transitional_device) {
            hp_attr->pci_attr.device_id = SNAP_PCI_DEVICE_ID_VIRTIO_BLK_TRANSITIONAL;
            hp_attr->pci_attr.revision_id = SNAP_PCI_REVISION_ID_VIRTIO_BLK_TRANSITIONAL;
            hp_attr->pci_attr.subsystem_id = SNAP_VIRTIO_BLK_VIRTIO_DEVICE_ID;
        } else {
            hp_attr->pci_attr.device_id = attr->pci.id ? :
                                      SNAP_PCI_DEVICE_ID_VIRTIO_BLK;
            hp_attr->pci_attr.revision_id = SNAP_PCI_REVISION_ID_VIRTIO_BLK;
            hp_attr->pci_attr.subsystem_id = attr->pci.ssid ? :
                                         SNAP_PCI_DEVICE_ID_BLUEFIELD_VIRTIO_BLK;
        }
        hp_attr->pci_attr.vendor_id = SNAP_PCI_VENDOR_ID_REDHAT; // Mandatory by spec.
        hp_attr->pci_attr.subsystem_vendor_id = attr->pci.ssvid ? :
                                                SNAP_PCI_VENDOR_ID_MELLANOX;
        break;
    case SNAP_VIRTIO_FS:
        hp_attr->pci_attr.class_code = attr->pci.class_code ? :
                                       SNAP_PCI_CLASS_CODE_MASS_STORAGE_CTRL;
        hp_attr->pci_attr.device_id = attr->pci.id ? :
                                      SNAP_PCI_DEVICE_ID_VIRTIO_FS;
        hp_attr->pci_attr.vendor_id = SNAP_PCI_VENDOR_ID_REDHAT; // Mandatory by spec.

        /* Subsystem device id 6004 – fs
         * Subsystem vendor id should be Mellanox.
         */
        hp_attr->pci_attr.subsystem_id = attr->pci.ssid ? :
                                         SNAP_PCI_SSID_VIRTIO_FS;
        hp_attr->pci_attr.subsystem_vendor_id = attr->pci.ssvid ? :
                                                SNAP_PCI_SSVID_VIRTIO_FS;
        break;
    default:
        SPDK_WARNLOG("Invalid device type %d\n", hp_attr->type);
        return;
    };
    nvmx_debug("Hotplug PCI attributes: id 0x%04x vid 0x%04x ssid 0x%04x "
               "ssvid 0x%04x revid 0x%04x class_code 0x%06x num_msix %d",
               hp_attr->pci_attr.device_id, hp_attr->pci_attr.vendor_id,
               hp_attr->pci_attr.subsystem_id,
               hp_attr->pci_attr.subsystem_vendor_id,
               hp_attr->pci_attr.revision_id, hp_attr->pci_attr.class_code,
               hp_attr->pci_attr.num_msix);
}

static int
spdk_emu_hotplug_init_nvme(struct snap_hotplug_attr *hp_attr,
                           struct spdk_emu_device_attach_attr *init_attr,
                           const struct snap_nvme_caps *caps)
{
    unsigned int mjr = 1, mnr = 3, ter = 0;

    hp_attr->use_default_regs = false;
    hp_attr->regs.nvme.cap.bits.mpsmax = 4;
    hp_attr->regs.nvme.cap.bits.mpsmin = 0;
    hp_attr->regs.nvme.cap.bits.css = 1;
    hp_attr->regs.nvme.cap.bits.to = 0xFF;
    hp_attr->regs.nvme.cap.bits.cqr = 1;
    if (init_attr->queue_depth > caps->max_queue_depth) {
        nvmx_error("Requested Queue Depth (%lu) exceeds maximal supported value (%lu)",
                   init_attr->queue_depth, caps->max_queue_depth);
        return -EINVAL;
    }
    hp_attr->regs.nvme.cap.bits.mqes = init_attr->queue_depth ? :
                                       NVME_REGS_DEFAULT_QUEUE_SIZE;
    hp_attr->regs.nvme.cap.bits.mqes--; // zero-based value

    if (init_attr->nvme.version) {
        if (sscanf(init_attr->nvme.version, "%u.%u.%u",
                   &mjr, &mnr, &ter) != 3) {
            nvmx_error("Invalid version number \'%s\'",
                       init_attr->nvme.version);
            return -EINVAL;
        }
    }

    hp_attr->regs.nvme.vs.bits.mjr = mjr;
    hp_attr->regs.nvme.vs.bits.mnr = mnr;
    hp_attr->regs.nvme.vs.bits.ter = ter;
    return 0;
}

static int
spdk_emu_hotplug_init_virtio_blk(struct snap_hotplug_attr *hp_attr,
                                 struct spdk_emu_device_attach_attr *init_attr,
                                 const struct snap_virtio_caps *caps)
{
    struct spdk_bdev *bdev;
    struct virtio_blk_regs_set_attr regs_attr = {};

    regs_attr.max_queues = nvmx_min(caps->max_emulated_virtqs,
                                    hp_attr->pci_attr.num_msix - 1);
    if (init_attr->virtio_blk.num_queues) {
        if (init_attr->virtio_blk.num_queues > caps->max_emulated_virtqs) {
            SPDK_ERRLOG("Requested num_queues (%u) exceeds device capacity (%u)\n",
                        init_attr->virtio_blk.num_queues, caps->max_emulated_virtqs);
            return -EINVAL;
        }

        if (init_attr->virtio_blk.num_queues > hp_attr->pci_attr.num_msix - 1) {
            SPDK_WARNLOG("Requested num_queues (%u) exceeds valid MSI-X. "
                         "Performance might be harmed\n",
                         init_attr->virtio_blk.num_queues);
        }
        regs_attr.max_queues = init_attr->virtio_blk.num_queues;
    }

    regs_attr.queue_depth = init_attr->queue_depth ? :
                            VIRTIO_BLK_REGS_DEF_QUEUE_SIZE;
    if (regs_attr.queue_depth > VIRTIO_BLK_REGS_MAX_QUEUE_SIZE) {
        SPDK_ERRLOG("Requested queue_depth (%u) exceeds device capacity (%u)\n",
                    regs_attr.queue_depth, VIRTIO_BLK_REGS_MAX_QUEUE_SIZE);
        return -EINVAL;
    }
    /* Virtio blk needs at least 3 descriptors: header, data and status.*/
    if (regs_attr.queue_depth < 3 ) {
        SPDK_ERRLOG("Requested queue_depth (%u) cannot be less than 3.\n",
                    regs_attr.queue_depth);
        return -EINVAL;
    }

    if (init_attr->virtio_blk.bdev_type) {
        if (strncmp(init_attr->virtio_blk.bdev_type, "spdk", 4) &&
            strncmp(init_attr->virtio_blk.bdev_type, "none", 4)) {
            SPDK_ERRLOG("Bdev type '%s' is not supported\n",
                         init_attr->virtio_blk.bdev_type);
            return -ENOTSUP;
        } else if (!strncmp(init_attr->virtio_blk.bdev_type, "none", 4) &&
                   init_attr->virtio_blk.bdev) {
            SPDK_ERRLOG("Bdev type '%s' requires no \"bdev\" attribute\n",
                        init_attr->virtio_blk.bdev_type);
            return -EINVAL;
        } else if (!strncmp(init_attr->virtio_blk.bdev_type, "spdk", 4) &&
                   !init_attr->virtio_blk.bdev) {
            SPDK_ERRLOG("Bdev type '%s' requires \"bdev\" attribute\n",
                        init_attr->virtio_blk.bdev_type);
            return -EINVAL;
        }
    }

    regs_attr.size_max = VIRTIO_BLK_REGS_DEF_SIZE_MAX;
    regs_attr.seg_max = VIRTIO_BLK_REGS_DEF_SEG_MAX;
    if (init_attr->virtio_blk.bdev) {
        bdev = spdk_bdev_get_by_name(init_attr->virtio_blk.bdev);
        if (bdev) {
            regs_attr.num_blocks = spdk_bdev_get_num_blocks(bdev);
            regs_attr.blk_size = spdk_bdev_get_block_size(bdev);
        }
    }

    /* Set bar registers with hotplug attributes */
    virtio_blk_regs_set(&regs_attr, &hp_attr->regs.virtio_blk);
    /* special non-modifiable feature which can be set only during hotplug */
    hp_attr->regs.virtio_blk.device_features |= 1ULL << VIRTIO_F_IOMMU_PLATFORM;
    if (hp_attr->max_vfs)
        hp_attr->regs.virtio_blk.device_features |= 1ULL << VIRTIO_F_SR_IOV;

    return 0;
}

static int
spdk_emu_hotplug_init_virtio_fs(struct snap_hotplug_attr *hp_attr,
                                struct spdk_emu_device_attach_attr *init_attr,
                                const struct snap_virtio_caps *caps)
{
    struct virtio_fs_regs_set_attr regs_attr = {};

    regs_attr.max_queues = nvmx_min(caps->max_emulated_virtqs,
                                    hp_attr->pci_attr.num_msix - 1);
    if (init_attr->virtio_fs.num_queues) {
        if (init_attr->virtio_fs.num_queues > caps->max_emulated_virtqs) {
            SPDK_ERRLOG("Requested num_queues (%u) exceeds device capacity (%u)\n",
                        init_attr->virtio_fs.num_queues, caps->max_emulated_virtqs);
            return -EINVAL;
        }

        if (init_attr->virtio_blk.num_queues > hp_attr->pci_attr.num_msix - 1) {
            SPDK_WARNLOG("Requested num_queues (%u) exceeds valid MSI-X. "
                         "Performance might be harmed\n",
                         init_attr->virtio_fs.num_queues);
        }
        regs_attr.max_queues = init_attr->virtio_blk.num_queues;
    }

    regs_attr.queue_depth = init_attr->queue_depth ? :
                            VIRTIO_FS_REGS_DEF_QUEUE_SIZE;
    if (regs_attr.queue_depth > VIRTIO_FS_REGS_MAX_QUEUE_SIZE) {
        SPDK_ERRLOG("Requested queue_depth (%u) exceeds device capacity (%u)\n",
                    regs_attr.queue_depth, VIRTIO_FS_REGS_MAX_QUEUE_SIZE);
        return -EINVAL;
    }

    if (init_attr->virtio_fs.bdev_type) {
        if (strncmp(init_attr->virtio_fs.bdev_type, "none", 4)) {
            SPDK_ERRLOG("Dev type '%s' is not supported\n",
                         init_attr->virtio_fs.bdev_type);
            return -ENOTSUP;
        }
    }

    /*
     * 5.11.3 Feature bits
     * There are currently no feature bits defined.
     */

    /* Set bar registers with hotplug attributes */
    virtio_fs_regs_set(&regs_attr, &hp_attr->regs.virtio_fs);
    /* special non-modifiable feature which can be set only during hotplug */

    // TODO - check
    hp_attr->regs.virtio_fs.device_features |= 1ULL << VIRTIO_F_IOMMU_PLATFORM;
    if (hp_attr->max_vfs)
        hp_attr->regs.virtio_fs.device_features |= 1ULL << VIRTIO_F_SR_IOV;

    return 0;
}

static const struct spdk_json_object_decoder
spdk_emu_device_attach_decoder[] = {
    {
        "emulation_manager",
        offsetof(struct spdk_emu_device_attach_attr, emu_manager),
        spdk_json_decode_string
    },
    {
        "device_type",
        offsetof(struct spdk_emu_device_attach_attr, device_type),
        spdk_json_decode_string
    },
    {
        "total_vf",
        offsetof(struct spdk_emu_device_attach_attr, total_vf),
        spdk_json_decode_uint16,
        true
    },
    {
        "id",
        offsetof(struct spdk_emu_device_attach_attr, pci.id),
        spdk_json_decode_uint16,
        true
    },
    {
        "vid",
        offsetof(struct spdk_emu_device_attach_attr, pci.vid),
        spdk_json_decode_uint16,
        true
    },
    {
        "revid",
        offsetof(struct spdk_emu_device_attach_attr, pci.revid),
        spdk_json_decode_uint16,
        true
    },
    {
        "class_code",
        offsetof(struct spdk_emu_device_attach_attr, pci.class_code),
        spdk_json_decode_uint32,
        true
    },
    {
        "ssid",
        offsetof(struct spdk_emu_device_attach_attr, pci.ssid),
        spdk_json_decode_uint16,
        true
    },
    {
        "ssvid",
        offsetof(struct spdk_emu_device_attach_attr, pci.ssvid),
        spdk_json_decode_uint16,
        true
    },
    {
        "num_msix",
        offsetof(struct spdk_emu_device_attach_attr, pci.num_msix),
        spdk_json_decode_uint16,
        true
    },
    {
        "num_queues",
        offsetof(struct spdk_emu_device_attach_attr, virtio_blk.num_queues),
        spdk_json_decode_uint16,
        true
    },
    {
        "queue_depth",
        offsetof(struct spdk_emu_device_attach_attr, queue_depth),
        spdk_json_decode_uint16,
        true
    },
    {
        "bdev_type",
        offsetof(struct spdk_emu_device_attach_attr, virtio_blk.bdev_type),
        spdk_json_decode_string,
        true
    },
    {
        "bdev",
        offsetof(struct spdk_emu_device_attach_attr, virtio_blk.bdev),
        spdk_json_decode_string,
        true
    },
    {
        "transitional_device",
        offsetof(struct spdk_emu_device_attach_attr, virtio_blk.transitional_device),
        spdk_json_decode_bool,
        true
    },
    {
        "version",
        offsetof(struct spdk_emu_device_attach_attr, nvme.version),
        spdk_json_decode_string,
        true
    },
};

static void
spdk_emu_rpc_device_attach(struct spdk_jsonrpc_request *request,
                           const struct spdk_json_val *params)
{
    struct spdk_emu_device_attach_attr *attr = NULL;
    struct spdk_json_write_ctx *w;
    struct snap_context *sctx;
    struct snap_hotplug_attr hp_attr = {};
    struct snap_pci *hotplug;
    int err;

    attr = calloc(1, sizeof(*attr));
    if (!attr)
        goto invalid;

    if (spdk_json_decode_object(params,
            spdk_emu_device_attach_decoder,
            SPDK_COUNTOF(spdk_emu_device_attach_decoder),
            attr)) {
        SPDK_ERRLOG("Failed to decode parameters\n");
        goto free_attr;
    }

    sctx = spdk_emu_find_hotplug_manager(attr->emu_manager);
    if (!sctx) {
        SPDK_ERRLOG("Cannot find hotplug snap context\n");
        goto free_attr_strings;
    }

    if (!strncmp(attr->device_type, "nvme", 4))
        hp_attr.type = SNAP_NVME;
    else if (!strncmp(attr->device_type, "virtio_blk", 10))
        hp_attr.type = SNAP_VIRTIO_BLK;
    else if (!strncmp(attr->device_type, "virtio_fs", 9))
        hp_attr.type = SNAP_VIRTIO_FS;
    else {
        SPDK_ERRLOG("Invalid device type: %s\n", attr->device_type);
        goto free_attr_strings;
    }

    if (!(hp_attr.type & sctx->emulation_caps)) {
        SPDK_ERRLOG("Device type %s is not supported for emu manager %s\n",
                    attr->device_type, attr->emu_manager);
        goto free_attr_strings;
    }

    if (!(hp_attr.type & sctx->hotplug.supported_types)) {
        SPDK_ERRLOG("Device type %s hotplug is not supported for emu manager %s\n",
                    attr->device_type, attr->emu_manager);
        goto free_attr_strings;
    }

    if (attr->virtio_blk.transitional_device &&
        !sctx->hotplug.virtio_transitional_device_hotplug) {
        SPDK_ERRLOG("virtio blk transitional device hotplug is not supported for emu manager %s\n",
                    attr->emu_manager);
        goto free_attr_strings;
    }

    if (attr->total_vf > sctx->hotplug.max_total_vfs) {
        SPDK_ERRLOG("Cannot open %d VFs, max is %d\n", attr->total_vf,
                    sctx->hotplug.max_total_vfs);
        goto free_attr_strings;
    }
    hp_attr.max_vfs = attr->total_vf;

    spdk_emu_hotplug_init_pci(&hp_attr, attr);
    if (hp_attr.type == SNAP_NVME)
        err = spdk_emu_hotplug_init_nvme(&hp_attr, attr, &sctx->nvme_caps);
    else if (hp_attr.type == SNAP_VIRTIO_BLK)
        err = spdk_emu_hotplug_init_virtio_blk(&hp_attr, attr,
                                               &sctx->virtio_blk_caps);
    else
        err = spdk_emu_hotplug_init_virtio_fs(&hp_attr, attr,
                                              &sctx->virtio_fs_caps);
    if (err) {
        SPDK_ERRLOG("Failed to init hotplug attributes\n");
        goto free_attr_strings;
    }

    hotplug = snap_hotplug_pf(sctx, &hp_attr);
    if (!hotplug) {
        SPDK_ERRLOG("Failed to hotplug device\n");
        if (attr->pci.num_msix)
            SPDK_WARNLOG("Please verify num_msix value provided does not "
                         "exceed value configured in mlxconfig\n");
        goto free_attr_strings;
    }

    SPDK_NOTICELOG("A new %s PCI function was attached successfully\n",
                   attr->device_type);
    w = spdk_jsonrpc_begin_result(request);
    spdk_json_write_object_begin(w);
    spdk_json_write_named_string(w, "emulation_manager",
                                 ibv_get_device_name(sctx->context->device));
    spdk_json_write_named_string(w, "emulation_type", attr->device_type);
    spdk_json_write_named_string(w, "pci_type", "physical function");
    spdk_json_write_named_uint32(w, "pci_index", hotplug->id);
    spdk_json_write_object_end(w);
    spdk_jsonrpc_end_result(request, w);

    if (hp_attr.type == SNAP_VIRTIO_BLK) {
        free(attr->virtio_blk.bdev);
        free(attr->virtio_blk.bdev_type);
    }
    if (hp_attr.type == SNAP_VIRTIO_FS) {
        free(attr->virtio_fs.bdev);
        free(attr->virtio_fs.bdev_type);
    }
    free(attr->device_type);
    free(attr->emu_manager);
    free(attr);
    return;

free_attr_strings:
    if (hp_attr.type == SNAP_VIRTIO_BLK) {
        free(attr->virtio_blk.bdev);
        free(attr->virtio_blk.bdev_type);
    }
    if (hp_attr.type == SNAP_VIRTIO_FS) {
        free(attr->virtio_fs.bdev);
        free(attr->virtio_fs.bdev_type);
    }

    free(attr->device_type);
    free(attr->emu_manager);
free_attr:
    free(attr);
invalid:
    spdk_jsonrpc_send_error_response(request,
                     SPDK_JSONRPC_ERROR_INVALID_PARAMS,
                     "Invalid parameters");
}

SPDK_RPC_REGISTER("emulation_device_attach",
                  spdk_emu_rpc_device_attach, SPDK_RPC_RUNTIME)

#define MAX_EXECV_ARG_LEN                   128
#define MAX_RPC_PROGRAM_ARGS                64
#define MAX_PROGRAM_DEFAULT_OUTPUT_LEN      (32 * 1024)
#define MAX_PROGRAM_MAX_OUTPUT_LEN          (4 * MAX_PROGRAM_DEFAULT_OUTPUT_LEN)

/* return buffer when succeed both on exec program and read output,
 * return NULL otherwise. */
static char *spdk_exec_program(char **execv_args, int *status)
{
    int fd[2];
    pid_t pid, wpid;
    ssize_t nbytes, total_read, output_len;
    char *output, *tmp;

    if (pipe(fd)) {
        SPDK_ERRLOG("Failed to open pipe. errno:%d\n", errno);
        goto out;
    }

    pid = fork();
    if (pid < 0) {
        SPDK_ERRLOG("Failed to do fork. errno:%d\n", errno);
        goto close_pipe;
    } else if (pid == 0) {  /* child process */
        close(fd[0]);
        dup2(fd[1], 1);  /* redirect stdout */
        dup2(fd[1], 2);  /* redirect stderr */
        if(execv(execv_args[0], execv_args) < 0)
            SPDK_ERRLOG("child process invoke execv() failed, errno:%d\n", errno);
        exit(EXIT_FAILURE);
    }

    /* parent process */
    wpid = wait(status);
    if (wpid == -1) {
        if (errno == EINTR)
            SPDK_ERRLOG("wait child process failed due to parent got a signal\n");
        else
            SPDK_ERRLOG("wait child process failed. errno:%d\n", errno);
        goto close_pipe;
    }

    close(fd[1]);

    if (*status)
        SPDK_ERRLOG("Child Process Returned with error: %d\n", *status);
    else
        SPDK_NOTICELOG("Child process finished successfully\n");

    output_len = MAX_PROGRAM_DEFAULT_OUTPUT_LEN;
    output = calloc(1, output_len);
    if (!output) {
        SPDK_ERRLOG("alloc memory for child process `output` failed\n");
        goto close_fd0;
    }

    /* Try to read child process output no matter what is its return status */
    /* fd[0] will read output of `stdout` from child process
     * when program execute status is 0, otherwisw will read
     * the output of `stderr` from child process. */
    total_read = 0;
    while ((nbytes = read(fd[0], output + total_read,
                          output_len - total_read)) > 0) {
        total_read += nbytes;
        if (total_read >= output_len && total_read < MAX_PROGRAM_MAX_OUTPUT_LEN) {
            output_len = output_len * 2;
            tmp = realloc(output, output_len);
            if (!tmp) {
                SPDK_ERRLOG("realloc output buffer to size:%ld failed\n", output_len);
                free(output);
                goto close_fd0;
            }
            output = tmp;
        } else if (total_read >= MAX_PROGRAM_MAX_OUTPUT_LEN) {
            SPDK_ERRLOG("output from child process exceed limitation(%d)\n", MAX_PROGRAM_MAX_OUTPUT_LEN);
            free(output);
            goto close_fd0;
        }
    }

    output[total_read] = '\0'; /* buffer from realloc() is not set to zero */
    close(fd[0]);

    if (nbytes < 0) {
        SPDK_ERRLOG("read() child process output failed. errno:%d\n", errno);
        free(output);
        goto out;
    }

    return output;

close_pipe:
    close(fd[1]);

close_fd0:
    close(fd[0]);

out:
    return NULL;
}

struct rpc_args_list {
    size_t num_args;
    char *args[MAX_RPC_PROGRAM_ARGS];
};

static int spdk_decode_args_list(const struct spdk_json_val *val, void *out)
{
    struct rpc_args_list *args = out;

    return spdk_json_decode_array(val, spdk_json_decode_string, args->args,
                                  MAX_RPC_PROGRAM_ARGS, &args->num_args, sizeof(char *));
}

struct storage_admin_rpc_attr {
    char *protocol;
    char *op;
    char *policy;
    char *qn;
    char *hostqn;
    struct rpc_args_list paths;
    char *dev;
    char *type;
};

static const struct spdk_json_object_decoder
storage_admin_rpc_decoder[] = {
    {
        "protocol",
        offsetof(struct storage_admin_rpc_attr, protocol),
        spdk_json_decode_string
    },
    {
        "op",
        offsetof(struct storage_admin_rpc_attr, op),
        spdk_json_decode_string
    },
    {
        "policy",
        offsetof(struct storage_admin_rpc_attr, policy),
        spdk_json_decode_string,
        true
    },
    {
        "qn",
        offsetof(struct storage_admin_rpc_attr, qn),
        spdk_json_decode_string,
        true
    },
    {
        "hostqn",
        offsetof(struct storage_admin_rpc_attr, hostqn),
        spdk_json_decode_string,
        true
    },
    {
        "paths",
        offsetof(struct storage_admin_rpc_attr, paths),
        spdk_decode_args_list,
        true
    },
    {
        "dev",
        offsetof(struct storage_admin_rpc_attr, dev),
        spdk_json_decode_string,
        true
    },
    {
        "type",
        offsetof(struct storage_admin_rpc_attr, type),
        spdk_json_decode_string,
        true
    },
};

static char **
spdk_storage_admin_rpc_build_execv_args(struct storage_admin_rpc_attr *attr)
{
    size_t execv_nargs;
    int execv_next_arg = 0;
    char **execv_args;
    const char *script_path;
    int i;
    char arg_buf[MAX_EXECV_ARG_LEN] = {};

    /* Prepare execv command */
    execv_nargs = 2; // 1 for executable name, +1 for the null-terminated list
    execv_nargs += SPDK_COUNTOF(storage_admin_rpc_decoder); // All possible args
    execv_nargs += attr->paths.num_args - 1; // Additional paths, if exist
    if (execv_nargs > MAX_RPC_PROGRAM_ARGS) {
        SPDK_ERRLOG("Too many args provided\n");
        goto err;
    }

    execv_args = calloc(execv_nargs, sizeof(char *));
    if (!execv_args) {
        SPDK_ERRLOG("allocate memory for execv() argument failed\n");
        goto err;
    }

    script_path = getenv(STORAGE_ADMIN_SCRIPT_PATH_ENV);
    if (!script_path) {
        SPDK_ERRLOG("Unable to locate storage_admin script\n");
        goto free_args;
    } else if (access(script_path, F_OK | X_OK) == -1) {
        SPDK_ERRLOG("Script File %s does not exit or missing "
                    "execution permissions\n", script_path);
        goto free_args;
    }
    execv_args[execv_next_arg++] = strndup(script_path, MAX_EXECV_ARG_LEN);

    snprintf(arg_buf, MAX_EXECV_ARG_LEN - 1, "--protocol=%s", attr->protocol);
    execv_args[execv_next_arg++] = strndup(arg_buf, MAX_EXECV_ARG_LEN);

    snprintf(arg_buf, MAX_EXECV_ARG_LEN - 1, "--op=%s", attr->op);
    execv_args[execv_next_arg++] = strndup(arg_buf, MAX_EXECV_ARG_LEN);

    if (attr->policy) {
        snprintf(arg_buf, MAX_EXECV_ARG_LEN - 1, "--policy=%s", attr->policy);
        execv_args[execv_next_arg++] = strndup(arg_buf, MAX_EXECV_ARG_LEN);
    }

    if (attr->qn) {
        snprintf(arg_buf, MAX_EXECV_ARG_LEN - 1, "--qn=%s", attr->qn);
        execv_args[execv_next_arg++] = strndup(arg_buf, MAX_EXECV_ARG_LEN);
    }

    if (attr->hostqn) {
        snprintf(arg_buf, MAX_EXECV_ARG_LEN - 1, "--hostqn=%s", attr->hostqn);
        execv_args[execv_next_arg++] = strndup(arg_buf, MAX_EXECV_ARG_LEN);
    }

    for (i = 0; i < attr->paths.num_args; i++) {
        snprintf(arg_buf, MAX_EXECV_ARG_LEN - 1, "--path=%s", attr->paths.args[i]);
        execv_args[execv_next_arg++] = strndup(arg_buf, MAX_EXECV_ARG_LEN);
    }

    if (attr->dev) {
        snprintf(arg_buf, MAX_EXECV_ARG_LEN - 1, "--dev=%s", attr->dev);
        execv_args[execv_next_arg++] = strndup(arg_buf, MAX_EXECV_ARG_LEN);
    }

    if (attr->type) {
        snprintf(arg_buf, MAX_EXECV_ARG_LEN - 1, "--type=%s", attr->type);
        execv_args[execv_next_arg++] = strndup(arg_buf, MAX_EXECV_ARG_LEN);
    }

    return execv_args;

free_args:
    free(execv_args);
err:
    return NULL;
}

static void
spdk_rpc_free_execv_args(char **execv_args)
{
    char **arg = execv_args;

    while (*arg)
        free(*(arg++));
    free(execv_args);
}


static char *
spdk_exec_storage_admin_method(const void *arg, int *status)
{
    int i;
    char *output;
    struct storage_admin_rpc_attr *attr = NULL;
    char **execv_args;
    const struct spdk_json_val *params;

    attr = calloc(1, sizeof(*attr));
    if (!attr) {
        SPDK_ERRLOG("alloc memory for `attr` failed\n");
        goto out;
    }

    params = (struct spdk_json_val *)arg;
    if (spdk_json_decode_object(params, storage_admin_rpc_decoder,
            SPDK_COUNTOF(storage_admin_rpc_decoder), attr)) {
        SPDK_ERRLOG("Failed to decode parameters\n");
        goto free_attr;
    }

    execv_args = spdk_storage_admin_rpc_build_execv_args(attr);
    if (!execv_args)
        goto free_attr_strings;

    output = spdk_exec_program(execv_args, status);
    if (!output) {
        SPDK_ERRLOG("Failed to exec program\n");
        goto free_execv_args;
    }

    spdk_rpc_free_execv_args(execv_args);
    free(attr->protocol);
    free(attr->op);
    free(attr->policy);
    free(attr->qn);
    free(attr->hostqn);
    for (i = 0; i < attr->paths.num_args; i++)
        free(attr->paths.args[i]);
    free(attr->dev);
    free(attr->type);
    free(attr);

    return output;

free_execv_args:
    spdk_rpc_free_execv_args(execv_args);
free_attr_strings:
    free(attr->protocol);
    free(attr->op);
    free(attr->policy);
    free(attr->qn);
    free(attr->hostqn);
    for (i = 0; i < attr->paths.num_args; i++)
        free(attr->paths.args[i]);
    free(attr->dev);
    free(attr->type);
free_attr:
    free(attr);
out:
    return NULL;
}

static void spdk_storage_admin_rpc(struct spdk_jsonrpc_request *request,
                                   const struct spdk_json_val *params)
{
    int status;
    char *output;
    struct spdk_json_write_ctx *w;

    output = spdk_exec_storage_admin_method(params, &status);
    if (!output) {
        SPDK_ERRLOG("Exec storage_admin method failed\n");
        goto invalid;
    }

    w = spdk_jsonrpc_begin_result(request);
    spdk_json_write_object_begin(w);
    spdk_json_write_named_uint32(w, "status", status);
    spdk_json_write_named_string(w, "output", output);
    spdk_json_write_object_end(w);
    spdk_jsonrpc_end_result(request, w);

    free(output);

    return;

invalid:
    spdk_jsonrpc_send_error_response(request,
                     SPDK_JSONRPC_ERROR_INVALID_PARAMS,
                     "Invalid parameters");
}

SPDK_RPC_REGISTER("storage_admin", spdk_storage_admin_rpc, SPDK_RPC_RUNTIME)

struct network_admin_rpc_attr {
    char *op;
    char *action;
    int port;
    char *ipv4_addr;
    int ipv4_prefix;
    char *ipv6_addr;
    int ipv6_prefix;
    int mtu;
    char *ipv4_gateway;
    char *ipv6_gateway;
    struct rpc_args_list ipv4_nameservers;
    struct rpc_args_list ipv6_nameservers;
    struct rpc_args_list domains;
    char *trust;
    char *ecn;
    char *dscp2prio;
    char *prio_tc;
    char *pfc;
    int cable_len;
    char *prio2buffer;
    char *ratelimit;
    char *buffer_size;
    char *roce_accl;
#if NVME_CTRL_JSON_RPC_2_0_MJR > 1
    int vlan;
    char *skprio_up_egress;
    char *up_skprio_ingress;
#endif
};

static const struct spdk_json_object_decoder
network_admin_rpc_decoder[] = {
    {
        "op",
        offsetof(struct network_admin_rpc_attr, op),
        spdk_json_decode_string
    },
    {
        "action",
        offsetof(struct network_admin_rpc_attr, action),
        spdk_json_decode_string
    },
    {
        "port",
        offsetof(struct network_admin_rpc_attr, port),
        spdk_json_decode_int32,
        true
    },
    {
        "ipv4_addr",
        offsetof(struct network_admin_rpc_attr, ipv4_addr),
        spdk_json_decode_string,
        true
    },
    {
        "ipv4_prefix",
        offsetof(struct network_admin_rpc_attr, ipv4_prefix),
        spdk_json_decode_int32,
        true
    },
    {
        "ipv6_addr",
        offsetof(struct network_admin_rpc_attr, ipv6_addr),
        spdk_json_decode_string,
        true
    },
    {
        "ipv6_prefix",
        offsetof(struct network_admin_rpc_attr, ipv6_prefix),
        spdk_json_decode_int32,
        true
    },
    {
        "mtu",
        offsetof(struct network_admin_rpc_attr, mtu),
        spdk_json_decode_int32,
        true
    },
    {
        "ipv4_gateway",
        offsetof(struct network_admin_rpc_attr, ipv4_gateway),
        spdk_json_decode_string,
        true
    },
    {
        "ipv6_gateway",
        offsetof(struct network_admin_rpc_attr, ipv6_gateway),
        spdk_json_decode_string,
        true
    },
    {
        "ipv4_nameservers",
        offsetof(struct network_admin_rpc_attr, ipv4_nameservers),
        spdk_decode_args_list,
        true
    },
    {
        "ipv6_nameservers",
        offsetof(struct network_admin_rpc_attr, ipv6_nameservers),
        spdk_decode_args_list,
        true
    },
    {
        "domains",
        offsetof(struct network_admin_rpc_attr, domains),
        spdk_decode_args_list,
        true
    },
    {
        "trust",
        offsetof(struct network_admin_rpc_attr, trust),
        spdk_json_decode_string,
        true
    },
    {
        "ecn",
        offsetof(struct network_admin_rpc_attr, ecn),
        spdk_json_decode_string,
        true
    },
    {
        "dscp2prio",
        offsetof(struct network_admin_rpc_attr, dscp2prio),
        spdk_json_decode_string,
        true
    },
    {
        "prio_tc",
        offsetof(struct network_admin_rpc_attr, prio_tc),
        spdk_json_decode_string,
        true
    },
    {
        "pfc",
        offsetof(struct network_admin_rpc_attr, pfc),
        spdk_json_decode_string,
        true
    },
    {
        "cable_len",
        offsetof(struct network_admin_rpc_attr, cable_len),
        spdk_json_decode_int32,
        true
    },
    {
        "prio2buffer",
        offsetof(struct network_admin_rpc_attr, prio2buffer),
        spdk_json_decode_string,
        true
    },
    {
        "ratelimit",
        offsetof(struct network_admin_rpc_attr, ratelimit),
        spdk_json_decode_string,
        true
    },
    {
        "buffer_size",
        offsetof(struct network_admin_rpc_attr, buffer_size),
        spdk_json_decode_string,
        true
    },
    {
        "roce_accl",
        offsetof(struct network_admin_rpc_attr, roce_accl),
        spdk_json_decode_string,
        true
    },
#if NVME_CTRL_JSON_RPC_2_0_MJR > 1
    {
        "vlan",
        offsetof(struct network_admin_rpc_attr, vlan),
        spdk_json_decode_int32,
        true
    },
    {
        "skprio_up_egress",
        offsetof(struct network_admin_rpc_attr, skprio_up_egress),
        spdk_json_decode_string,
        true
    },
    {
        "up_skprio_ingress",
        offsetof(struct network_admin_rpc_attr, up_skprio_ingress),
        spdk_json_decode_string,
        true
    },
#endif
};

static char **
spdk_network_admin_rpc_build_execv_args(struct network_admin_rpc_attr *attr)
{
    size_t execv_nargs;
    int execv_next_arg = 0;
    char **execv_args;
    const char *script_path;
    int i;
    char arg_buf[MAX_EXECV_ARG_LEN] = {};

    /* Prepare execv command */
    /* 1 for executable name, +1 for the null-terminated list */
    execv_nargs = 2;
    /* All possible args */
    execv_nargs += SPDK_COUNTOF(network_admin_rpc_decoder);
    /* Additional ipv4 name servers, if exist */
    execv_nargs += attr->ipv4_nameservers.num_args - 1;
    /* Additional ipv6 name servers, if exist */
    execv_nargs += attr->ipv6_nameservers.num_args - 1;
    /* Additional domains, if exist */
    execv_nargs += attr->domains.num_args - 1;
    if (execv_nargs > MAX_RPC_PROGRAM_ARGS) {
        SPDK_ERRLOG("Too many args provided\n");
        goto err;
    }

    execv_args = calloc(execv_nargs, sizeof(char *));
    if (!execv_args) {
        SPDK_ERRLOG("allocate memory for execv() argument failed\n");
        goto err;
    }

    script_path = getenv(NETWORK_ADMIN_SCRIPT_PATH_ENV);
    if (!script_path) {
        SPDK_ERRLOG("Unable to locate storage_admin script\n");
        goto free_args;
    } else if (access(script_path, F_OK | X_OK) == -1) {
        SPDK_ERRLOG("Script File %s does not exit or missing "
                    "execution permissions\n", script_path);
        goto free_args;
    }
    execv_args[execv_next_arg++] = strndup(script_path, MAX_EXECV_ARG_LEN);

    snprintf(arg_buf, MAX_EXECV_ARG_LEN - 1, "--op=%s", attr->op);
    execv_args[execv_next_arg++] = strndup(arg_buf, MAX_EXECV_ARG_LEN);

    snprintf(arg_buf, MAX_EXECV_ARG_LEN - 1, "--action=%s", attr->action);
    execv_args[execv_next_arg++] = strndup(arg_buf, MAX_EXECV_ARG_LEN);

    if (attr->port >= 0) {
        snprintf(arg_buf, MAX_EXECV_ARG_LEN - 1, "--port=%d", attr->port);
        execv_args[execv_next_arg++] = strndup(arg_buf, MAX_EXECV_ARG_LEN);
    }

    if (attr->ipv4_addr) {
        snprintf(arg_buf, MAX_EXECV_ARG_LEN - 1, "--ipv4_addr=%s", attr->ipv4_addr);
        execv_args[execv_next_arg++] = strndup(arg_buf, MAX_EXECV_ARG_LEN);
    }

    if (attr->ipv4_prefix >= 0) {
        snprintf(arg_buf, MAX_EXECV_ARG_LEN - 1, "--ipv4_prefix=%d", attr->ipv4_prefix);
        execv_args[execv_next_arg++] = strndup(arg_buf, MAX_EXECV_ARG_LEN);
    }

    if (attr->ipv6_addr) {
        snprintf(arg_buf, MAX_EXECV_ARG_LEN - 1, "--ipv6_addr=%s", attr->ipv6_addr);
        execv_args[execv_next_arg++] = strndup(arg_buf, MAX_EXECV_ARG_LEN);
    }

    if (attr->ipv6_prefix >= 0) {
        snprintf(arg_buf, MAX_EXECV_ARG_LEN - 1, "--ipv6_prefix=%d", attr->ipv6_prefix);
        execv_args[execv_next_arg++] = strndup(arg_buf, MAX_EXECV_ARG_LEN);
    }

    if (attr->mtu >= 0) {
        snprintf(arg_buf, MAX_EXECV_ARG_LEN - 1, "--mtu=%d", attr->mtu);
        execv_args[execv_next_arg++] = strndup(arg_buf, MAX_EXECV_ARG_LEN);
    }

    if (attr->ipv4_gateway) {
        snprintf(arg_buf, MAX_EXECV_ARG_LEN - 1, "--ipv4_gateway=%s", attr->ipv4_gateway);
        execv_args[execv_next_arg++] = strndup(arg_buf, MAX_EXECV_ARG_LEN);
    }

    if (attr->ipv6_gateway) {
        snprintf(arg_buf, MAX_EXECV_ARG_LEN - 1, "--ipv6_gateway=%s", attr->ipv6_gateway);
        execv_args[execv_next_arg++] = strndup(arg_buf, MAX_EXECV_ARG_LEN);
    }

    for (i = 0; i < attr->ipv4_nameservers.num_args; i++) {
        snprintf(arg_buf, MAX_EXECV_ARG_LEN - 1, "--ipv4_nameserver=%s",
                 attr->ipv4_nameservers.args[i]);
        execv_args[execv_next_arg++] = strndup(arg_buf, MAX_EXECV_ARG_LEN);
    }

    for (i = 0; i < attr->ipv6_nameservers.num_args; i++) {
        snprintf(arg_buf, MAX_EXECV_ARG_LEN - 1, "--ipv6_nameserver=%s",
                 attr->ipv6_nameservers.args[i]);
        execv_args[execv_next_arg++] = strndup(arg_buf, MAX_EXECV_ARG_LEN);
    }

    for (i = 0; i < attr->domains.num_args; i++) {
        snprintf(arg_buf, MAX_EXECV_ARG_LEN - 1, "--domain=%s",
                 attr->domains.args[i]);
        execv_args[execv_next_arg++] = strndup(arg_buf, MAX_EXECV_ARG_LEN);
    }

    if (attr->trust) {
        snprintf(arg_buf, MAX_EXECV_ARG_LEN - 1, "--trust=%s", attr->trust);
        execv_args[execv_next_arg++] = strndup(arg_buf, MAX_EXECV_ARG_LEN);
    }

    if (attr->ecn) {
        snprintf(arg_buf, MAX_EXECV_ARG_LEN - 1, "--ecn=%s", attr->ecn);
        execv_args[execv_next_arg++] = strndup(arg_buf, MAX_EXECV_ARG_LEN);
    }

    if (attr->dscp2prio) {
        snprintf(arg_buf, MAX_EXECV_ARG_LEN - 1, "--dscp2prio=%s", attr->dscp2prio);
        execv_args[execv_next_arg++] = strndup(arg_buf, MAX_EXECV_ARG_LEN);
    }

	if (attr->prio_tc) {
		snprintf(arg_buf, MAX_EXECV_ARG_LEN - 1, "--prio_tc=%s", attr->prio_tc);
		execv_args[execv_next_arg++] = strndup(arg_buf, MAX_EXECV_ARG_LEN);
	}

    if (attr->pfc) {
        snprintf(arg_buf, MAX_EXECV_ARG_LEN - 1, "--pfc=%s", attr->pfc);
        execv_args[execv_next_arg++] = strndup(arg_buf, MAX_EXECV_ARG_LEN);
    }

    if (attr->cable_len >= 0) {
        snprintf(arg_buf, MAX_EXECV_ARG_LEN - 1, "--cable_len=%d", attr->cable_len);
        execv_args[execv_next_arg++] = strndup(arg_buf, MAX_EXECV_ARG_LEN);
    }

    if (attr->prio2buffer) {
        snprintf(arg_buf, MAX_EXECV_ARG_LEN - 1, "--prio2buffer=%s", attr->prio2buffer);
        execv_args[execv_next_arg++] = strndup(arg_buf, MAX_EXECV_ARG_LEN);
    }

    if (attr->ratelimit) {
        snprintf(arg_buf, MAX_EXECV_ARG_LEN - 1, "--ratelimit=%s", attr->ratelimit);
        execv_args[execv_next_arg++] = strndup(arg_buf, MAX_EXECV_ARG_LEN);
    }

    if (attr->buffer_size) {
        snprintf(arg_buf, MAX_EXECV_ARG_LEN - 1, "--buffer_size=%s", attr->buffer_size);
        execv_args[execv_next_arg++] = strndup(arg_buf, MAX_EXECV_ARG_LEN);
    }

    if (attr->roce_accl) {
        snprintf(arg_buf, MAX_EXECV_ARG_LEN - 1, "--roce_accl=%s", attr->roce_accl);
        execv_args[execv_next_arg++] = strndup(arg_buf, MAX_EXECV_ARG_LEN);
    }

#if NVME_CTRL_JSON_RPC_2_0_MJR > 1
    if (attr->vlan >= -1) {
        snprintf(arg_buf, MAX_EXECV_ARG_LEN - 1, "--vlan=%d", attr->vlan);
        execv_args[execv_next_arg++] = strndup(arg_buf, MAX_EXECV_ARG_LEN);
    }

    if (attr->skprio_up_egress) {
        snprintf(arg_buf, MAX_EXECV_ARG_LEN - 1, "--skprio_up_egress=%s", attr->skprio_up_egress);
        execv_args[execv_next_arg++] = strndup(arg_buf, MAX_EXECV_ARG_LEN);
    }

    if (attr->up_skprio_ingress) {
        snprintf(arg_buf, MAX_EXECV_ARG_LEN - 1, "--up_skprio_ingress=%s", attr->up_skprio_ingress);
        execv_args[execv_next_arg++] = strndup(arg_buf, MAX_EXECV_ARG_LEN);
    }
#endif

    return execv_args;

free_args:
    free(execv_args);
err:
    return NULL;
}


static char *
spdk_exec_network_admin_method(const void *arg, int *status)
{
    int i;
    char *output;
    struct network_admin_rpc_attr *attr = NULL;
    char **execv_args;
    const struct spdk_json_val *params;

    attr = calloc(1, sizeof(*attr));
    if (!attr) {
        SPDK_ERRLOG("alloc memory for `attr` failed\n");
        goto out;
    }

    attr->port = -1;
    attr->ipv4_prefix = -1;
    attr->ipv6_prefix = -1;
    attr->mtu = -1;
    attr->cable_len = -1;
#if NVME_CTRL_JSON_RPC_2_0_MJR > 1
    attr->vlan = -2;  /* -1 is a valid value to vlan */
#endif

    params = (struct spdk_json_val *)arg;
    if (spdk_json_decode_object(params, network_admin_rpc_decoder,
            SPDK_COUNTOF(network_admin_rpc_decoder), attr)) {
        SPDK_ERRLOG("Failed to decode parameters\n");
        goto free_attr;
    }

    execv_args = spdk_network_admin_rpc_build_execv_args(attr);
    if (!execv_args)
        goto free_attr_strings;

    output = spdk_exec_program(execv_args, status);
    if (!output) {
        SPDK_ERRLOG("Failed to exec program\n");
        goto free_execv_args;
    }

    spdk_rpc_free_execv_args(execv_args);
    free(attr->op);
    free(attr->action);
    free(attr->ipv4_addr);
    free(attr->ipv6_addr);
    free(attr->ipv4_gateway);
    free(attr->ipv6_gateway);
    for (i = 0; i < attr->ipv4_nameservers.num_args; i++)
        free(attr->ipv4_nameservers.args[i]);
    for (i = 0; i < attr->ipv6_nameservers.num_args; i++)
        free(attr->ipv6_nameservers.args[i]);
    for (i = 0; i < attr->domains.num_args; i++)
        free(attr->domains.args[i]);
    free(attr->trust);
    free(attr->ecn);
    free(attr->dscp2prio);
    free(attr->prio_tc);
    free(attr->pfc);
    free(attr->prio2buffer);
    free(attr->ratelimit);
    free(attr->buffer_size);
    free(attr->roce_accl);
#if NVME_CTRL_JSON_RPC_2_0_MJR > 1
    free(attr->skprio_up_egress);
    free(attr->up_skprio_ingress);
#endif
    free(attr);

    return output;

free_execv_args:
    spdk_rpc_free_execv_args(execv_args);
free_attr_strings:
    free(attr->op);
    free(attr->action);
    free(attr->ipv4_addr);
    free(attr->ipv6_addr);
    free(attr->ipv4_gateway);
    free(attr->ipv6_gateway);
    for (i = 0; i < attr->ipv4_nameservers.num_args; i++)
        free(attr->ipv4_nameservers.args[i]);
    for (i = 0; i < attr->ipv6_nameservers.num_args; i++)
        free(attr->ipv6_nameservers.args[i]);
    for (i = 0; i < attr->domains.num_args; i++)
        free(attr->domains.args[i]);
    free(attr->trust);
    free(attr->ecn);
    free(attr->dscp2prio);
    free(attr->prio_tc);
    free(attr->pfc);
    free(attr->prio2buffer);
    free(attr->ratelimit);
    free(attr->buffer_size);
    free(attr->roce_accl);
#if NVME_CTRL_JSON_RPC_2_0_MJR > 1
    free(attr->skprio_up_egress);
    free(attr->up_skprio_ingress);
#endif
free_attr:
    free(attr);
out:
    return NULL;
}

static void spdk_network_admin_rpc(struct spdk_jsonrpc_request *request,
                                   const struct spdk_json_val *params)
{
    int status;
    char *output;
    struct spdk_json_write_ctx *w;

    output = spdk_exec_network_admin_method(params, &status);
    if (!output) {
        SPDK_ERRLOG("Exec storage_admin method failed\n");
        goto invalid;
    }

    w = spdk_jsonrpc_begin_result(request);
    spdk_json_write_object_begin(w);
    spdk_json_write_named_uint32(w, "status", status);
    spdk_json_write_named_string(w, "output", output);
    spdk_json_write_object_end(w);
    spdk_jsonrpc_end_result(request, w);

    free(output);

    return;

invalid:
    spdk_jsonrpc_send_error_response(request,
                     SPDK_JSONRPC_ERROR_INVALID_PARAMS,
                     "Invalid parameters");
}

SPDK_RPC_REGISTER("network_admin", spdk_network_admin_rpc, SPDK_RPC_RUNTIME)

static void
spdk_emu_rpc_mempool_get_debugstat(struct spdk_jsonrpc_request *request,
                                    const struct spdk_json_val *params)
{
    int thread_id, counter_id;
    size_t *counters;
    struct spdk_json_write_ctx *w;

    if (!memp) {
        spdk_jsonrpc_send_error_response(request,
                SPDK_JSONRPC_ERROR_INVALID_REQUEST,
                     "Memory pool is not created");
        return;
    }

    w = spdk_jsonrpc_begin_result(request);
    spdk_json_write_object_begin(w);
    spdk_json_write_named_array_begin(w, "threads");

    for (thread_id = -1; thread_id < (int)spdk_io_mgr_get_num_threads(); thread_id++) {
        counters = memp_get_counters(memp, thread_id);

        spdk_json_write_object_begin(w);
        spdk_json_write_named_int32(w, "id", thread_id);

        for (counter_id = 0; counter_id < MEMP_COUNTERS_COUNT; counter_id++)
            spdk_json_write_named_uint64(w, memp_counter_name(counter_id), counters[counter_id]);

        spdk_json_write_object_end(w);
    }

    spdk_json_write_array_end(w);
    spdk_json_write_object_end(w);
    spdk_jsonrpc_end_result(request, w);

    return;
}

SPDK_RPC_REGISTER("mempool_get_debugstat",
        spdk_emu_rpc_mempool_get_debugstat, SPDK_RPC_RUNTIME)

#ifdef HAVE_SNAP
struct jsonrpc_request {
    struct spdk_json_val *version;
    struct spdk_json_val *method;
    struct spdk_json_val *params;
    struct spdk_json_val *id;

    struct spdk_json_val *values;
};

static struct nvme_snap_method spdk_snap_blacklist[] = {
    {
        "storage_admin",
        spdk_exec_storage_admin_method,
        NULL
    },
    {
        "network_admin",
        spdk_exec_network_admin_method,
        NULL
    },
    {
        NULL
    }
};

static int
capture_val(const struct spdk_json_val *val, void *out)
{
    const struct spdk_json_val **vptr = out;

    *vptr = val;
    return 0;
}

static const struct spdk_json_object_decoder jsonrpc_request_decoder[] = {
    {"jsonrpc", offsetof(struct jsonrpc_request, version), capture_val, true},
    {"method", offsetof(struct jsonrpc_request, method), capture_val},
    {"params", offsetof(struct jsonrpc_request, params), capture_val, true},
    {"id", offsetof(struct jsonrpc_request, id), capture_val, true},
};

static int spdk_check_json_request(struct jsonrpc_request *req)
{
    if (req->version && (req->version->type != SPDK_JSON_VAL_STRING ||
                         !spdk_json_strequal(req->version, "2.0"))) {
        goto invalid;
    }

    if (!req->method || req->method->type != SPDK_JSON_VAL_STRING) {
        goto invalid;
    }

    if (req->id) {
        if (req->id->type != SPDK_JSON_VAL_STRING &&
            req->id->type != SPDK_JSON_VAL_NUMBER &&
            req->id->type != SPDK_JSON_VAL_NULL) {
            goto invalid;
        }
    }

    if (req->params) {
        /* null json value is as if there were no parameters */
        if (req->params->type != SPDK_JSON_VAL_NULL) {
            if (req->params->type != SPDK_JSON_VAL_ARRAY_BEGIN &&
                req->params->type != SPDK_JSON_VAL_OBJECT_BEGIN) {
                goto invalid;
            }
        }
    }

    return 0;

invalid:
    return -1;
}

static struct jsonrpc_request *
spdk_parse_json_request(void *json_buf)
{
    ssize_t rc;
    size_t values_cnt, size;
    void *json_start, *json_end, *end;
    struct jsonrpc_request *req;

    json_start = strchr(json_buf, '{');
    if (!json_start) {
        SPDK_ERRLOG("Json format wrong\n");
        goto out;
    }

    json_end = strrchr(json_buf, '}');
    if (!json_end) {
        SPDK_ERRLOG("Json format wrong\n");
        goto out;
    }

    size = json_end - json_start + 1;

    rc = spdk_json_parse(json_start, size, NULL, 0, &end, 0);
    if (rc < 0) {
        SPDK_ERRLOG("Parsing(1st) JSON Configuration failed (%zd)\n", rc);
        goto out;
    }

    req = calloc(1, sizeof(struct jsonrpc_request));
    if (req == NULL) {
        SPDK_ERRLOG("allocate memory for `req` failed\n");
        goto out;
    }

    values_cnt = rc;
    req->values = calloc(values_cnt, sizeof(struct spdk_json_val));
    if (!req->values) {
        SPDK_ERRLOG("allocate memory for `req->values` failed\n");
        goto free_req;
    }

    rc = spdk_json_parse(json_start, size, req->values, values_cnt, &end, 0);
    if (rc != values_cnt) {
        SPDK_ERRLOG("Parsing(2nd) JSON configuration failed (%zd)\n", rc);
        goto free_values;
    }

    if (spdk_json_decode_object(req->values, jsonrpc_request_decoder,
                    SPDK_COUNTOF(jsonrpc_request_decoder),
                    req)) {
        SPDK_ERRLOG("Decode Json Object Failed\n");
        goto free_values;
    }

    if (spdk_check_json_request(req)) {
        SPDK_ERRLOG("Check for Json Request Failed\n");
        goto free_values;
    }

    return req;

free_values:
    free(req->values);

free_req:
    free(req);

out:
    return NULL;
}

static void spdk_free_jsonrpc_request(struct jsonrpc_request *req)
{
    if (req) {
        free(req->values);
        free(req);
    }
}

void *spdk_get_snap_blacklist_method(nvme_ctrl_t *ctrl)
{
    int i;
    void *json_buf;
    struct jsonrpc_request *req;
    struct snap_json_rpc_client *client;

    client = ctrl->rpc_client;
    json_buf = client->send_buf + client->send_offset;

    req = spdk_parse_json_request(json_buf);
    if (!req) {
        SPDK_ERRLOG("Parse Json Request Failed\n");
        return NULL;
    }

    for (i = 0; spdk_snap_blacklist[i].name != NULL; i++) {
        if (!strncmp(req->method->start, spdk_snap_blacklist[i].name, req->method->len)) {
            spdk_snap_blacklist[i].priv = req;
            return &spdk_snap_blacklist[i];
        }
    }

    spdk_free_jsonrpc_request(req);
    return NULL;
}

void spdk_handle_escape_char(char *orig, char *escaped, int len)
{
    int i, j;

    for (i = 0, j = 0; i < len; i++) {
        switch (orig[i]) {
        case '\a': // Alert
        case '\b': // Backspace
        case 0x1B: // Escape character
        case '\f': // Formfeed Page Break
        case '\n': // Newline
        case '\r': // Carriage Return
        case '\t': // Horizontal Tab
        case '\v': // Vertical Tab
        case '\\': // Backslash
        case '\'': // Apostrophe
        case '\"': // Double quotation mark
        case '\?': // Question mark
            escaped[j++] = '\\';
            /* Fallthrough */
        default:
            escaped[j++] = orig[i];
	};
    }
}

/* return 0 in success, return -1 otherwise */
int spdk_handle_snap_blacklist_method(void *ctx, nvme_ctrl_t *ctrl)
{
    int status;
    char *output;
    char *output_escaped;
    struct jsonrpc_request *req;
    struct snap_json_rpc_client_response *rsp;
    size_t output_len, json_rep_len;
    struct nvme_snap_method *rpc = (struct nvme_snap_method *)ctx;

    req = rpc->priv;

    output = rpc->rpc_method_cb(req->params, &status);
    if (!output) {
        SPDK_ERRLOG("Exec %s method failed\n", rpc->name);
        goto invalid;
    }

    output_len = strlen(output);
    while (output[output_len - 1] == '\n') {
        output[output_len - 1] = '\0';
        output_len--;
    }

    output_escaped = calloc(output_len * 2, sizeof(char));
    if (!output_escaped) {
        SPDK_ERRLOG("Allocate memory for output_escaped failed\n");
        goto free_output;
    }

    spdk_handle_escape_char(output, output_escaped, output_len);

    json_rep_len = strlen(output_escaped) + 128; /* + 128 for json format prefix */

    /* We support only 1 in-flight req/rsp pair. Override the previous rsp. */
    if (ctrl->rpc_rsp)
        snap_json_rpc_put_response(ctrl->rpc_rsp);

    rsp = calloc(1, sizeof(*rsp));
    if (!rsp) {
        SPDK_ERRLOG("Allocate memory for json response failed\n");
        goto free_output_escaped;
    }

    rsp->length = json_rep_len + 1;
    rsp->buf = calloc(1, rsp->length);
    if (!rsp->buf) {
        SPDK_ERRLOG("Allocate memory for response buffer failed\n");
        goto free_rsp;
    }

    snprintf(rsp->buf, json_rep_len,
             "{\"jsonrpc\": \"2.0\", \"id\": 1, \"result\": {\"status\": %d, \"output\": \"%s\"}}",
             status, output_escaped);

    ctrl->rpc_rsp = rsp;

    free(output);
    free(output_escaped);
    spdk_free_jsonrpc_request(req);
    snap_json_rpc_client_reset_send_buf(ctrl->rpc_client);

    return 0;

free_rsp:
    free(rsp);
free_output_escaped:
    free(output_escaped);
free_output:
    free(output);
invalid:
    spdk_free_jsonrpc_request(req);
    snap_json_rpc_client_reset_send_buf(ctrl->rpc_client);

    return -1;
}
#endif
