/*
 *   Copyright © 2021 NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 *     * Neither the name of Intel Corporation nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include <stdlib.h>
#include <stdbool.h>
#include <infiniband/verbs.h>

#include "compiler.h"
#include "nvme_emu_log.h"
#include "spdk_ext_io.h"

static struct spdk_memory_domain *spdk_ext_io_domain;
static struct spdk_memory_domain_ctx spdk_ext_io_domain_ctx;

static int spdk_ext_io_domain_translate_memory_cb(
        struct spdk_memory_domain *src_domain,
        void *src_domain_ctx, struct spdk_memory_domain *dst_domain,
        struct spdk_memory_domain_translation_ctx *dst_domain_ctx,
        void *addr, size_t len,
        struct spdk_memory_domain_translation_result *result)
{
    int domain_type;
    spdk_ext_io_ctx_t *ext_io_ctx;

    if (unlikely(!len))
        return 0;

    ext_io_ctx = src_domain_ctx;
    domain_type = spdk_memory_domain_get_dma_device_type(dst_domain);

    switch (domain_type) {
    case SPDK_DMA_DEVICE_TYPE_RDMA:
#ifdef HAVE_SPDK_DMA_DEVICE_TYPE_RDMA_TCP
    case SPDK_DMA_DEVICE_TYPE_RDMA_TCP:
#endif
        return spdk_ext_io_zcopy_memory_domain_translate_memory_cb(
                        src_domain, &ext_io_ctx->zcopy_ctx, dst_domain, dst_domain_ctx,
                        addr, len, result);
#ifdef HAVE_SPDK_DMA_DEVICE_TYPE_HOST
    case SPDK_DMA_DEVICE_TYPE_HOST:
        return spdk_ext_io_effdma_memory_domain_translate_memory_cb(
                        src_domain, &ext_io_ctx->effdma_ctx, dst_domain, dst_domain_ctx,
                        addr, len, result);
#endif
    default:
#ifdef DEBUG_ENABLED
       nvmx_assertv_always(false, "Unsupported dst domain");
#endif
       break;
    }

    return -1;
}

#ifdef HAVE_SPDK_MEMORY_DOMAIN_SET_MEMZERO
static void spdk_ext_io_memzero_write_comp(void *cpl_cb, void *cpl_cb_arg,
                                           int status)
{
    spdk_memory_domain_data_cpl_cb spdk_cpl_cb = cpl_cb;

    spdk_cpl_cb(cpl_cb_arg, status);
}

static int spdk_ext_io_domain_memzero_cb(struct spdk_memory_domain *domain,
        void *domain_ctx, struct iovec *iov, uint32_t iovcnt,
        spdk_memory_domain_data_cpl_cb cpl_cb, void *cpl_cb_arg)
{
    spdk_ext_io_ctx_t *ext_io_ctx = domain_ctx;
    spdk_ext_io_memzero_ctx_t *memzero_ctx = &ext_io_ctx->memzero_ctx;

    if (unlikely(!memzero_ctx->enabled))
        return -1;

    return memzero_ctx->ops->write(memzero_ctx->ops, iov, iovcnt,
                                   (void *)cpl_cb, cpl_cb_arg,
                                   spdk_ext_io_memzero_write_comp);
}
#endif

int spdk_ext_io_get_bdev_status(struct spdk_bdev *bdev, int domain_type)
{
    bool domain_supported = false;
    int bdev_domains_count, domain_idx;
    struct spdk_memory_domain **bdev_domains;

    bdev_domains_count = spdk_bdev_get_memory_domains(bdev, NULL, 0);

    if (bdev_domains_count < 0) {
        nvmx_error("Namespace %s: Failed to get memory domains count, rc %d",
                   spdk_bdev_get_name(bdev), bdev_domains_count);
        return -1;
    } else if (bdev_domains_count == 0)
        goto out;

    nvmx_debug("Namespace %s: Reported memory domains count %d",
                   spdk_bdev_get_name(bdev), bdev_domains_count);

    bdev_domains = calloc((size_t)bdev_domains_count, sizeof(*bdev_domains));
    if (!bdev_domains) {
        nvmx_error("Namespace %s: Failed to allocate memory domains",
                    spdk_bdev_get_name(bdev));
        return -ENOMEM;
    }

    spdk_bdev_get_memory_domains(bdev, bdev_domains, bdev_domains_count);

    for (domain_idx = 0; domain_idx < bdev_domains_count; domain_idx++) {
        if (spdk_memory_domain_get_dma_device_type(bdev_domains[domain_idx]) == domain_type) {
            domain_supported = true;
            break;
        }
    }

    free(bdev_domains);

out:
    return domain_supported ? 0 : -ENOTSUP;
}

int spdk_ext_io_init()
{
    int rc;

    spdk_ext_io_domain_ctx.size = sizeof(spdk_ext_io_domain_ctx);
    rc = spdk_memory_domain_create(&spdk_ext_io_domain,
#ifdef HAVE_SPDK_DMA_DEVICE_TYPE_HOST
                                   SPDK_DMA_DEVICE_TYPE_HOST,
#else
                                   SPDK_DMA_DEVICE_TYPE_RDMA,
#endif
                                   &spdk_ext_io_domain_ctx, NULL);
    if (rc) {
        nvmx_error("spdk_ext_io: Failed to create memory domain %d", rc);
        return -1;
    }

    spdk_memory_domain_set_translation(spdk_ext_io_domain,
                        spdk_ext_io_domain_translate_memory_cb);

#ifdef HAVE_SPDK_MEMORY_DOMAIN_SET_MEMZERO
    spdk_memory_domain_set_memzero(spdk_ext_io_domain,
                        spdk_ext_io_domain_memzero_cb);
#endif

    nvmx_debug("spdk_ext_io domain configured");

    return 0;
}

void spdk_ext_io_clear()
{
    spdk_memory_domain_destroy(spdk_ext_io_domain);
    spdk_ext_io_domain = NULL;
    nvmx_debug("spdk_ext_io cleared");
}

void spdk_ext_io_context_init(spdk_ext_io_ctx_t *ext_io_ctx, void *sdev,
                              void *key_table,
                              struct effdma_domain *domain,
                              memzero_ops_t *memzero_ops)
{
    memset(ext_io_ctx, 0, sizeof(spdk_ext_io_ctx_t));

    ext_io_ctx->ext_io_opts.size = sizeof(ext_io_ctx->ext_io_opts);
    ext_io_ctx->ext_io_opts.memory_domain = spdk_ext_io_domain;
    ext_io_ctx->ext_io_opts.memory_domain_ctx = ext_io_ctx;

    if (key_table)
        spdk_ext_io_zcopy_init(&ext_io_ctx->zcopy_ctx, key_table, sdev);

    if (domain)
        spdk_ext_io_effdma_init(&ext_io_ctx->effdma_ctx, domain);

    if (memzero_ops)
        spdk_ext_io_memzero_init(&ext_io_ctx->memzero_ctx, memzero_ops);
}

void spdk_ext_io_context_clear(spdk_ext_io_ctx_t *ext_io_ctx)
{
    spdk_ext_io_zcopy_clear(&ext_io_ctx->zcopy_ctx);
    spdk_ext_io_effdma_clear(&ext_io_ctx->effdma_ctx);
    spdk_ext_io_memzero_clear(&ext_io_ctx->memzero_ctx);
}
