#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <stdbool.h>
#include <unistd.h>
#include <errno.h>
#include <glob.h>
#include <arpa/inet.h>
#include <ifaddrs.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <linux/limits.h>
#include <linux/if_ether.h>
#include <libgen.h>
#include <ctype.h>

#include "config.h"
#include "utils.h"
#include "nvme_emu_log.h"
#include "snap_conf.h"

#define IB_PREFIX "/sys/class/infiniband/"
#define NET_PREFIX "/sys/class/net/"
#define SUFFIX "/device/resource"
#define SF_SUFFIX "/ports/1/gid_attrs/ndevs/0"
#define MLX_SUFFIX "/device/infiniband/*"

int compare_uint32_t(const void *p, const void *q)
{
    uint32_t l = *(uint32_t *)p;
    uint32_t r = *(uint32_t *)q;
    return l - r;
}

static int diff_files(const char *fpath1, const char *fpath2)
{
    FILE *f1, *f2;
    int c1, c2;
    int ret = 0;

    f1 = fopen(fpath1, "r");
    if (!f1) {
        nvmx_error("Failed to open file %s: %m", fpath1);
        return -errno;
    }

    f2 = fopen(fpath2, "r");
    if (!f2) {
        nvmx_error("Failed to open file %s: %m", fpath2);
        fclose(f1);
        return -errno;
    }

    do {
        c1 = getc(f1);
        c2 = getc(f2);
    } while ((c1 == c2) && c1 != EOF && c2 != EOF);

    if (c1 != EOF || c2 != EOF)
        ret = 1;

    fclose(f2);
    fclose(f1);
    return ret;
}

/* Checks whether a network interface is a representor,
 * Based on it's dev_port attribute.
 * The check follows the nvmx kernel convention given in
 * commit 505cb6ad on nvmx-kernel repository
 *
 * returns: On success - 1 if representor, 0 if not.
 *          On failure - ret < 0
 */
static int is_representor(const char *iface)
{
    FILE *fp;
    char dev_port_path[PATH_MAX];
    char line[256];
    int ret = 0;

    snprintf(dev_port_path, sizeof(dev_port_path),
             NET_PREFIX"%s/dev_port", iface);

    fp = fopen(dev_port_path, "r");
    if (!fp) {
        nvmx_error("Failed to open file %s: %m", dev_port_path);
        return -errno;
    }

    //dev_port should be 0 for device, and != for representors
    if (NULL == fgets(line, sizeof(line)-1, fp)) {
        nvmx_error("Failed to read file %s: %m", dev_port_path);
        ret = -errno;
        fclose(fp);
        return ret;
    }
    if (atoi(line) != 0)
        ret = 1;

    //dev_port should be single-lined file
    if(fgets(line, sizeof(line)-1, fp))
        ret = -EINVAL;

    fclose(fp);
    return ret;
}

int dev_to_iface(const char *dev, char *iface)
{
    char ibdev_res_path[PATH_MAX], netdev_res_path[PATH_MAX], **p;
    char ibdev_sf_res_path[PATH_MAX];
    char line[256];
    const char *_iface;
    glob_t glob_entry = {0,};
    int i = 0;
    FILE *fp;
    int ret = -ENODEV;

    snprintf(ibdev_sf_res_path, sizeof(ibdev_sf_res_path), IB_PREFIX"%s"SF_SUFFIX, dev);

    /* first try to check SF path */
    fp = fopen(ibdev_sf_res_path, "r");
    if (!fp) {
        nvmx_debug("Failed to open file %s: %m. fallback to non SF check",
                  ibdev_sf_res_path);
        goto fallback;
    };

    if (fgets(line, sizeof(line), fp) == NULL) {
        nvmx_debug("Failed to read file %s: %m. fallback to non SF check",
                  ibdev_sf_res_path);
        fclose(fp);
        goto fallback;
    } else {
        while (line[i++] != '\n');
        line[i - 1] = '\0';
        snprintf(netdev_res_path, sizeof(netdev_res_path), "%s%s",
                 NET_PREFIX, line);
        if (access(netdev_res_path, F_OK) != -1) {
            memcpy(iface, line, IFACE_MAX_LEN);
            ret = 0;
        } else {
            ret = -ENODEV;
        }
        fclose(fp);
        goto out;
    }

fallback:
    snprintf(ibdev_res_path, sizeof(ibdev_res_path), IB_PREFIX"%s"SUFFIX, dev);
    glob(NET_PREFIX"*", 0, 0, &glob_entry);
    p = glob_entry.gl_pathv;
    ret = -ENODEV;
    if (glob_entry.gl_pathc >= 1)
        for (i = 0; i < glob_entry.gl_pathc; i++, p++) {
            snprintf(netdev_res_path, sizeof(netdev_res_path), "%s"SUFFIX, *p);
            _iface = *p + strlen(NET_PREFIX);
            if (access(netdev_res_path, F_OK) != -1 &&
                diff_files(ibdev_res_path, netdev_res_path) == 0) {
                if (is_representor(_iface))
                    continue;
                strncpy(iface, _iface, IFACE_MAX_LEN - 1);
                ret = 0;
                break;
            }
        }
    globfree(&glob_entry);

out:
    return ret;
}

int ipaddr_to_iface(const char *addr, char *iface, size_t iface_sz)
{
    struct ifaddrs *addrs, *iap;
    struct sockaddr_in *sa;
    char ip_addr[INET_ADDRSTRLEN];
    bool found = false;

    getifaddrs(&addrs);
    for (iap = addrs; iap != NULL; iap = iap->ifa_next) {
        if (iap->ifa_addr) {
            sa = (struct sockaddr_in *)(iap->ifa_addr);
            if (inet_ntop(iap->ifa_addr->sa_family, (void *)&(sa->sin_addr),
                          ip_addr, (socklen_t)INET_ADDRSTRLEN)) {
                if (!strcmp(addr, ip_addr) && strlen(iap->ifa_name) < iface_sz) {
                    found = true;
                    strcpy(iface, iap->ifa_name);
                    break;
                }
            }
        }
    }
    freeifaddrs(addrs);

    return found ? 0: -ENODEV;
}

int iface_to_mlx_dev(const char *iface, char *dev, size_t dev_sz)
{
    glob_t glob_entry = {};
    char ibdev_path[PATH_MAX];
    int ret = -ENODEV;

    snprintf(ibdev_path, sizeof(ibdev_path),
             NET_PREFIX"%s"MLX_SUFFIX, iface);
    ret = glob(ibdev_path, 0, NULL, &glob_entry);
    if (ret)
        goto out;

    if (glob_entry.gl_pathc == 1 && strlen(glob_entry.gl_pathv[0]) < dev_sz) {
        strcpy(dev, basename(glob_entry.gl_pathv[0]));
        ret = 0;
    }

    globfree(&glob_entry);
out:
    return ret;
}

int iface_to_ipaddr(const char *iface, char *ip_addr, size_t ip_addr_sz)
{
    int fd;
    struct ifreq ifr;
    int err = 0;

    fd = socket(AF_INET, SOCK_DGRAM, 0);
    if (fd < 0) {
        nvmx_error("Failed to open socket");
        return -EINVAL;
    }

    ifr.ifr_addr.sa_family = AF_INET;
    strncpy(ifr.ifr_name, iface, IFNAMSIZ-1);
    err = ioctl(fd, SIOCGIFADDR, &ifr);
    close(fd);

    if (err) {
        nvmx_error("Failed to get IP address from interface: %s (%s)",
                   iface, strerror(errno));
        return err;
    }
    // YFYI inet_ntop supports both AF_INET and AF_INET6
    if (NULL == inet_ntop(ifr.ifr_addr.sa_family,
                          &(((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr),
                          ip_addr, (socklen_t)ip_addr_sz)) {
        nvmx_error("inet_ntop error: %s", strerror(errno));
        return errno;
    }
    return err;
}

int iface_rgid_to_rmac(const char *if_name, uint8_t *rgid, uint8_t *rmac)
{
    int                 s;
    struct arpreq       areq;
    struct sockaddr_in *sin;
    struct ifreq ifr;
    uint32_t ip_addr;

    if (rgid[0] != 0x0 || rgid[1] != 0x0 ||
        rgid[10] != 0xff || rgid[11] != 0xff) {
        nvmx_error("%s: dest gid is not ipv4", if_name);
        return -EINVAL;
    }

    /* Get an internet domain socket. */
    if ((s = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
        nvmx_error("%s: failed to open datagram socket: err=%d %m", if_name, errno);
        return -errno;
    }

    /* check if ip address is actually our own. */
    ifr.ifr_addr.sa_family = AF_INET;
    strncpy(ifr.ifr_name, if_name, IFNAMSIZ-1);
    if (ioctl(s, SIOCGIFADDR, &ifr) == -1)
        goto err;

    ip_addr = *(uint32_t *)&rgid[12];

    if (((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr.s_addr == ip_addr) {
        nvmx_debug("LOCAL IP address");
        if (ioctl(s, SIOCGIFHWADDR, &ifr) < 0)
            goto err;

        memcpy(rmac, &ifr.ifr_addr.sa_data, ETH_ALEN);
    }
    else {
        /* Make the ARP request. */
        memset(&areq, 0, sizeof(areq));
        sin = (struct sockaddr_in *) &areq.arp_pa;
        sin->sin_family = AF_INET;
        sin->sin_addr.s_addr = ip_addr;

        sin = (struct sockaddr_in *) &areq.arp_ha;
        sin->sin_family = ARPHRD_ETHER;

        strncpy(areq.arp_dev, if_name, IFACE_MAX_LEN - 1);

        if (ioctl(s, SIOCGARP, (caddr_t) &areq) == -1) {
            nvmx_error("%s: unable to make ARP request: err=%d %m", if_name, errno);
            goto err;
        }

        memcpy(rmac, &areq.arp_ha.sa_data, ETH_ALEN);
    }

    close(s);
    return 0;

err:
    close(s);
    return -errno;
}

int rdma_dev_name_to_id(const char *rdma_dev_name)
{
    char rdma_dev[MAX_RDMA_DEV_LEN];
    char *str, *next;
    int ret = 0;

    snprintf(rdma_dev, MAX_RDMA_DEV_LEN, "%s", rdma_dev_name);
    next = rdma_dev;

    do {
        str = next;
        while (str[0] != '\0' && !isdigit(str[0]))
            str++;

        if (str[0] == '\0')
            break;
        else
            ret = strtol(str, &next, 0);
    } while (true);

    return ret;
}

bool is_global_id_equal(nvme_id_ns_descriptor_t *id_desc,
                        int nidt, nvme_id_ns_global_t *nid)
{
    if (id_desc->nidt == nidt) {
        if (nidt == NVME_NS_IDENTIFIER_EUI64)
            return !memcmp(id_desc->nid.eui64.raw, nid->eui64.raw,
                           sizeof(nid->eui64.raw));
        else
            /*
             * All 16 bytes should be equal, so doesn't matter
             * which type will be used
             */
            return !memcmp(id_desc->nid.nguid.raw, nid->nguid.raw,
                           sizeof(nid->nguid.raw));
    }

    return false;
}

unsigned ulog2(uint32_t v)
{
   unsigned l;

   for (l = 0; v > 1; v >>= 1, l++);
   return l;
}

unsigned char char_to_hex(unsigned char c)
{
    if (c > '9')
        c += 9;
    return (c &= 0x0F);
}

